/****************************************************************************

DESCRIPTION: Linux driver header

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/

#ifndef ATDL_H
#define ATDL_H


#define EXPORT_SYMTAB

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>

#ifndef UTS_RELEASE
#include <generated/utsrelease.h>
#endif
#endif

#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/ctype.h>
#include <linux/percpu.h>

#include <asm/page.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <asm/pgtable.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/vmalloc.h>     /* vmap() */
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/moduleparam.h>
#include <linux/blkdev.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/completion.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/init.h> /* Linux module initialization */

#include <scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_tcq.h>

#include "atto_features.h"
#include "atto_ioctl.h"
#include "atsi.h"
#include "atto.h"
#include "atto_log.h"
#include "atto_cfg2.h"
#include "atdf.h"

#if defined(PROTO_SAS)
#include <linux/hdreg.h>
#endif

#ifdef FEAT_CDEV_IOCTL
#include <linux/cdev.h>
#endif

#ifdef FEAT_TRANSPORT
#include <scsi/scsi_transport.h>
#if defined(PROTO_FC)
#include <scsi/scsi_transport_fc.h>
#elif defined(PROTO_SAS)
#include <scsi/scsi_transport_sas.h>
#endif
#endif


#define ATTONODE_NAME       "ATTONode"
#define ATTO_VENDOR_ID      0x117C


/* MSI-X vector context */

struct atdl_msi_vector {
    struct atdl_adap*           hostadap;
    cpumask_t                   aff_mask;
    struct work_struct          q_work;
    struct workqueue_struct*    wq;
    int                         irq;        /* OS IRQ                   */
    u32                         vector;     /* Device relative vector   */
};


/* Host context registered with attocfg for any callbacks */

struct attohba_host {
    struct atdl_adap* hostadap;
};


/* The payload pointed to by scsi_device::hostdata (transport layer only) */

struct atdl_targ {
    u16     target_id;
    u8      port_num;
    u8      rsvd;
};


static inline struct atdl_targ* atdl_sd_to_targ( struct scsi_device* sdev )
{
    return (struct atdl_targ*) sdev->hostdata;
}


static inline u16 atdl_sd_to_targ_id( struct scsi_device* sdev )
{
#if defined(FEAT_TRANSPORT)
    struct atdl_targ* targ = atdl_sd_to_targ( sdev );

    if( targ )
        return targ->target_id;
#endif

    return (u16) sdev->id;
}


static inline u8 atdl_sd_to_port_num( struct scsi_device* sdev )
{
#if defined(FEAT_TRANSPORT)
    struct atdl_targ* targ = atdl_sd_to_targ( sdev );

    if( targ )
        return targ->port_num;
#endif

    return (u8) sdev->channel;
}


#if defined(FEAT_TRANSPORT) && defined(PROTO_SAS)

/* Represents a SAS PHY and wraps Linux sas_phy structure.  An appropriately
sized array of PHYs are attached to an expander/host device and linked to SAS ports.
*/

struct atdl_sas_phy {
    struct sas_phy*     phy;
    u64                 attach_sasaddr;
};


/* Represents a device of any type - host, expander or end device - and wraps 
the Linux sas_rphy for expanders and end device.  It's the central structure of
the topology.  A list of devices is maintained in the topology structure.
*/

struct atdl_sas_dev {
    struct list_head    link_topo;      /* Entry on topology list           */
    struct atdl_sas_dev*parent_dev;     /* Parent device                    */

    /* 'device' shortcut for Scsi_Host or sas_rphy depending on device type */
    struct device*      dev;            /* Linux device pointer             */

    /* Children object lists */
    struct atdl_sas_phy*phy_list;       /* Host/expander PHY array          */
    struct list_head    port_list;      /* List of ports on this device     */

    /* Device properties */
    struct sas_rphy*    rphy;           /* Device remote phy                */
    u64                 sasaddr;        /* SAS Address of this device       */
    u8                  num_phys;       /* Number of PHYs in this device    */
    u8                  dev_type;       /* Device type                      */
        #define ATDL_SAS_DT_HOST    0x01    /* Device is host               */
        #define ATDL_SAS_DT_END_DEV 0x02    /* Device is end device         */
        #define ATDL_SAS_DT_EXP     0x03    /* Device is expander           */

    u16                 sgid_pers;      /* Persistent SG ID                 */
    u16                 sgid_rel;       /* Relative SG ID                   */
    struct atdl_targ    target;         /* Common end device data           */
};


/* Represents a logical SAS port and wraps Linux sas_port structure */

struct atdl_sas_port {
    struct list_head    link_dev;       /* Entry of parent device list      */
    struct atdl_sas_dev*parent_dev;     /* Parent device                    */
    struct atdl_sas_dev*attach_dev;     /* Attached device                  */
    struct sas_port*    port;           /* Pointer to SAS port              */
};


/* Root SAS topology structure that lives in the adapter */

struct atdl_sas_topo {
    struct list_head    dev_list;       /* List of all devices in topology  */
    struct atdl_sas_dev*host_dev;       /* The host device in this topology */
    struct mutex        topo_mutex;     /* Mutex for accessing topology     */
    u16                 last_sg_id;     /* Last used SG ID                  */
        #define ATDL_XPORT_NUM_SG_IDS   1024
        #define ATDL_XPORT_INV_SG_ID    0xFFFF

    u16                 sg_ids[ATDL_XPORT_NUM_SG_IDS];
                                         /* Used SG IDs, by target ID       */
};

#endif /* FEAT_TRANSPORT && PROTO_SAS */


/* The payload pointed to by Scsi_Host::hostdata */

struct atdl_adap {
    struct atcl_adap*       coreadap;

    struct atsi_adap        adap;
    struct work_struct      dpc_work_s;
    struct workqueue_struct*dpc_work_q;
    struct atdl_msi_vector* msi_vectors;
    struct kmem_cache*      ioreq_pool;
    struct atdf_timer       timer;
    spinlock_t              mem_lock;
    struct completion       compl_adap_rst;
    struct completion       compl_bus_rst;
    struct completion       compl_shutdown;
    struct attohba_host     hba_host;   /* Cookie for HBA service routines */
    struct attocfg_host*    cfg_host;
    char                    rsrc_name[32];
    u32                     msi_cnt;
    u32                     flags;
        #define ATDL_AF_ALLOW_DPC       0x00000001
#if defined(FEAT_TRANSPORT)
        #define ATDL_AF_XPORT_STARTED   0x00000002
#endif

#ifdef FEAT_CDEV_IOCTL
    struct cdev             ioctl_cdev;
    struct device*          ioctl_dev;
#endif
#if defined(FEAT_TRANSPORT) && defined(PROTO_SAS)
    struct atdl_sas_topo    sas_topo;
#endif
#if defined(PROTO_FC)
    HBA_PORTATTRIBUTES      port_attr;
    HWIHBA_FC_PHY_STATS     phy_stats;
#ifdef FEAT_TRANSPORT
    struct fc_host_statistics
                            host_stats;
#endif
#endif
#if defined(FEAT_NVME_INI)
    struct atdl_nvme_ini*   nvme_ini;
#endif

#if defined(FEAT_NVME_TGT)
    struct atdl_nvme_tgt*   nvme_tgt;
#endif  /* FEAT_NVME_TGT */
};


static inline struct atdl_adap* atdl_sh_to_hadap( struct Scsi_Host *host )
{
    return (struct atdl_adap*) &host->hostdata;
}

static inline struct atcl_adap* atdl_sh_to_cadap( struct Scsi_Host *host )
{
    return atdl_sh_to_hadap( host )->coreadap;
}


/* atdl_req - (I/O) request information */

struct atdl_req {
    struct completion       compl;
} __attribute__((packed,aligned(8)));


/* Globals */

struct atdl_globals {
    struct task_struct*             worker_thread;
    wait_queue_head_t               worker_waitq;
    spinlock_t                      adap_lock;
    spinlock_t                      dwq_lock;
#ifdef FEAT_TRANSPORT
    struct scsi_transport_template* xport_templ;
#endif
#ifdef FEAT_PROC_IOCTL
    struct Scsi_Host*               proc_host;
    int                             proc_major;
#endif  /* FEAT_PROC_IOCTL */
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,16,0) && !IS_SLES_RELEASE_OR_LATER(15,5,0)
extern struct device_attribute*     g_atdl_host_attrs[];
#else
extern const struct attribute_group*g_atdl_host_groups[];
#endif
extern struct scsi_host_template    g_atdl_scsi_templ;
extern struct atdl_globals          g_atdl_vars;
extern struct atgi_connect_ops      g_atgi_ops;


/* Misc. functions */

int     atde_worker_execute( void* data );
void    atde_dpc_work_q( struct work_struct* work );


/* Interrupt interface */

#ifndef HW_NO_LEGACY_INTS
void    atdl_intr_legacy_disable( struct atdl_adap* hostadap );
int     atdl_intr_legacy_enable( struct atdl_adap* hostadap );
#endif
void    atdl_intr_msi_disable( struct atdl_adap* hostadap );
u8      atdl_intr_msi_enable( struct atdl_adap* hostadap );
void    atdl_intr_msix_disable( struct atdl_adap* hostadap );
u8      atdl_intr_msix_enable( struct atdl_adap* hostadap, u8 num_vect );


/* IOCTL interface */

#ifdef DIRECT2GPU
int     atdl_d2gpu_ioctl_handler( int cmd, void* arg );
int     atdl_d2gpu_ioctl_io( struct scsi_device* sd, void* arg );
#endif
int     atdl_ioctl_handler( struct atdl_adap* hostadap, struct scsi_device* sdev,
                            u32 cmd, void* userptr );
int     atdl_ioctl_legacy( struct atdl_adap* hostadap, u32 cmd, 
                           struct atdf_mapd_sgl* msgl );
#ifdef FEAT_CDEV_IOCTL
int     atdl_cdev_create( struct atdl_adap* hostadap, u32 instance );
void    atdl_cdev_destroy( struct atdl_adap* hostadap );
int     atdl_cdev_rsrc_alloc( void );
void    atdl_cdev_rsrc_free( void );
#endif


/* NVMe interface */

#if defined(FEAT_NVME_INI)
int     atdl_nvme_ini_lport_create( struct atdl_adap* hostadap, struct device* dev );
void    atdl_nvme_ini_lport_destroy( struct atdl_adap* hostadap );
void    atdl_nvme_ini_lport_free( struct atdl_adap* hostadap );
void    atdl_nvme_rpt_targ_change( struct atsi_dev_work_op* dwo );
#endif
#if defined(FEAT_NVME_TGT)
int     atdl_nvme_tgt_tport_create( struct atdl_adap* hostadap, struct device *dev );
void    atdl_nvme_tgt_tport_destroy( struct atdl_adap* hostadap );
void    atdl_nvme_tgt_tport_free( struct atdl_adap* hostadap );
#endif


/* ATA interface */

#ifdef PROTO_ATA
int     atdl_ata_ioctl_handler( struct scsi_device *sd, u32 cmd, void* arg );
#endif


/* Logging interface */

/**
 * atdl_log_lvl() - Log a device event, filtering by ATTO level.
 *                  Message is prefixed with driver name.
 */

#define atdl_log_print( dev, lnx_lvl, atto_lvl, func_area, fmt, ... )           \
    do {                                                                        \
        if( (func_area & atto_log_mask) && (atto_lvl <= log_level) )            \
        {                                                                       \
            printk( lnx_lvl ATTO_DRVR_NAME ": %s: " fmt, dev, ##__VA_ARGS__ );  \
            atci_serdbg_print( fmt, ##__VA_ARGS__ );                            \
        }                                                                       \
    } while(0)

#define atdl_log_lvl( ha, lnx_lvl, atto_lvl, func_area, fmt, ... ) \
        atdl_log_print( (ha)->adap.log_str, lnx_lvl, atto_lvl, func_area, fmt, ##__VA_ARGS__ )

/* atdl_adap based wrappers for atdl_log_lvl() for each logging level */

#define atdl_log_alert( ha, func_area, fmt, ... ) \
        atdl_log_lvl( ha, KERN_ALERT, HWLOG_LVL_ALERT, func_area, fmt, ##__VA_ARGS__ )

#define atdl_log_crit( ha, func_area, fmt, ... ) \
        atdl_log_lvl( ha, KERN_CRIT, HWLOG_LVL_CRITICAL, func_area, fmt, ##__VA_ARGS__ )

#define atdl_log_err( ha, func_area, fmt, ... ) \
        atdl_log_lvl( ha, KERN_ERR, HWLOG_LVL_ERROR, func_area, fmt, ##__VA_ARGS__ )

#define atdl_log_warn( ha, func_area, fmt, ... ) \
        atdl_log_lvl( ha, KERN_WARNING, HWLOG_LVL_WARNING, func_area, fmt, ##__VA_ARGS__ )

#define atdl_log_note( ha, func_area, fmt, ... ) \
        atdl_log_lvl( ha, KERN_NOTICE, HWLOG_LVL_NOTICE, func_area, fmt, ##__VA_ARGS__ )

#define atdl_log_info( ha, func_area, fmt, ... ) \
        atdl_log_lvl( ha, KERN_INFO, HWLOG_LVL_INFO, func_area, fmt, ##__VA_ARGS__ )

#define atdl_log_dbg( ha, func_area, fmt, ... ) \
        atdl_log_lvl( ha, KERN_DEBUG, HWLOG_LVL_VERBOSE1, func_area, fmt, ##__VA_ARGS__ )


/* SCSI interface */

int     atdl_scsi_dev_set_state( struct atsi_dev_work_op* dwo, u16 lun, int verbose );
void    atdl_scsi_dev_upd_state( struct atdl_adap* hostadap,
                                 struct scsi_device* sdev,
                                 enum atsi_dev_action action );


/* Protocol Transport Layer interface */

#ifdef FEAT_TRANSPORT
void    atdl_xport_adap_free( struct atdl_adap* hostadap );
int     atdl_xport_adap_init( struct atdl_adap* hostadap );
void    atdl_xport_adap_stop( struct atdl_adap* hostadap );
void    atdl_xport_rpt_targ_change( struct atsi_dev_work_op* dwo );
int     atdl_xport_slave_alloc( struct atdl_adap* hostadap, struct scsi_device* sdev );
struct scsi_transport_template*
        atdl_xport_templ_alloc( void );
void    atdl_xport_templ_free( struct scsi_transport_template* stt );
#endif


/* Request interface */

/**
 * atdl_req_init() - Initialize request defaults
 */

static inline void atdl_req_init( struct atsi_req* req,
                                  struct atdl_adap* hostadap,
                                  int cpu_id )
{
    struct atdl_req* adreq = (struct atdl_req*) (req + 1);

    req->hostadap = hostadap;
    req->cpu_id = cpu_id;
    req->flags = 0;
    req->compl = &adreq->compl;
    req->rq = (struct atcl_req*) (adreq + 1);
}


/**
 * atdl_req_alloc() - Allocate a request
 * 
 * Returns: Pointer to the atsi_req (which is followed by PIOREQ)
 */

static inline struct atsi_req* atdl_req_alloc( struct atdl_adap* hostadap, int cpu_id )
{
    struct atsi_req* req;
    
    req = kmem_cache_alloc_node( hostadap->ioreq_pool, GFP_ATOMIC, cpu_to_node(cpu_id) );

    if( req )
        atdl_req_init( req, hostadap, cpu_id );

    return req;
}


/**
 * atdl_req_free() - Free a request
 */

static inline void atdl_req_free( struct atsi_req* req )
{
    kmem_cache_free( req->hostadap->ioreq_pool, req );
}


#endif /* ATDL_H */
