/****************************************************************************

DESCRIPTION: Core/driver layer interface

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef ATSI_H
#define ATSI_H


/* Driver configuration - uses driver specific compilation switch to further
configure the driver build.  Also see hwshared.h for additional features.

Protocols:
PROTO_FC    
PROTO_SAS
PROTO_NVME
PROTO_ATA

Features:
FEAT_BAD_BLK_EVT
FEAT_DISABLE_DEV_SCAN
FEAT_SAS_DEV_ORDER
FEAT_SRIOV
FEAT_VSA
FEAT_ZERO_EXP_CHANGE_CNT

Defined in makefile:
FEAT_NVME_INI
FEAT_NVME_TGT
FEAT_TRANSPORT
*/

#define FEAT_CDEV_IOCTL
#define FEAT_LEGACY_IOCTL

#define ATTO_DRVR_RAWNAME  ATTO_HBA_MOD_RAWNAME    /* Set by makefile */


#if defined(_MP_)
#define ATDL_DRVR_TYPE      " Multipathing "
#elif defined(DIRECT2GPU)
#define ATDL_DRVR_TYPE      " Direct2GPU "
#else
#define ATDL_DRVR_TYPE      " "
#endif


#ifdef CELERITY4
#define ATDL_ADAP_NAME      "ATTO Celerity 8Gb" ATDL_DRVR_TYPE "Fibre Channel Adapter"
#define PROTO_FC
#endif

#ifdef CELERITY5
#define ATDL_ADAP_NAME      "ATTO Celerity 16/32/64Gb" ATDL_DRVR_TYPE "Fibre Channel Adapter"
#define PROTO_FC
#define FEAT_SRIOV
#endif

#ifdef ESASHBA4
#define ATDL_ADAP_NAME      "ATTO ExpressSAS 12Gb" ATDL_DRVR_TYPE "Adapter"
#define FEAT_VSA
#define FEAT_BAD_BLK_EVT
#define FEAT_DISABLE_DEV_SCAN
#define FEAT_ZERO_EXP_CHANGE_CNT
#define PROTO_ATA
#define PROTO_SAS
#endif

#ifdef ESASHBA5
#define ATDL_ADAP_NAME      "ATTO ExpressSAS 12Gb GT" ATDL_DRVR_TYPE "Adapter"
#define FEAT_VSA
#define PROTO_ATA
#define PROTO_SAS
#define BRCM_SAS
#endif

#ifdef ESASHBA6
#define ATDL_ADAP_NAME      "ATTO ExpressSAS 24Gb" ATDL_DRVR_TYPE "Adapter"
#define FEAT_VSA
#define PROTO_ATA
#define PROTO_SAS
#define BRCM_SAS
#endif

#ifdef NVMESC1
#define ATDL_ADAP_NAME      "ATTO ExpressNVM" ATDL_DRVR_TYPE "NVMe Adapter"
#define PROTO_NVME
#undef FEAT_LEGACY_IOCTL
#define HW_NO_LEGACY_INTS
#endif

/* Derived from above and makefile */

#if defined(PROTO_SAS) && defined(FEAT_TRANSPORT)
#define FEAT_SAS_DEV_ORDER
#endif


/* In lieu of core/driver base type headers, inline a conditional section here */

#ifdef __KERNEL__

#include <linux/types.h>
#include <linux/interrupt.h>

#endif  /* __KERNEL__ */


#include <hwtypes.h>
#ifdef PROTO_FC
#include <hwfc.h>
#include <smhba.h>
#endif
#ifdef PROTO_SAS
#include <hwsas.h>
#include <hwsassmp.h>
#endif
#include <hwidata.h>
#include <hwpublic.h>


/* Misc. definitions */

#define ATSI_MAX_LUN       256
#define ATTO_DRVR_NAME      STRINGIFY( ATTO_DRVR_RAWNAME )
#define ATSI_MAX_ADAPTERS   136


#if defined(PROTO_FC)

/* See hwifctm.h */
#define CORE_FCTM_XFER_FLAG_AUTO_RSP     0x00000001
#define CORE_FCTM_XFER_FLAG_CONTIG       0x00000002
#define CORE_FCTM_XFER_FLAG_OS_SGL       0x00000004
#define CORE_FCTM_XFER_FLAG_WRITE_DATA   0x00000010

#if defined (_MP_)
#define ATSI_MAX_ID         (max_luns)
#else
#define ATSI_MAX_ID         (max_logins)
#endif

#elif defined(PROTO_SAS)
#define ATSI_MAX_ID         (num_targets)
#else
#define ATSI_MAX_ID         256
#endif


/* Forward declarations of kernel objects so the core layer can use typed pointers */

struct device;
struct pci_dev;
struct scsi_cmnd;
struct Scsi_Host;
struct scatterlist;
struct nvmefc_ls_req;
struct nvmefc_fcp_req;
struct completion;
struct kmem_cache;
#ifndef __KERNEL__
typedef struct __wait_queue_head wait_queue_head_t;
typedef struct spinlock spinlock_t;
#endif

/* Forward declarations of internal objects */

struct atdl_adap;
struct atdf_mapd_sgl;
typedef struct atcl_adap ADAPTER, *PADAPTER;
typedef struct atcl_req IOREQ, *PIOREQ;


/* Device Work Queue mechanism
 *
 * Device state and action to take as a result are bedfellows.  atsi_dev_action
 * essentially rebrands device states and provides a way to define other work types.
 */

enum atsi_dev_action {
    ATSI_DEV_ACT_REMOVE         = HWTS_NOT_PRESENT,
    ATSI_DEV_ACT_DEACTIVATE     = HWTS_INACTIVE,
    ATSI_DEV_ACT_RELOCATE       = HWTS_ACTIVE_LOCATION,
    ATSI_DEV_ACT_REACTIVATE     = HWTS_ACTIVE_INFO,
    ATSI_DEV_ACT_ACTIVATE       = HWTS_ACTIVE,
    ATSI_DEV_ACT_ADD            = HWTS_PRESENT,
    ATSI_DEV_ACT_RESCAN         = HWTS_LUN_CHANGE,
    ATSI_DEV_ACT_TRANSITION     = HWTS_IN_TRANSITION,
    ATSI_DEV_ACT_ONLINE         = 97,
    ATSI_DEV_ACT_OFFLINE        = 98,
};


enum atsi_dev_type {
    ATSI_DEV_TYPE_SCSI = HWTARG_TYPE_SCSI,
    ATSI_DEV_TYPE_NVME = HWTARG_TYPE_NVME,
    ATSI_DEV_TYPE_EXPANDER = HWTARG_TYPE_EXPANDER
};


struct atsi_dev_work_op {
    struct atsi_adap*                   adap;
    enum atsi_dev_action                action;
    enum atsi_dev_type                  type;
    u16                                 targ_id;
    u8                                  port_num;
};


struct atsi_dev_work_item {
    struct atsi_dev_work_item* volatile next;
    struct atsi_dev_work_item* volatile prev;
    struct atsi_dev_work_op             dev_op;
    u32                                 queue_time;
    u32                                 delay_time;
};


struct atsi_dev_work_q {
    struct atsi_dev_work_item* volatile first;
    struct atsi_dev_work_item* volatile last;
    struct atsi_dev_work_item*          head;
    spinlock_t*                         lock;
};


/* atsi_pci_id - Pseudo pci_device_id clone for passing PCI IDs across the core
 * interface.  The first six fields match so table entries in ospciids.h can
 * pull double duty.
 */

struct atsi_pci_id {
    u16 ven_id;
    u16 dev_id;
    u16 ss_ven_id;
    u16 ss_dev_id;
    u32 class_code;
    u32 class_mask;
    u8  rev_id;
};


/* atsi_adap - Adapter information */

struct atsi_adap {
    struct pci_dev*         pcidev;
    struct atcl_adap*       coreadap;
    struct atdl_adap*       hostadap;
    struct atdf_timer*      timer;
    spinlock_t*             mem_lock;
    struct completion*      compl_adap_rst;
    struct completion*      compl_bus_rst;
    u32*                    cpu_to_msi_vect;
    char*                   scratch_buf;
    u32                     scratch_len;
    u32                     timer_period;
    u32 volatile            ref_cnt;
    u32                     flags;
        #define ATSI_AF_ALLOW_WORK      0x00000001
        #define ATSI_AF_HW_INIT_DONE    0x00000002
        #define ATSI_AF_HW_INIT_FAIL    0x00000004
        #define ATSI_AF_SCSI_HOST_ADDED 0x00000008
        #define ATSI_AF_TERMINATING     0x00000010

    HWINIT_MSI              msi_cfg;
    struct atsi_pci_id      pci_id;
    char                    log_str[16];
    const char*             model_name;
    u8                      intr_mode;
        #define ATSI_INTR_MODE_UNCLAIMED    0
        #define ATSI_INTR_MODE_LEGACY       1
        #define ATSI_INTR_MODE_MSI          2
        #define ATSI_INTR_MODE_MSIX         3


    struct {
        u8 bus;
        u8 dev;
        u8 func;
        u8 devfunc;
        u32 domain;
#ifdef FEAT_SRIOV
        u8 vfcnt;
#endif
    } pci;

    /* Only valid after the Scsi_Host has been added to the system */

    struct Scsi_Host*       scsihost;
    unsigned short          hostnum;
};


/* atsi_req - (I/O) request information */

struct atsi_req {
    struct atdl_adap*       hostadap;
    struct atcl_req*        rq;
    struct scatterlist*     sgl;
    struct completion*      compl;
    u32                     sge_cnt;
    int                     cpu_id;
    u32                     timeout;
    u32                     data_len;
    u32                     flags;
        #define ATSI_RF_DATA_DIR_WRITE  0x00000001
        #define ATSI_RF_DATA_DIR_READ   0x00000002
        #define ATSI_RF_DMA_MAPPING     0x00000004
#ifdef _MP_
        #define ATSI_RF_TUR_POLL        0x40000000
#endif
#ifdef DIRECT2GPU
        #define ATSI_RF_DIRECT2GPU      0x80000000
#endif

    u16                     targ_id;

    union {
        struct {
            union {
                struct scsi_cmnd*       parent;
#ifdef DIRECT2GPU
                struct atsi_d2gpu_io*   d2g_parent;
#endif
            };

            u8*     cdb;
            u8*     sense_buf;
            u16     lun;
            u8      sense_len;
            u8      cdb_len;
#if defined(DIRECT2GPU) || defined(_MP_)
            u8      cdb_array[16];
#endif
#ifdef DIRECT2GPU
            u32     sg_offset;
#endif
            u8      dev_type;
#ifdef BRCM_SAS
            u32     dma_dir_cmd;
            u32     dma_dir_map;
#endif
        } scsi;

        struct {
            struct nvmefc_fcp_req* parent;

            void*   cmd_virt;
            void*   rsp_virt;
            u64     cmd_dma;
            u64     rsp_dma;
            u16     cmd_len;
            u16     rsp_len;
            u16     sqid;
        } nvme;

        struct {
            struct nvmefc_ls_req* parent;

            void*   req_virt;
            void*   rsp_virt;
            u64     req_dma;
            u64     rsp_dma;
            u32     req_len;
            u32     rsp_len;
        } nvmels;
    };
} __attribute__((packed,aligned(8)));


#if defined(PROTO_SAS)

/* SMP request data */

struct atsi_smp_task {
    u64     SasAddress;
    void*   Request;
    int     RequestLen;
    void*   Response;
    int     ResponseLen;
    u8      Result;       /* SMP Response Result */
};

#endif /* PROTO_SAS */


/* atsi_globals - shared globals */

struct atsi_globals {
    const char*             kernel_ver;
    struct atsi_adap*       adap_list[ATSI_MAX_ADAPTERS];
    struct kmem_cache*      dwi_pool;
    struct atsi_dev_work_q  dwq;
    u32                     page_size;
    u32                     cpu_cnt;
    u32                     node_cnt;
    u8                      adap_cnt;
};

extern struct atsi_globals g_atsi_vars;

static inline struct atsi_adap* atsi_get_adap( u32 index )
{
    if( index < ATSI_MAX_ADAPTERS
     && g_atsi_vars.adap_list[index]
     && (g_atsi_vars.adap_list[index]->flags & ATSI_AF_SCSI_HOST_ADDED) )
        return g_atsi_vars.adap_list[index];

    return NULL;
}

static inline struct atcl_adap* atsi_get_coreadap( u32 index )
{
    struct atsi_adap* adap = atsi_get_adap( index );

    if( adap )
        return adap->coreadap;

    return NULL;
}


/* Module parameters */

#ifdef HW_VSA_REBUILD
extern unsigned int raid_rebuild_mode;
#endif
#if defined(HW_TRACING)
extern unsigned int trace_buf_size;
extern unsigned int trace_buf_wrap;
#endif
#if defined(FEAT_TRANSPORT)
extern unsigned int use_transport_layer;
#endif
extern unsigned int log_level;
extern unsigned int atto_log_mask;
extern unsigned int change_notification;
#if defined(PROTO_SAS) || defined(PROTO_FC)
extern unsigned int num_sg_lists;
extern unsigned int cmd_per_lun;
extern unsigned int can_queue;
extern unsigned int quick_init;
extern unsigned int sgl_page_size;
extern unsigned int cmd_retry_count;
#endif
extern unsigned int max_xfer_size;
extern unsigned int cpu_affinity;
#ifdef PROTO_FC
extern unsigned int max_logins;
extern unsigned int rport_timeout;
#if defined(CELERITY4)
extern unsigned int force_low_dma;
extern unsigned int fab_alternate_exchanges;
#endif
#if defined(CELERITY5)
extern unsigned int multi_tgt_queue;
extern unsigned int reg_sgl_percent;
extern unsigned int eq_delay;
#endif
#endif /* PROTO_FC */

#ifdef HW_NPIV
extern unsigned int npiv_enable;
extern unsigned int max_npiv_vports;
extern unsigned int max_npiv_logins;
#endif

#ifdef FEAT_SRIOV
extern unsigned int sriov_numvfs;
extern unsigned int sriov_mode;
#define ATSI_SRIOV_MODE_DIS     0
#define ATSI_SRIOV_MODE_ALL     1
#define ATSI_SRIOV_MODE_PF      2
#endif

#if defined(HW_SCSI_TGT) || defined(HW_NVME_TGT)
extern unsigned int target_mode;
#define ATTO_TARGET_MODE_DISABLED   0
#define ATTO_TARGET_MODE_SCSI       1
#define ATTO_TARGET_MODE_NVME       2
#if defined(FEAT_NVME_TGT)
#define PARAM_TARGET_MODE_MAX 3
#else
#define PARAM_TARGET_MODE_MAX 1
#endif
#ifndef ESASHBA6
extern unsigned int multimode_ports;
#endif
extern unsigned int target_mode_link_delay;
#if defined(PROTO_FC)
extern unsigned int target_mode_mask;
extern unsigned int initiator_mode_mask;
extern unsigned int num_xcb;
extern unsigned int num_scb;
#elif defined(PROTO_SAS)
extern unsigned int num_tgt_req;
extern unsigned int initiator_phy_mask;
extern unsigned int target_phy_mask;
#endif
#endif
#if defined(FEAT_DISABLE_DEV_SCAN)
extern unsigned int disable_device_scan;
#endif
#if defined(FEAT_ZERO_EXP_CHANGE_CNT)
extern unsigned int zero_eecc;
#endif

#ifdef _MP_
extern unsigned int max_luns;
extern unsigned int mp_init_wait;
#ifdef SF_DEBUG
extern unsigned int atto_sflib_mask;
#endif
#endif /* _MP_ */

#if defined(PROTO_SAS)
extern unsigned int num_sas_addr;
extern unsigned int fw_log_size;
extern unsigned int num_targets;
#endif

#ifdef FEAT_SAS_DEV_ORDER
extern unsigned int sas_dev_order;
#endif


/* Timer interface */

void    atci_timer_execute( void* adapter );


/* DPC interface */

void    atci_dpc_execute( struct atcl_adap* coreadap );
void    atdi_dpc_schedule( struct atdl_adap* hostadap );


/* Interrupt interface */

#ifndef HW_NO_LEGACY_INTS
int     atci_intr_legacy_handler( struct atcl_adap* a );
#endif
void    atci_intr_msi_handler( struct atcl_adap* a, u32 vector );


/* Debugging interface */

#ifdef HW_DEBUG
void    atci_dbg_print( const char* fmt, ... );
void    atci_dbg_str( const char* str );
#else
#define atci_dbg_print( fmt, ... )
#define atci_dbg_str( str )
#endif

#ifdef SER_DEBUG
void    atci_serdbg_init( void );
void    atci_serdbg_print( const char* fmt, ... );
void    atci_serdbg_str( const char* str );
void    atci_serdbg_vprint( const char* fmt, va_list args );
#else
#define atci_serdbg_init()
#define atci_serdbg_print( fmt, ... )
#define atci_serdbg_str( str )
#define atci_serdbg_vprint( fmt, args )
#endif

#ifdef SER_DEFER
void    atci_serdbg_cache( void );
void    atci_serdbg_resume( void );
#else
#define atci_serdbg_cache()
#define atci_serdbg_resume()
#endif


/* Logging interface */

void atdi_log_print( u8 atto_lvl, char* fmt, ... ) __attribute__ ((format (printf, 2, 3)));

#define atdi_log_lvl( atto_lvl, func_area, fmt, arg... )           \
    do {                                                                \
        if( (func_area & atto_log_mask) && (atto_lvl <= log_level) )    \
            atdi_log_print( atto_lvl, ATTO_DRVR_NAME ": " fmt, ##arg );  \
    } while(0)

/* Wrappers for atdi_log_lvl() for each logging level */

#define atdi_log_alert( func_area, fmt, arg... ) \
        atdi_log_lvl( HWLOG_LVL_ALERT, func_area, fmt, ##arg )

#define atdi_log_crit( func_area, fmt, arg... ) \
        atdi_log_lvl( HWLOG_LVL_CRITICAL, func_area, fmt, ##arg )

#define atdi_log_err( func_area, fmt, arg... ) \
        atdi_log_lvl( HWLOG_LVL_ERROR, func_area, fmt, ##arg )

#define atdi_log_warn( func_area, fmt, arg... ) \
        atdi_log_lvl( HWLOG_LVL_WARNING, func_area, fmt, ##arg )

#define atdi_log_note( func_area, fmt, arg... ) \
        atdi_log_lvl( HWLOG_LVL_NOTICE, func_area, fmt, ##arg )

#define atdi_log_info( func_area, fmt, arg... ) \
        atdi_log_lvl( HWLOG_LVL_INFO, func_area, fmt, ##arg )

#define atdi_log_dbg( func_area, fmt, arg... ) \
        atdi_log_lvl( HWLOG_LVL_VERBOSE1, func_area, fmt, ##arg )


/* SCSI interface */

int     atdi_scsi_dev_set_state( struct atsi_dev_work_op* dwo, u16 lun );
void    atdi_scsi_dev_upd_state( struct atdl_adap* hostadap, struct atsi_req* req,
                                 enum atsi_dev_action action );
u8      atci_scsi_eh_abort( struct atcl_adap* a, struct scsi_cmnd* cmd );
u8      atci_scsi_eh_adap_reset( struct atcl_adap* a );
u8      atci_scsi_eh_bus_reset( struct atcl_adap* a );
u8      atci_scsi_eh_lun_reset( struct atcl_adap* a, u16 target, u16 lun );
u8      atci_scsi_eh_target_reset( struct atcl_adap* a, u16 target, u16 lun );
int     atci_scsi_req_start( struct atcl_adap* a, struct atcl_req* rq, struct atsi_req* req );

void    atdi_scsi_bus_reset_detected( struct atdl_adap* hostadap );
int     atdi_scsi_host_in_recovery( struct Scsi_Host* sh );
void    atdi_scsi_req_complete( struct atsi_req* req, int scsi_stat, u8 hw_sts, int resid );
u32     atdi_scsi_req_move_data( struct atsi_req* req, void* buf, u32 buf_len, u32 offset, int read );


/* NVMe interface */

#if defined(PROTO_FC)
#if defined(HW_NVME_INI) || defined(FEAT_NVME_INI)
int     atci_nvme_ini_start_ls( struct atcl_adap* a, struct atcl_req* rq, struct atsi_req* req );
int     atci_nvme_ini_start_io( struct atcl_adap* a, struct atcl_req* rq, struct atsi_req* req );
void    atci_nvme_ini_abort_io( struct atcl_adap* a, struct atcl_req* rq, struct atsi_req* req );
void    atci_nvme_ini_abort_ls( struct atcl_adap* a, struct atcl_req* rq, struct atsi_req* req );
#endif
void    atdi_nvme_ini_io_comp( struct atdl_adap* hostadap,
                               struct atsi_req* req,
                               u8 hw_sts, u16 rsp_len, u32 resid );
void    atdi_nvme_ini_ls_comp( struct atdl_adap* hostadap,
                               struct atsi_req* req, u8 hw_sts );

#if defined(HW_NVME_TGT) || defined(FEAT_NVME_TGT)
void    atci_nvme_tgt_abort_req( struct atcl_adap* a, void* fctmd_cmd_ctx );
int     atci_nvme_tgt_connect( struct atcl_adap* a, struct atdl_adap* hostadap );
void    atci_nvme_tgt_disconnect( struct atcl_adap* a );
void    atci_nvme_tgt_set_tpatmd_ctx( void* fctmd_cmd_ctx, void* tpatmd_xfer_ctx );
void    atci_nvme_tgt_send_io_rsp( void* fctmd_cmd_ctx, void* rsp_virt, u32 rsp_len );
void    atci_nvme_tgt_send_ls_rsp( void* fctmd_cmd_ctx, void* tpatmd_xfer_ctx,
                                   void* rsp_virt, u64 rsp_dma, u32 rsp_len );
void    atci_nvme_tgt_xfer_data( void* tpatmd_xfer_ctx, void* fctmd_cmd_ctx, void* pagelist,
                                 u32 reloffset, u32 len, u32 flags );
#endif
void    atdi_nvme_tgt_handle_abort( struct atdl_adap* hostadap, void* tpatmd_xfer_ctx );
int     atdi_nvme_tgt_handle_io( void* fctmd_cmd_ctx, struct atdl_adap* hostadap,
                                 void* cmd_iu, u32 cmd_iu_len, u16 qid );
int     atdi_nvme_tgt_handle_ls( void* fctmd_cmd_ctx, struct atdl_adap* hostadap,
                                 void* ls_req, u32 ls_req_len );
void    atdi_nvme_tgt_xfer_comp( void* FctmdCmdContext, void* tpatmd_xfer_ctx, u8 status );
#endif /* PROTO_FC */


/* Initialization interface */

void    atci_core_free( void );
int     atci_core_init( void );
void    atci_free_adapter( struct atcl_adap* a );
int     atci_init_adap_cfg( struct atcl_adap* a );
int     atci_init_adap_operational( struct atcl_adap* a );
int     atci_init_adap_mem( struct atcl_adap* a, struct atdl_adap* hostadap,
                            struct atsi_adap* adap, u32 instance );
int     atci_init_adap_resume( struct atcl_adap* a );
void    atci_stop_adapter( struct atcl_adap* a );

void    atdi_free_interrupts( struct atdl_adap* hostadap );
u8      atdi_init_interrupts( struct atdl_adap* hostadap, u8 max_vectors );
int     atdi_pci_mem_map( struct atdl_adap* hostadap, int* bars, int num_bars, void** maps );
void    atdi_pci_mem_unmap( struct atdl_adap* hostadap, int* bars, int num_bars, void** maps );

u32     atci_obj_size( u32 objid );

enum {
    ATSI_OBJ_ID_ADAPTER,
    ATSI_OBJ_ID_REQUEST,
};


/* Protocol utilities */

#ifdef PROTO_SAS
int     atci_get_dev_info( struct atcl_adap* a, u16 targid, PHWIHBA_SAS_DEV_INFO devinfo );
int     atci_get_dev_addr( struct atcl_adap* a, u16 targid, PSASADDR devaddr );
int     atci_get_exp_info( struct atcl_adap* a, u16 targid, PHWIHBA_EXP_INFO expinfo );
void    atci_get_host_addr( struct atcl_adap* a, PSASADDR sasaddr );
u8      atci_get_host_phys( struct atcl_adap* a );
int     atci_get_phy_info( struct atcl_adap* a, u8 phyid, PHWIHBA_SAS_PHY_INFO phyinfo );
int     atci_smp_command( struct atcl_adap* a, struct atsi_smp_task* task );
int     atci_smp_discover( struct atcl_adap* a, PSASADDR sasaddr, u8 phyid,
                           PHWIHBA_SAS_PHY_INFO phyinfo );
#endif

#ifdef PROTO_FC
int     atci_get_dev_info( struct atcl_adap* a, u16 targid, PHWIHBA_FC_DEV_INFO devinfo );
int     atci_get_phy_stats( struct atcl_adap* a, PHWIHBA_FC_PHY_STATS PhyStats );
int     atci_get_port_attr( struct atcl_adap* a, PHBA_PORTATTRIBUTES portattr );
int     atci_get_port_info( struct atcl_adap* a, u32 port_num, PHWIHBA_FC_PORT_INFO portinfo );
void    atci_get_sym_name( struct atcl_adap* a, char* dest, u32 len );
void    atci_reset_bus( struct atcl_adap* a );
#endif  /* PROTO_FC */


/* IOCTL interface */

u8      atci_ioctl_handler( struct atcl_adap* a, int cmd, void* ioctl_data, u32 len,
                            struct atdf_mapd_sgl* msgl );
u32     atci_ioctl_legacy_size( u32 ioctl_cmd, void* user_buf );

#ifdef HW_SCSI_TGT
u8      atci_tgt_ioctl_handler( struct atcl_adap* a, int cmd, void* ioctl_data );
#endif


/* Worker interface */

u32     atci_worker_execute( void );
int     atci_worker_has_work( void );
void    atdi_worker_schedule( u32 delay );


/* Misc. functions that live in the core driver */

int     atci_adap_addr_str_get( struct atcl_adap* a, int port, char* buf );
int     atci_adap_addr_str_set( struct atcl_adap* a, int port, const char* buf, int len );
char*   atci_dev_action_str( enum atsi_dev_action action );
u32     atci_proc_write( struct atcl_adap* a, const char* buffer, u32 len );
u32     atci_proc_read( struct atcl_adap* a );
void    atci_sflib_mask_set( u32 newval );


/* Misc. functions that live in the driver layer */

void    atdi_adap_dereference( struct atsi_adap* adap );
int     atdi_adap_reference( struct atsi_adap* adap );
#if defined(FEAT_VSA) || defined(_MP_)
void    atdi_fire_atto_evt( struct atevent_hba_hdr* data, u32 len );
#endif
#ifdef FEAT_BAD_BLK_EVT
void    atdi_fire_bad_blk_evt( void* data, u32 len );
#endif
void*   atdi_mem_dma_alloc( struct atdl_adap* hostadap, u32 size, u64* dma_addr );
void*   atdi_mem_dma_alloc_node( struct atdl_adap* hostadap, u32 size, u64* dma_addr, u32 vector );
void    atdi_mem_dma_flush( struct atdl_adap* hostadap, u32 size, void* cpu_addr, u64 dma_addr );
void    atdi_mem_dma_free( struct atdl_adap* hostadap, u32 size, void* cpu_addr, u64 dma_addr );
void    atdi_mem_dma_invalidate( struct atdl_adap* hostadap, u32 size, void* cpu_addr, u64 dma_addr );
struct atsi_req*
        atdi_req_alloc( struct atdl_adap* hostadap );
void    atdi_req_free( struct atsi_req* req );
void    atdi_rpt_targ_change( struct atsi_dev_work_op* dwo );
#ifdef FEAT_SAS_DEV_ORDER
int     atdi_xport_print_dev_list( struct atdl_adap* hostadap, char* buf );
#endif


/* Direct2GPU interface */

#ifdef DIRECT2GPU

struct atdl_d2gpu_mem;

struct atsi_d2gpu_io {
    struct atioctl_d2gpu_io d2gio;
    u16                     targ_id;
    u16                     lun;
    struct completion*      cmpl;
    volatile u32            count;
    struct atcl_req*        queue;
    u8                      hw_sts;
};

int     atci_d2gpu_start_io( struct atcl_adap* a, struct atsi_d2gpu_io* d2g_io_cmd );
int     atci_d2gpu_is_supported( void );
u32     atdi_d2gpu_get_sge( struct atdl_d2gpu_mem* gpumem, u32 offset, PPHYSADDR addr );

#endif /* DIRECT2GPU */


/* DPM interface */

#if defined(_MP_)
#define OS_DPM_HELP_QUICK_INIT \
"Enable quick link initialization.\n" \
"Default: Enabled"

#define OS_DPM_HELP_MAX_LUNS \
"Maximum number of multipathed SCSI LUNs across all targets.\n" \
"Default: 1024, Minimum: 512, Maximum: 32768"

#define OS_DPM_HELP_MP_INIT_WAIT \
"The time in seconds that the driver waits for a multipathing configuration\n" \
"to be loaded from the ATTO Config Tool before proceeding with defaults.\n" \
"Default: 60"
#elif !defined(PROTO_NVME)
#define OS_DPM_HELP_QUICK_INIT \
"Enable quick link initialization.\n" \
"Default: Disabled"
#endif /* _MP_ */

#if defined(CELERITY4)
#define OS_DPM_HELP_FORCE_LOW_DMA \
"Force all DMA mappings to below 4GB.  This may severely impact performance\n" \
"and should only be used on systems with more than 512GB of physical RAM or\n" \
"unusual physical memory layouts.\n" \
"Default: Disabled"
#endif

#if defined(PROTO_FC) && (defined(HW_SCSI_TGT) || defined(HW_NVME_TGT))
#define OS_DPM_HELP_INI_MODE_MASK \
"Mask of adapter channels that support initiator mode.  Bit 0\n" \
"corresponds to channel 1.\n" \
"Default: 0xFFFFFFFF (when initiator mode is enabled)"

#define OS_DPM_HELP_TGT_MODE_MASK \
"Mask of adapter channels that support target mode.  Bit 0\n" \
"corresponds to channel 1.\n" \
"Default: 0xFFFFFFFF (when target mode is enabled)"
#endif

#ifdef FEAT_SAS_DEV_ORDER
#define OS_DPM_HELP_SAS_DEV_ORDER \
"The display order of targets in procfs output.  Non-zero values add an sgID\n" \
"value to the display - a relative ID indicating the order in which the targets\n" \
"were reported to Linux.\n" \
"0 - ATTO-assigned target ID order.\n" \
"1 - The order devices were reported to Linux.\n" \
"2 - ATTO-assigned target ID order, sgID is dynamic based on current devices.\n" \
"3 - ATTO-assigned target ID order, sgID is persistent for each device.\n" \
"Default: 0."
#endif

#ifdef FEAT_SRIOV
#define OS_DPM_HELP_SRIOV_MODE \
"Control SR-IOV initialization.  Valid values are:\n" \
"0 - Disable SR-IOV.\n" \
"1 - Control virtual functions (VFs) and physical functions (PFs).\n" \
"2 - Control PFs only.  Typically, this will be set for the PF driver.\n" \
"The VFs created by the PF driver are assigned to VMs.  The VF driver loaded\n" \
"in the VM uses the default setting.\n" \
"Default: 1"
#endif

#define OS_DPM_HELP_DEV_CHG_NOTIFY \
"Enable reporting device changes after the initial bus scan.\n" \
"Default: True"

#define OS_DPM_HELP_CMD_RETRY_CNT \
"Set SCSI command retry count.  Overrides the per-device max_retries\n" \
"setting in sysfs.  Set to 0 to disable retries.\n" \
"Default: -1 (use sysfs setting)"

#define OS_DPM_HELP_CMD_PER_LUN \
"Maximum number of commands per SCSI LUN.\n" \
"Default 1024, Minimum: 1, Maximum: 2048"

#ifdef PROTO_FC
#define OS_DPM_HELP_MAX_XFER_SIZE \
"The maximum data transfer size for a command.\n" \
"Default: 512KB, Minimum: 64KB, Maximum: 32MB"
#else
#define OS_DPM_HELP_MAX_XFER_SIZE \
"The maximum data transfer size for a command.\n" \
"Default: 8MB, Minimum: 64KB, Maximum: 32MB"
#endif

#define OS_DPM_HELP_CPU_AFFINITY \
"Enable affinity between CPU and MSI-X vector for load balancing.\n" \
"0 - Disable load balancing\n" \
"1 - ATTO load balancing policy\n" \
"2 - Linux load balancing policy\n" \
"Default: 1"

#if defined(FEAT_TRANSPORT)
#define OS_DPM_HELP_XPORT_LAYER \
"Attach to the protocol transport layer.\n" \
"Default: Enabled."

#if defined(PROTO_FC)
#define OS_DPM_HELP_RPORT_TIMEOUT \
"Time (in seconds) before a missing FC rport device is removed by the OS.\n" \
"Default: 30"
#endif
#endif /* FEAT_TRANSPORT */

#if defined(HW_DEBUG) || defined(HW_TRACING)
#define OS_DPM_HELP_EVT_TYPE_MASK \
"Mask of event types to post to the system log. See atto.h for mask values.\n" \
"Default: 0xbfffffff"
#else
#define OS_DPM_HELP_EVT_TYPE_MASK \
"Mask of event types to post to the system log. See atto.h for mask values.\n" \
"Default: 0xbffffdff"
#endif

#ifdef SF_DEBUG
#define OS_DPM_HELP_SF_DBG_MASK \
"Mask of multipathing debug info to post to the system log. See sfdbg.h.\n" \
"Default: 0 (driver selects mask)"
#endif


#endif /* ATSI_H */
