/****************************************************************************

DESCRIPTION: Linux distribution compatibility header

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/

#if !defined(ATTO_DISTRO_H)
#define ATTO_DISTRO_H


#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif


#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
#error "This driver will only work on kernel version 3.10 or later"
#endif


/* RHEL/CentOS versioning */

#ifdef RHEL_RELEASE_CODE
#define IS_RHEL_RELEASE(a, b)             (RHEL_RELEASE_CODE == RHEL_RELEASE_VERSION(a,b))
#define IS_RHEL_RELEASE_OR_LATER(a, b)    (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(a,b))
#define IS_RHEL_RELEASE_BEFORE(a, b)      (RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(a,b))
#else
#define IS_RHEL_RELEASE(a, b)             (0)
#define IS_RHEL_RELEASE_OR_LATER(a, b)    (0)
#define IS_RHEL_RELEASE_BEFORE(a, b)      (0)
#endif

/* 
 * SLES Versioning
 *
 * The SLES_VERSION / SLES_VERSION_CODE macros exist in some SLES
 * releases, but not SLES 11.  If they are not defined the following
 * defines will create them.  This code needs to be updated everytime a new
 * SLES SP is released.
 *
 * SLES releases kernel updates without bumping the service pack number.
 * The kernel versions used may overlap between service packs, meaning
 * that SP2 may be released with kernel 3.12.32, while SP1 may eventually
 * get updated to 3.12.48. That makes the implementation below inherently 
 * broken, as we will assume a system running 3.12.48 is using the later (SP2)
 * release.
 *
 * The SLES kernels for each SP are listed at:
 * https://wiki.microfocus.com/index.php/SUSE/SLES/Kernel_versions
 */

#ifdef CONFIG_SUSE_KERNEL

#ifndef SLE_VERSION
#define SLE_VERSION(a,b,c) KERNEL_VERSION(a,b,c)
#endif

/* Do not modify the below comment! The install script uses this to define   */
/* SLES versions with kernels that have backports from later kernel versions */

/* ***SUSE Install script overrides*** */

#ifndef SLE_VERSION_CODE
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,14))
#define SLE_VERSION_CODE SLE_VERSION(12,4,0)

/* The kernel version for SP3 is defined as the SP3 kernel released on
   April 23, 2018. This prevents conflicts with SP2 which is the base for
   the OSS operating system as the last released kernel for SP2 was 4.4.121 */

#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,126))
#define SLE_VERSION_CODE SLE_VERSION(12,3,0)
#endif /* LINUX_VERSION_CODE == KERNEL_VERSION(x,y,z) */
#endif

#endif /* CONFIG_SUSE_KERNEL */


#ifdef SLE_VERSION_CODE
#define IS_SLES_RELEASE(a, b, c)             (SLE_VERSION_CODE == SLE_VERSION(a,b,c))
#define IS_SLES_RELEASE_OR_LATER(a, b, c)    (SLE_VERSION_CODE >= SLE_VERSION(a,b,c))
#define IS_SLES_RELEASE_BEFORE(a, b, c)      (SLE_VERSION_CODE < SLE_VERSION(a,b,c))
#else
#define IS_SLES_RELEASE(a, b, c)             (0)
#define IS_SLES_RELEASE_OR_LATER(a, b, c)    (0)
#define IS_SLES_RELEASE_BEFORE(a, b, c)      (0)
#endif


/* ***Debian Install script overrides*** */

#ifdef DEB_VERSION_CODE
#define IS_DEB_RELEASE(a, b, c)             (DEB_VERSION_CODE == DEB_VERSION(a,b,c))
#define IS_DEB_RELEASE_OR_LATER(a, b, c)    (DEB_VERSION_CODE >= DEB_VERSION(a,b,c))
#define IS_DEB_RELEASE_BEFORE(a, b, c)      (DEB_VERSION_CODE < DEB_VERSION(a,b,c))
#else
#define IS_DEB_RELEASE(a, b, c)             (0)
#define IS_DEB_RELEASE_OR_LATER(a, b, c)    (0)
#define IS_DEB_RELEASE_BEFORE(a, b, c)      (0)
#endif


#endif  /* ATTO_DISTRO_H */

