/****************************************************************************

DESCRIPTION: Supported features based on kernel version and CONFIG_ options

Copyright (C) 2018  ATTO TECHNOLOGY, INC.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*****************************************************************************/

#if !defined(ATTO_FEATURES_H)
#define ATTO_FEATURES_H


#include "atto_distro.h"


/* This is to get around a SLES 11.4 build issue.
 * IS_ENABLED is not defined in kernel versions less than 3.1,
 * so just define it if it isn't defined instead of adding kernel
 * checks everywhere.
 */

#ifndef IS_ENABLED
#define IS_ENABLED(x) 0
#endif

/* Remove proc IOCTL node for kernel 6.2 and later */

#if defined(CONFIG_PROC_FS) && LINUX_VERSION_CODE < KERNEL_VERSION(6,2,0)
#define FEAT_PROC_IOCTL
#endif

/* Just because a family supports it, doesn't mean the current kernel does.
 * See /lib/modules/`uname -r`/build/.config for names.
 */

#if defined (FEAT_NVME_INI)

#if !IS_ENABLED(CONFIG_NVME_FC)
#undef FEAT_NVME_INI
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4,15,0) \
      && !IS_RHEL_RELEASE_OR_LATER(7, 5)
#undef FEAT_NVME_INI
#endif  /* !IS_ENABLED(CONFIG_NVME_FC) */

#endif  /* FEAT_NVME_INI  */


#if defined (FEAT_NVME_TGT)

#if !IS_ENABLED(CONFIG_NVME_TARGET_FC)
#undef FEAT_NVME_TGT
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0) \
      && !IS_RHEL_RELEASE_OR_LATER(7, 5)
#undef FEAT_NVME_TGT
#endif  /* !IS_ENABLED(CONFIG_NVME_TARGET_FC) */

#endif  /* FEAT_NVME_TGT */


#endif  /* ATTO_FEATURES_H */
