/****************************************************************************

DESCRIPTION: Linux driver ioctl definitions

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2021
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef ATTO_IOCTL_H
#define ATTO_IOCTL_H


/* IOCTL commands - coordinated with legacy command codes below.  Only the
first two are passed to ioctl().  The remainder are set in the header.
*/

#define ATTO_IOCTL_CMD_USER             0x4580  /* Started from user space  */
#define ATTO_IOCTL_CMD_KRNL             0x4581  /* Started from kernel space*/
#define ATTO_IOCTL_CMD_ADAP_INFO        0x4582  /* ATIOCTL_ADAP_INFO        */
#define ATTO_IOCTL_CMD_MP               0x4583  /* MULTIPATH_IOCTL          */
#define ATTO_IOCTL_CMD_SCSI_DEV_INFO    0x4584  /* ATIOCTL_SCSI_DEV_INFO    */
#define ATTO_IOCTL_CMD_SMP              EXPRESS_HWIOCTL_SMP
#define ATTO_IOCTL_CMD_HBA              EXPRESS_HWIOCTL_HBA
#define ATTO_IOCTL_CMD_VSA              EXPRESS_HWIOCTL_VSA
#define ATTO_IOCTL_CMD_FCTM             FCTM_IOCTL_CONTROL
#define ATTO_IOCTL_CMD_SASTM            SASTM_IOCTL_CONTROL


/* IOCTL header preceding all commands that use ATTO_IOCTL_CMD_XXX */

typedef struct atioctl_hdr {
    u32     version;        /* 00 Header version                            */
        #define ATIHDR_VER_0    0

    u8      signature[12];  /* 04 Signature, NULL terminated                */
        #define ATIHDR_SIGNATURE    "atto_ioctl"

    u32     datalen;        /* 10 Data length following the header          */
    u32     ioctlcmd;       /* 14 Defines payload that follows this header  */
                            /*    See ATTO_IOCTL_CMD_XXX                    */
    u8      rsvd[0x28];     /* 18 Reserved                                  */
} ATIOCTL_HDR,              /* 40 Size of structure                         */
*PATIOCTL_HDR;


/* IOCTL payload for ATTO_IOCTL_CMD_SCSI_DEV_INFO.  This IOCTL must be sent to
devices attached to adapters (e.g. /dev/sda), not the adapters themselves.
This is essentially a replacement for SCSI_IOCTL_GET_IDLUN without truncation.
*/

typedef struct atioctl_scsi_dev_info {
    u32     host_num;       /* 00 SCSI host number                          */
    u32     channel;        /* 04 SCSI bus number or virtual port           */
    u32     atto_targ_id;   /* 08 ATTO's target ID assignment               */
    u32     os_targ_id;     /* 0C Linux's target ID assignment              */
    u64     lun;            /* 10 Logical unit number                       */
    u8      rsvd[8];        /* 18 Reserved                                  */
} ATIOCTL_SCSI_DEV_INFO,    /* 20 Size of structure                         */
*PATIOCTL_SCSI_DEV_INFO;


/* Direct2GPU interface (does not use ATIOCTL_HDR) */

#define ATTO_IOCTL_D2GPU_MEM_PIN        0x4590  /* atioctl_d2gpu_mem        */
#define ATTO_IOCTL_D2GPU_MEM_UNPIN      0x4591  /* atioctl_d2gpu_mem        */
#define ATTO_IOCTL_D2GPU_IO             0x4592  /* atioctl_d2gpu_io         */

typedef struct atioctl_d2gpu_mem {
    u32     signature;
        #define ATID2GPU_SIGNATURE  0x08675309

    u32     version;
        #define ATID2GPU_VERSION    0

    u32     size;
    void*   user_ptr;
    void*   handle;
} ATIOCTL_D2GPU_MEM,
*PATIOCTL_D2GPU_MEM;


typedef struct atioctl_d2gpu_io {
    unsigned long block;
    u32     block_count;
    u32     block_size;
    u32     flags;
        #define ATID2GPU_IO_WRITE       0x00000001 /* Write (1) or read (0) */
        #define ATID2GPU_IO_Q_MASK      0x00000006 /* Queueing type mask    */
        #define ATID2GPU_IO_Q_SIMPLE    0x00000000 /* Simple queueing       */
        #define ATID2GPU_IO_Q_ORDERED   0x00000002 /* Ordered queueing      */
        #define ATID2GPU_IO_Q_HEAD      0x00000004 /* Head-of-queue         */

    void*   d2g_handle;
    u32     d2g_offset;
    u32     blocks_per_req;
    u32     max_pending;
} ATIOCTL_D2GPU_IO,
*PATIOCTL_D2GPU_IO;


/*
 * Event Interface
 */

/* Port numbers for sockets */

#define ATTO_EVT_PORT_HBA1      54414
#define ATTO_EVT_PORT_HBA2      54415
#define ATTO_EVT_PORT_TARG      54416


/* General event header for ATTO HBA events.  See atevent_targ_hdr for additional
signatures used.
*/

typedef struct atevent_hba_hdr {
    u32     signature;
        #define ATTO_EVT_SIG_VDA        0x4154544F  /* Obsolete         */
        #define ATTO_EVT_SIG_MP         0x4F545441  /* atevent_mp       */
        #define ATTO_EVT_SIG_VSA        0x4F545442  /* atevent_vsa      */

    u8      bus_number;
    u8      devfn;
    u8      pad[2];
} ATEVENT_HBA_HDR,
*PATEVENT_HBA_HDR;


/* General event header for ATTO target events.  Note these events use a different
port than atevent_hba_hdr based events so signatures don't NEED to be coordinated,
but make them unique all the same.
*/

typedef struct atevent_targ_hdr {
    u32     signature;
        #define ATTO_EVT_SIG_TARG_ADD   0x47226A74
        #define ATTO_EVT_SIG_TARG_RMV   0x47226A75
        #define ATTO_EVT_SIG_BAD_BLK    0x47756E74

    u16     hba_index;
    u16     targ_id;
} ATEVENT_TARG_HDR,
*PATEVENT_TARG_HDR;


/*
 * Deprecated Legacy Interface 
 */

#define EXPRESS_IOCTL_SIGNATURE        "Express"
#define EXPRESS_IOCTL_SIGNATURE_SIZE   8

/* Structure definitions for IOCTLs */

typedef struct _express_ioctl_header {
    u8      Signature [EXPRESS_IOCTL_SIGNATURE_SIZE];
    u8      ReturnCode;

    #define IOCTL_SUCCESS               0   /* All good                     */

    /* Note: HWI_FCTM_STS_XXX status codes start at 0 so reserve 100 values */
    /* for HW layer interfaces.                                             */

    #define IOCTL_ERR_INVCMD          101   /* Unrecognized IOCTL code      */
    #define IOCTL_INIT_FAILED         102   /* Adapter init failed          */
    #define IOCTL_NOT_IMPLEMENTED     103   /* Unimplemented IOCTL          */
    #define IOCTL_BAD_CHANNEL         104   /* Invalid Channel number       */
    #define IOCTL_TARGET_OVERRUN      105   /* Data overrun (GET_INI_LIST)  */
    #define IOCTL_TARGET_NOT_ENABLED  106   /* Target Mode not enabled      */
    #define IOCTL_BAD_FLASH_IMGTYPE   107   /* Invalid flash image type     */
    #define IOCTL_OUT_OF_RESOURCES    108   /* Temporarily out of resources */
    #define IOCTL_GENERAL_ERROR       109   /* Unspecified error occurred   */
    #define IOCTL_INVALID_PARAM       110   /* Invalid parameter            */

    /* NOTE - if Channel == 0xFF, the request is */
    /* handled on the adapter it came in on.     */

    u8      Channel;
    u8      Retries;
    u8      Pad[1];

    /* NOTE - Recipients of this IOCTL type should NOT rely on this field.  */
    /* It was not a part of the initial API and is only used by the app     */
    /* layer as temporary storage of the payload length.                    */

    u32     DataLen;
} EXPRESS_IOCTL_HEADER,
*PEXPRESS_IOCTL_HEADER;


/* IOCTL cmd definitions */
/* Note that we don't use the size parameter in the IOCTL definitions,  */
/* because EXPRESS_IOCTL is too large to be contained in this param.    */
/* The struct associated with the code is listed after the definition   */

#define EXPRESS_IOCTL_MIN           0x4500
                                                /* IOCTL structure */
#define EXPRESS_OBSOLETE_4500       0x4500
#define EXPRESS_OBSOLETE_4501       0x4501
#define EXPRESS_OBSOLETE_4502       0x4502
#define EXPRESS_OBSOLETE_4503       0x4503
#define EXPRESS_IOCTL_GET_CHANNELS  0x4504      /* CHANNELLIST   */
#define EXPRESS_IOCTL_CHAN_INFO     0x4505      /* CHANNELINFO   */
#define EXPRESS_OBSOLETE_4506       0x4506
#define EXPRESS_OBSOLETE_4507       0x4507
#define EXPRESS_OBSOLETE_4508       0x4508
#define EXPRESS_OBSOLETE_4509       0x4509
#define EXPRESS_HWIOCTL_SMP         0x450A      /* HWIOCTL_SMP   */
#define EXPRESS_CSMI                0x450B      /* CSMI          */
#define EXPRESS_HWIOCTL_HBA         0x450C      /* HWIOCTL_HBA   */
#define EXPRESS_OBSOLETE_450D       0x450D
#define EXPRESS_OBSOLETE_450E       0x450E
#define EXPRESS_IOCTL_GET_MOD_INFO  0x450F      /* MODULE_INFO   */
#define EXPRESS_HWIOCTL_VSA         0x4510      /* HWIOCTL_VSA   */

#define EXPRESS_IOCTL_MAX           0x4510

/* The following are target mode ioctls.  These should *not* be used */
/* with the ioctl () call, rather they should be referenced through  */
/* the target_connect function - see the scsitmd sample driver.      */

#define FCTM_IOCTL_OBSOLETE         0x4520

#ifdef FCTM_DEPRECATED
#define FCTM_IOCTL_INITIALIZE       0x4521  /* HWIOCTL_FCTM_INIT_IN/OUT     */
#define FCTM_IOCTL_SHUTDOWN         0x4522  /* none                         */
#define FCTM_IOCTL_GET_INFO         0x4523  /* HWIOCTL_FCTM_INFO            */
#define FCTM_IOCTL_GET_INI_LIST     0x4524  /* HWIOCTL_FCTM_INI_LIST        */
#endif

#define FCTM_IOCTL_CONTROL          0x4530
#define FCTM_IOCTL_QUERY            0x4531

#define SASTM_IOCTL_CONTROL         0x4540
#define SASTM_IOCTL_QUERY           0x4541


/* 0x4590 - 0x459F reserved for Direct2GPU */

#endif /* ATTO_IOCTL_H */
