/****************************************************************************

DESCRIPTION: Logging functions

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/

#ifndef ATTO_LOG_H
#define ATTO_LOG_H


#include <hwpublic.h>


/**
 * atto_log_lvl() - Log an event, filtering by ATTO level.
 *                  Message is prefixed with driver name.
 */

#define atto_log_lvl( lnx_lvl, atto_lvl, func_area, fmt, arg... )       \
    do {                                                                \
        if( (func_area & atto_log_mask) && (atto_lvl <= log_level) )    \
            printk( lnx_lvl ATTO_DRVR_NAME ": " fmt, ##arg );           \
    } while(0)


/* Wrappers for atto_log_lvl() for each logging level */

#define atto_log_alert( func_area, fmt, arg... ) \
        atto_log_lvl( KERN_ALERT, HWLOG_LVL_ALERT, func_area, fmt, ##arg )

#define atto_log_crit( func_area, fmt, arg... ) \
        atto_log_lvl( KERN_CRIT, HWLOG_LVL_CRITICAL, func_area, fmt, ##arg )

#define atto_log_err( func_area, fmt, arg... ) \
        atto_log_lvl( KERN_ERR, HWLOG_LVL_ERROR, func_area, fmt, ##arg )

#define atto_log_warn( func_area, fmt, arg... ) \
        atto_log_lvl( KERN_WARNING, HWLOG_LVL_WARNING, func_area, fmt, ##arg )

#define atto_log_note( func_area, fmt, arg... ) \
        atto_log_lvl( KERN_NOTICE, HWLOG_LVL_NOTICE, func_area, fmt, ##arg )

#define atto_log_info( func_area, fmt, arg... ) \
        atto_log_lvl( KERN_INFO, HWLOG_LVL_INFO, func_area, fmt, ##arg )

#define atto_log_dbg( func_area, fmt, arg... ) \
        atto_log_lvl( KERN_DEBUG, HWLOG_LVL_VERBOSE1, func_area, fmt, ##arg )


#endif /* ATTO_LOG_H */
