/****************************************************************************

DESCRIPTION: Driver Parameter Management

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2020
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef HWDPM_H
#define HWDPM_H

/* NOTE: This is a special header.  It must be included in the Linux driver
layer.  Therefore, it must stand on its own.  The purpose is to share help
text between module parameters and the DPM IOCTL interface.
*/

/* HW_DPM_ID_TARGETS */
#define HW_DPM_HELP_TARGETS \
"The number of devices that can be discovered in the topology.  Devices that\n" \
"support both initiator and target mode count as one device.\n" \
"Default: 512, Minimum: 18, Maximum: 2048" 

/* HW_DPM_ID_INI_CMDS */
#define HW_DPM_HELP_INI_CMDS \
"The number of initiator mode commands that can be processed at one time.\n" \
"Default: 512, Minimum: 2, Maximum: 4096"

#ifdef HW_VSA_REBUILD
/* HW_DPM_ID_RAID_RBLD_MODE */
#define HW_DPM_HELP_RAID_RBLD_MODE \
"Set PowerCenter Pro RAID group rebuild mode.  Options are:\n" \
"0 - Driver performs rebuilds\n" \
"1 - Applications perform rebuilds (e.g. atpcpro CLI tool)\n" \
"Default: 0"
#endif

#if defined(HW_SCSI_TGT)

/* HW_DPM_ID_TGT_MODE */
#define HW_DPM_HELP_TGT_MODE \
"Enable target mode operation.  Options are:\n" \
"0 - Disable target mode\n" \
"1 - Enable SCSI target mode\n" \
"Default: 0"

/* HW_DPM_ID_TGT_LINK_DELAY */
#define HW_DPM_HELP_TGT_LINK_DELAY \
"Enable target mode link delay.  When operating in target mode only, the port\n" \
"is not enabled until the complete target mode solution is initialized.  This\n" \
"ensures that all incoming I/O is serviced successfully after link negotiation.\n" \
"Default: Enabled"

/* HW_DPM_ID_TGT_CMDS */
#define HW_DPM_HELP_TGT_CMDS \
"The number of target mode command resources.  A resource is used to process\n" \
"an incoming command and transfer a data sequence.  At least one data sequence\n" \
"should be allocated for each command.  The number of data sequences required\n" \
"depends on the command transfer size and the target mode solution's transfer\n" \
"size per sequence.\n" \
"Default: 1024, Minimum: 128, Maximum: 2048 if target mode is enabled"

/* HW_DPM_ID_TGT_PHY_MASK */
#define HW_DPM_HELP_TGT_PHY_MASK \
"Mask of PHYs for which the target mode enable setting applies.  Bit 0\n" \
"corresponds to PHY 0.\n" \
"Default: All PHYs enabled if target mode is enabled, otherwise no PHYs enabled"

/* HW_DPM_ID_TGT_INI_MASK */
#define HW_DPM_HELP_INI_PHY_MASK \
"Mask of PHYs for which initiator mode is enabled when target mode is enabled\n" \
"and multi-mode ports are enabled.  Bit 0 corresponds to PHY 0.\n" \
"Default: PHYs not set in the target mode PHY mask"

/* HW_DPM_ID_MULTI_MODE_PORTS */
#define HW_DPM_HELP_MULTI_MODE_PORTS \
"Enable simultaneous initiator and target mode operation.  By default,\n" \
"initiator mode is enabled.  If target mode is enabled, initiator mode is\n" \
"automatically disabled unless this setting is enabled.\n" \
"Default: Disabled"

#endif

#ifdef HW_TRACING

/* HW_DPM_ID_TRACE_BUF_SIZE */
#define HW_DPM_HELP_TRACE_BUF_SIZE \
"The size of the trace buffer for tracing drivers.\n" \
"Default: 1MB, Minimum: 4KB, Maximum: 32MB"

/* HW_DPM_ID_TRACE_BUF_WRAP */
#define HW_DPM_HELP_TRACE_BUF_WRAP \
"When disabled, tracing stops when the trace buffer is full.  When enabled,\n" \
"tracing continues maintaining a full buffer of the most recent events.\n" \
"Default: Enabled"

#endif

/* HW_DPM_ID_SGL_SEG_DATA_SGE */
#define HW_DPM_HELP_SGL_SEG_DATA_SGE \
"The number of entries in a S/G list segment for data transfer.  Multiply by\n" \
"the OS page size for an approximate data transfer size per segment.  Too\n" \
"many entries can waste memory for small data transfers.  Too few entries\n" \
"can lead to minor performance loss for large data transfers.\n" \
"Note the value may be adjusted for controller requirements.\n" \
"Default: 63, Minimum: 7, Maximum: 255"

/* HW_DPM_ID_SGL_SEGMENTS */
#define HW_DPM_HELP_SGL_SEGMENTS \
"The number of S/G list segments.  Segments are daisy chained to meet the\n" \
"data transfer size requirement for an I/O.  Exhausting this resource\n" \
"due to high queue depths may cause I/O latency.\n" \
"Default: 1024, Minimum: 4, Maximum: 3072"

/* HW_DPM_ID_PORT_ADDR */
#define HW_DPM_HELP_PORT_ADDR \
"Override the SAS address assigned to the adapter in NVRAM.\n" \
"Default: NVRAM value"

/* HW_DPM_ID_DIS_SPINDOWN */
#define HW_DPM_HELP_DIS_SPINDOWN \
"Disable drive spindown when the system powers down.\n" \
"Default: False"

/* HW_DPM_ID_MFG_TEST */
#define HW_DPM_HELP_MFG_TEST \
"Enable manufacturing test mode.\n" \
"Default: Disabled"

/* HW_DPM_ID_NUM_SAS_ADDR */
#define HW_DPM_HELP_NUM_SAS_ADDR \
"The number of unique SAS addresses assigned to the PHYs.  Valid values are\n" \
"1 to the number of PHYs in powers of 2 (1, 2, 4, 8, 16...).  This setting can force\n" \
"creation of narrow logical SAS ports.\n" \
"Default: 1 for 8 port adapters, 2 for 16 port adapters"

/* HW_DPM_ID_FW_LOG_SIZE */
#define HW_DPM_HELP_FW_LOG_SIZE \
"The size, in bytes, for firmware event logs.\n" \
"Default: Disabled, Minimum: 512KB, Maximum: 4MB"

/* HW_DPM_ID_FW_DBG_MASK_IOP */
#define HW_DPM_HELP_FW_DBG_MASK_IOP \
"Specifies the firmware logging mask for PCIe interface events."

/* HW_DPM_ID_FW_DBG_MASK_PL */
#define HW_DPM_HELP_FW_DBG_MASK_PL \
"Specifies the firmware logging mask for SAS interface events."

/* HW_DPM_ID_LOG_LEVEL */
#define HW_DPM_HELP_LOG_LEVEL \
"The driver logging level - starts at zero and increments to increase logging.\n" \
"The selected level logs messages for that level and lower.  Levels are:\n" \
"0  Disabled    All logging disabled\n" \
"1  Always      Informational, always logged when logging enabled\n" \
"2  Alert       Critical error requiring immediate attention\n" \
"3  Critical    Critical errors/info that may accompany an alert\n" \
"4  Error       Error that may require attention\n" \
"5  Warning     All other errors\n" \
"6  Notice      Informational message that may require attention\n" \
"7  Info        All other informational messages\n" \
"8+ Verbose     Excessive logging for issue analysis, may reduce performance\n" \
"Default: 7 for release drivers, 10 for tracing drivers"

/* HW_DPM_ID_BOOT_DRVR */
#define HW_DPM_HELP_BOOT_DRVR \
"Enable the boot driver.\n" \
"Default: Enabled"

/* HW_DPM_ID_NCQ */
#define HW_DPM_HELP_NCQ \
"Enable SATA native command queuing.  This allows multiple commands to be\n" \
"sent to a SATA device at one time which can improve performance.\n" \
"Default: Enabled"

/* HW_DPM_ID_PC_PRO */
#define HW_DPM_HELP_PC_PRO \
"Enable PowerCenter Pro which allows RAID groups to be created.  PowerCenter\n" \
"Pro may need to be disabled if devices are not visible in the system\n" \
"(devices with inaccessible media are hidden by default to protect data).\n" \
"Default: Enabled"

/* HW_DPM_ID_SSC */
#define HW_DPM_HELP_SSC \
"Enable Spread Spectrum Clocking (SSC).  Disable it if the attached device does\n" \
"not support SSC causing connection issues.\n" \
"Default: Enabled"

/* HW_DPM_ID_SGPIO_LOCATE */
#define HW_DPM_HELP_SGPIO_LOCATE \
"Set the locate mode for SGPIO transmission.  Modes are:\n" \
"0 - Blink\n" \
"1 - On (steady)\n" \
"2 - Off (disabled)\n" \
"If the device's locate LED blinks erratically, it may have its own blink\n" \
"mechanism that requires steady transmission from the adapter.\n" \
"Default: 0"

/* HW_DPM_ID_LED_CTRL */
#define HW_DPM_HELP_LED_CTRL \
"The adapter creates a virtual SES device for its PHYs.  This setting\n" \
"determines how incoming SES commands control device LEDs.  Modes are:\n" \
"0 - Disabled\n" \
"2 - SGPIO\n" \
"Default: 2"

/* HW_DPM_ID_T10_PI */
#define HW_DPM_HELP_T10_PI \
"Enable T10 protection information (T10-DIF).\n" \
"Default: Enabled"

/* HW_DPM_ID_INT_COAL */
#define HW_DPM_HELP_INT_COAL \
"The interrupt coalescing level.  Interrupt coalescing delays I/O response\n" \
"processing until an I/O count or time threshold is met in order to process\n" \
"multiple responses in a single interrupt.  This can improve performance\n" \
"depending on queue depth, but can also introduce I/O latency.  Values are:\n" \
"0 - Disabled\n" \
"1 - Low I/O count and time threshold\n" \
"2 - Medium I/O count and time threshold\n" \
"3 - High I/O count and time threshold\n" \
"Default: 0"

/* HW_DPM_ID_LINK_RATE_FIXED */
#define HW_DPM_HELP_LINK_RATE_FIXED \
"If enabled, the configured connection speed is fixed.  If the attached device\n" \
"doesn't support the fixed speed, a connection will not be established.\n" \
"If disabled, the configured speed and lower are allowed.\n" \
"Default: Enabled"

/* HW_DPM_ID_LINK_RATE */
#define HW_DPM_HELP_LINK_RATE \
"The connection speed.  Values are:\n" \
"0 - Auto speed negotiation\n" \
"3 - 3 Gb/s\n" \
"6 - 6 Gb/s\n" \
"12 - 12 Gb/s\n" \
"Default: 0"

/* HW_DPM_ID_EXEC_THR */
#define HW_DPM_HELP_EXEC_THR \
"The maximum number of I/O commands that are sent to a target at one time.\n" \
"0 selects the driver default.  255 disables any I/O throttling.\n" \
"Default: 0, Minimum: 0, Maximum: 255"

/* HW_DPM_ID_LINK_DOWN_TMO */
#define HW_DPM_HELP_LINK_DOWN_TMO \
"The amount of time in seconds to wait for a link to be re-established to a\n" \
"target after it disappears.  I/O is held in the driver during this time.\n" \
"If the time expires before the device is rediscovered, pending I/O is failed\n" \
"and the target is removed from the system.  0 selects the driver default.\n" \
"Default: 0, Minimum: 0, Maximum: 255"

/* HW_DPM_ID_DEV_SORT */
#define HW_DPM_HELP_DEV_SORT \
"Sort devices by SAS address when initial discovery is complete before\n" \
"assigning target IDs and presenting them to the OS.  If disabled, target IDs\n" \
"are assigned as devices are discovered.\n" \
"Default: Enabled"

/* HW_DPM_ID_DEV_WAIT_TIME */
#define HW_DPM_HELP_DEV_WAIT_TIME \
"The number of seconds to wait for devices to be discovered before exiting\n" \
"adapter initialization.  This can ensure all devices are found by the first\n" \
"OS device scan, and that devices are consistently assigned target IDs on\n" \
"each boot.  The time may be adjusted if adaptive discovery is enabled.\n" \
"Default: 0, Minimum: 0, Maximum: 255"

/* HW_DPM_ID_DEV_WAIT_COUNT */
#define HW_DPM_HELP_DEV_WAIT_COUNT \
"The number of devices that need to be discovered before exiting adapter\n" \
"initialization.  This can ensure all devices are found by the first OS device\n" \
"scan, and that devices are consistently assigned target IDs on each boot.\n" \
"Default: 0, Minimum: 0, Maximum: 255"

/* HW_DPM_ID_DEV_WAIT_ADAPT */
#define HW_DPM_HELP_DEV_WAIT_ADAPT \
"Adapt the time to wait for devices to be discovered.  The default time is\n" \
"determined by the device wait time setting or, if 0, 60 seconds.  Waiting\n" \
"will end before this time expires if any configured device count is met, or,\n" \
"no devices have been discovered for a few seconds.  Waiting will be extended\n" \
"if any configured device count is NOT met, or discovery is still active.\n" \
"If disabled, the configured device wait time is strictly honored.\n" \
"Default: Enabled"

/* HW_DPM_ID_DEV_WAIT_BOOT */
#define HW_DPM_HELP_DEV_WAIT_BOOT \
"During initial discovery, only wait for boot devices (configured by\n" \
"PowerCenter Pro).\n" \
"Default: Disabled"

/* HW_DPM_ID_DEV_MAINT */
#define HW_DPM_HELP_DEV_MAINT \
"Enable device maintenance mode.  Attached devices are not presented to the\n" \
"OS, but ATTO CLI tools can perform commands while the device is offline.\n" \
"Default: Disabled"

/* HW_DPM_ID_DEV_SPINUP_DELAY */
#define HW_DPM_HELP_DEV_SPINUP_DELAY \
"The time to wait between initiating spin-up to direct-attached devices.  This\n" \
"feature is useful for drive enclosures that do not supply enough power to\n" \
"spin up all HDDs simultaneously.  Peak power is typically consumed in the\n" \
"first 4 seconds of spin-up time.\n" \
"Default: 0, Minimum: 0, Maximum: 20"

#endif  /* HWDPM_H */
