/****************************************************************************

DESCRIPTION: Interface data.

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2006-2021
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef HWIDATA_H
#define HWIDATA_H


/* General IOCTL status codes.  See the HWIOCTL_HBA structure definition    */
/* for a description of error codes that apply to all IOCTLs.  See the      */
/* function definitions below for details on the remaining error codes.     */

#define HWI_STS_SUCCESS         0x00    /* Success                          */
#define HWI_STS_FAILED          0x01    /* Failure                          */
#define HWI_STS_INV_VERSION     0x02    /* Invalid version                  */
#define HWI_STS_OUT_OF_RSRC     0x03    /* Out of resources                 */
#define HWI_STS_INV_FUNC        0x04    /* Invalid function                 */
#define HWI_STS_UNSUPPORTED     0x05    /* IOCTL unsupported                */
#define HWI_STS_INV_ADAPTER     0x06    /* IOCTL for another adapter        */
#define HWI_STS_INV_DRVR_VER    0x07    /* New driver required              */
#define HWI_STS_INV_PARAM       0x08    /* Generic invalid parameter        */
#define HWI_STS_TIMEOUT         0x09    /* Request timeout                  */
#define HWI_STS_NOT_APPL        0x0A    /* Not applicable                   */
#define HWI_STS_DEGRADED        0x0B    /* Adapter is degraded              */
#define HWI_STS_DISABLED        0x0C    /* Feature is disabled              */
#define HWI_STS_INV_OEM         0x0D    /* Invalid OEM                      */
#define HWI_STS_INV_SUBFUNC     0x0E    /* Invalid subfunction              */
#define HWI_STS_INV_PORT_ID     0x0F    /* Invalid adapter port ID          */
#define HWI_STS_NOT_PRESENT     0x10    /* Item not present                 */
#define HWI_STS_NO_CHANGE       0x11    /* No change made by operation      */
#define HWI_STS_BUF_TOO_SMALL   0x12    /* Data area too small for response */
#define HWI_STS_UPDATE_PENDING  0x13    /* Previous update / change pending */


/* Link rates (same as multipathing API) */

#define HWI_LINK_RATE_1GB       0x00    /* 1 Gb/s                       */
#define HWI_LINK_RATE_1_5GB     0x01    /* 1.5 Gb/s                     */
#define HWI_LINK_RATE_2GB       0x02    /* 2 Gb/s                       */
#define HWI_LINK_RATE_3GB       0x03    /* 3 Gb/s                       */
#define HWI_LINK_RATE_4GB       0x04    /* 4 Gb/s                       */
#define HWI_LINK_RATE_6GB       0x05    /* 6 Gb/s                       */
#define HWI_LINK_RATE_8GB       0x06    /* 8 Gb/s                       */
#define HWI_LINK_RATE_10GB      0x07    /* 10 Gb/s                      */
#define HWI_LINK_RATE_UNKNOWN   0x08    /* Unknown                      */
#define HWI_LINK_RATE_NOT_APPL  0x09    /* N/A                          */
#define HWI_LINK_RATE_100MB     0x0A    /* 100 Mb/s                     */
#define HWI_LINK_RATE_40GB      0x0B    /* 40 Gb/s                      */
#define HWI_LINK_RATE_12GB      0x0C    /* 12 Gb/s                      */
#define HWI_LINK_RATE_16GB      0x0D    /* 16 Gb/s                      */
#define HWI_LINK_RATE_24GB      0x0E    /* 24 Gb/s                      */
#define HWI_LINK_RATE_32GB      0x0F    /* 32 Gb/s                      */
#define HWI_LINK_RATE_50GB      0x10    /* 50 Gb/s                      */
#define HWI_LINK_RATE_100GB     0x11    /* 100 Gb/s                     */
#define HWI_LINK_RATE_48GB      0x12    /* 48 Gb/s                      */
#define HWI_LINK_RATE_64GB      0x13    /* 64 Gb/s                      */
#define HWI_LINK_RATE_96GB      0x14    /* 96 Gb/s                      */
#define HWI_LINK_RATE_128GB     0x15    /* 128 Gb/s                     */
#define HWI_LINK_RATE_192GB     0x16    /* 192 Gb/s                     */
#define HWI_LINK_RATE_256GB     0x17    /* 256 Gb/s                     */
#define HWI_LINK_RATE_25GB      0x18    /* 25 Gb/s                      */
#define HWI_LINK_RATE_200GB     0x19    /* 200 Gb/s                     */


/* Miscellaneous definitions */

#define HWI_PCI_PORT_ID_INV     0xFF    /* Invalid PCIe Port ID             */
#define HWI_SAS_PHY_ID_INV      0xFF    /* Invalid SAS PHY ID               */
#define HWI_SAS_PORT_ID_INV     0xFF    /* Invalid SAS Port ID              */
#define HWI_SLOT_NUM_INV        0xFF    /* Invalid slot number              */
#define HWI_TARGET_ID_INV       0xFFFFFFFF  /* Invalid target ID            */
#define HWI_LUN_INV             0xFFFFFFFF  /* Invalid LUN                  */


#pragma pack (1)


/****************************************************************************
                                 Device info
*****************************************************************************/

typedef struct _HwiHbaSasDeviceInfo {

    #define HWI_SDI_MAX_PHYS_WIDE_PORT  16

    BYTE    PhyId[HWI_SDI_MAX_PHYS_WIDE_PORT];
                                /*  -00 PHY IDs of parent expander/adapter  */
        #define HWI_SDI_PHY_ID_INV      HWI_SAS_PHY_ID_INV

    DWORD   ExpTargetId;        /*  -10 Target ID of parent expander        */
    DWORD   SasPortMask;        /*  -14 SAS port mask device is present on  */
    BYTE    SasLevel;           /*  -18 SAS depth in topology               */
        #define HWI_SDI_SAS_LVL_INV     0xFF

    BYTE    SlotNum;            /*  -19 Slot number                         */
        #define HWI_SDI_SLOT_NUM_INV    HWI_SLOT_NUM_INV
    
    BYTE    DevType;            /*  -1A Device type                         */
        #define HWI_SDI_DT_END_DEVICE   0       /* End device               */
        #define HWI_SDI_DT_EXPANDER     1       /* Expander                 */
        #define HWI_SDI_DT_PORT_MULT    2       /* Port multiplier          */

    BYTE    IniFlags;           /*  -1B SMP Discover initiator flags        */
    BYTE    TgtFlags;           /*  -1C SMP Discover target flags           */
    BYTE    LinkRate;           /*  -1D SAS link rate (SMP_RATE_XXX)        */
    BYTE    LocFlags;           /*  -1E Location flags                      */

        /* If this field is zero, the location is not known.  The PhyId,    */
        /* ExpTargetId, SasLevel, SlotNum, and PmPort fields are not valid. */

        #define HWI_SDI_LF_DIRECT       0x01    /* Direct attached device   */
        #define HWI_SDI_LF_EXPANDER     0x02    /* Expander attached device */
                                                /* Expander fields valid    */
        #define HWI_SDI_LF_PORT_MULT    0x04    /* Port mult. attached dev  */
                                                /* Port mult. fields valid  */
        #define HWI_SDI_LF_IO_PATH      0x08    /* Used for I/O             */
                                                /* Added in version 1       */
    BYTE    PmPort;             /*  -1F Port multiplier port                */
} HWIHBA_SAS_DEV_INFO,          /*   20 Size of structure                   */
*PHWIHBA_SAS_DEV_INFO;


typedef struct _HwiHbaFcDeviceInfo {
#ifdef HWFC_H
    FC_ELS_SVC_PARAMS SvcParams;/*  -00 ELS service parameters              */
#else
    BYTE    SvcParams[0x70];    /*  -00 ELS service parameters              */
#endif
    DWORD   FcPortId;           /*  -70 FC port ID                          */
    BYTE    IniFlags;           /*  -74 Initiator mode flags                */
        #define HWI_FCDI_IF_SCSI        0x01    /* SCSI initiator           */
        #define HWI_FCDI_IF_NVME        0x02    /* NVMe initiator           */

    BYTE    TgtFlags;           /*  -75 Target mode flags                   */
        #define HWI_FCDI_TF_SCSI        0x01    /* SCSI target              */
        #define HWI_FCDI_TF_NVME        0x02    /* NVMe target              */
    
    BYTE    Flags;              /*   76 Flags                               */
        #define HWI_FCDIF_NVME_DISC     0x01    /* NVMe discovery device    */
        #define HWI_FCDIF_NVME_SLER     0x02    /* NVMe SLER enabled        */
        #define HWI_FCDIF_SCSI_SLER     0x04    /* SCSI SLER enabled        */
        #define HWI_FCDIF_NVME_CONF     0x08    /* NVMe confirmation enabled*/
        #define HWI_FCDIF_SCSI_CONF     0x10    /* SCSI confirmation enabled*/

    BYTE    Reserved[7];        /*   77 Reserved                            */
    WORD    MaxQueueDepth;      /*   7E Maximum queue depth                 */
} HWIHBA_FC_DEV_INFO,           /*   80 Size of structure                   */
*PHWIHBA_FC_DEV_INFO;


typedef struct _HwiHbaPciDeviceInfo {
    BYTE    DevType;            /*  -00 Device type                         */
        #define HWI_PDI_DT_NVME         0       /* NVMe device              */
        #define HWI_PDI_DT_SCSI         1       /* SCSI device              */

    BYTE    PciBus;             /*  -01 PCI bus number                      */
    BYTE    PciDevice;          /*  -02 PCI device number                   */
    BYTE    PciFunction;        /*  -03 PCI function number                 */
    BYTE    SlotNum;            /*  -04 Slot number                         */
        #define HWI_PDI_SLOT_NUM_INV    HWI_SLOT_NUM_INV

    BYTE    NvmeFlags;          /*  -05 NVMe feature flags                  */
        #define HWI_PDI_NF_SMBUS_BASIC  0x01    /* Basic Mgmt over SMBus    */
        #define HWI_PDI_NF_SMBUS_NVMEMI 0x02    /* NVMe-MI over SMBus       */

    BYTE    Reserved[2];        /*   06 Reserved                            */
} HWIHBA_PCI_DEV_INFO,          /*   08 Size of structure                   */
*PHWIHBA_PCI_DEV_INFO;


/****************************************************************************
                               Expander info
*****************************************************************************/

typedef struct _HwiHbaExpanderInfo {
    CHAR    VendorId[8];        /* 00 Vendor ID                             */
    CHAR    ProductId[16];      /* 08 Product ID                            */
    CHAR    Revision[4];        /* 18 Revision level                        */
    CHAR    ComponentVendor[8]; /* 1C Component Vendor ID                   */
    WORD    ComponentId;        /* 24 Component ID                          */
    BYTE    ComponentRev;       /* 26 Component Revision level              */
    BYTE    NumSlots;           /* 27 Number of end device slots            */
    BYTE    NumPhys;            /* 28 Number of PHYs                        */
    BYTE    Reserved;           /* 29 Alignment                             */
} HWIHBA_EXP_INFO,              /* 2A Size of structure                     */
*PHWIHBA_EXP_INFO;


/****************************************************************************
                                PHY stats
*****************************************************************************/

typedef struct _HwiHbaNetQueueStats {
    QWORD   Size;               /*  00 Size of the ring (in entries)        */
    QWORD   Resizes;            /*  08 Number of ring resizes               */
    QWORD   Total;              /*  10 Total number of packets sent/recvd   */
    QWORD   OctGood;            /*  18 Count of good octets sent/recvd      */
    /* This is where the version 4-6 structure ended */
    QWORD   LroPackets;         /*  20 LRO offloads performed               */
    QWORD   LroBytes;           /*  28 Received LRO'd bytes                 */
    QWORD   LroMaxPayload;      /*  30 LRO aborts by max IP payload         */
    QWORD   LroTimeout;         /*  38 LRO aborts by LRO timeout            */
    QWORD   LroException;       /*  40 LRO aborts by LRO exception          */
    QWORD   LroOther;           /*  48 LRO aborts by other                  */
    QWORD   SwLroQueued;        /*  50 SW LRO segments coalesced            */
    QWORD   SwLroFlushed;       /*  58 SW LROs performed                    */
    QWORD   SwLroSkipped;       /*  60 SW LROs skipped                      */
    QWORD   RxCsumNone;         /*  68 RX packets where no checksum offload */
                                /*  occurred                                */
    QWORD   RxWqeError;         /*  70 RX WQE errors                        */
    QWORD   RxBufsAllocated;    /*  78 RX buffers allocated for ring        */
    QWORD   RxAllocFails;       /*  80 Failed RX buffer allocations         */
    QWORD   RxIfcDownDrop;      /*  88 RX drops due to ifc down             */
    QWORD   TsoPackets;         /*  90 TSOs performed                       */
    QWORD   TsoBytes;           /*  98 TSO bytes sent                       */
    QWORD   TsoFails;           /*  A0 TSO failures                         */
    QWORD   TxCsumNone;         /*  A8 TX packets where checksum offloading */
                                /*  will not be performed                   */
    QWORD   TxInlined;          /*  B0 TX inlined data                      */
    QWORD   TxMapFails;         /*  B8 TX DMA mapping failures              */
    QWORD   TxRingFull;         /*  C0 TX ring fulls                        */
    QWORD   TxOtherFails;       /*  C8 TX other failures                    */
    QWORD   TxNop;              /*  D0 NOPs transmitted                     */
    QWORD   TxBufQueueHi;       /*  D8 Hi watermark for TX buf queue size   */
    QWORD   TxIfcDownDrop;      /*  E0 TX drops due to ifc down             */
    QWORD   TxInvalidEthHdrDrop;/*  E8 TX drops due to an invalid           */
    QWORD   XmitMore;           /*  F0 Xmit mores detected                  */
                                /*  ethernet header                         */
} HWIHBA_NET_QUEUE_STATS,       /*  F8 Size of structure                    */
*PHWIHBA_NET_QUEUE_STATS;


/* Define the statistics separately so apps and drivers can cache them */
#define HWI_NPS_NUM_TC      8   /* Number of traffic classes           */
#define HWI_NPS_NUM_PRIO    8   /* Number of priorities                */
#define HWI_NPS_NUM_LANES   4   /* Number of phy lanes                 */

typedef struct _HwiHbaNetPhyStats {
    QWORD   RxCrcErrs;          /*  00 Count of packets with CRC Errors     */
    QWORD   RxIllegalErrs;      /*  08  byte errors (eg, illegal symbol)    */
    QWORD   RxErrBytes;         /*  10  error bytes                         */
    QWORD   RxLenErr;           /*  18  Length field doesn't match pkt len  */
    QWORD   RxMissed[HWI_NPS_NUM_TC];
                                /*  20 Dropped- no buffer avail. (per TC)   */
    QWORD   MacLocalFault;      /*  60 Count of local MAC faults detected   */
    QWORD   MacRemoteFault;     /*  68 Count of remote MAC faults detected  */
    QWORD   Rx64;               /*  70 Rx packets with Length = 64 bytes    */
    QWORD   Rx127;              /*  78    65 <= Length <= 127               */
    QWORD   Rx255;              /*  80   128 <= Length <= 255               */
    QWORD   Rx511;              /*  88   256 <= Length <= 511               */
    QWORD   Rx1023;             /*  90   512 <= Length <= 1023              */
    QWORD   Rx1522;             /*  98  1024 <= Length <= Max Packet Size   */
    QWORD   RxGood;             /*  A0 Count of valid packets received      */
    QWORD   RxBcast;            /*  A8 Broadcast packets received           */
    QWORD   RxMcast;            /*  B0 Multicast packets received           */
    QWORD   RxOctGood;          /*  B8 Count of good octets received        */
    QWORD   TxOctGood;          /*  C0 Count of good octets transmitted     */
    QWORD   RxUndersized;       /*  C8 Length too small (otherwise good)    */
    QWORD   RxFragments;        /*  D0 Length too small and bad CRC         */
    QWORD   RxOversized;        /*  D8 Length too big (otherwise good)      */
    QWORD   RxJabber;           /*  E0 Length too big and bad CRC           */
    QWORD   RxMgmt;             /*  E8 Management packets received          */
    QWORD   RxMgmtDropped;      /*  F0 Management packets dropped           */
    QWORD   RxOctTotal;         /*  F8 Total octets received- good & erred  */
    QWORD   RxTotal;            /* 100 Total packets received- good & erred */
    QWORD   TxTotal;            /* 108 Total number of packets sent         */
    QWORD   TxMcast;            /* 110 Multicast packets sent               */
    QWORD   TxBcast;            /* 118 Broadcast packets sent               */
    QWORD   RxShort;            /* 120 Short packet discard packets rec'd   */
    QWORD   RxCsumErrs;         /* 128 Pkts received w/ checksum errors     */
    QWORD   RxAllCount;         /* 130 Good pkts, incl packet-filtered pkts */
    QWORD   RxAllBytes;         /* 138 Count of ALL bytes received          */
    QWORD   FcCrcErrs;          /* 140 FC CRC Errors                        */
    QWORD   FcRxDropped;        /* 148 FC pkts missed due to no descriptors */
    QWORD   FcLast;             /* 150 FC pkts received but no user buffers */
    QWORD   FcRx;               /* 158 FC packets received                  */
    QWORD   FcRxDwords;         /* 160 FC DWORDs received                   */
    QWORD   FcTx;               /* 168 FC packets transmitted               */
    QWORD   FcTxDwords;         /* 170 FC DWORDs transmitted                */
    /* This is where the version 0 structure ended */
    QWORD   TxQFull;            /* 178 Transmit queue full                  */
    QWORD   TxBusy;             /* 180 Transmit ring full                   */
    QWORD   TsoFails;           /* 188 TSO failures                         */
    QWORD   TxCopyFails;        /* 190 Failure during TX copy               */
    QWORD   TxIpChksumFails;    /* 198 IP checksum requested on non-IP TX   */
    QWORD   TxIpVerFails;       /* 1A0 IP version failure on TX             */
    QWORD   TxMapFails;         /* 1A8 Map failure on transmit              */
    QWORD   TxTooBig;           /* 1B0 Transmit too big                     */
    QWORD   TxLenMismatch;      /* 1B8 Transmit length mismatch             */
    QWORD   RxAllocFails;       /* 1C0 Receive buffer allocation failures   */
    QWORD   RxCopyFails;        /* 1C8 Receive buffer copy failures         */
    QWORD   TsoCount;           /* 1D0 TX Segmentation Offloads performed   */
    QWORD   RscCount;           /* 1D8 RSC offloads performed               */
    QWORD   TxOtherFails;       /* 1E0 Other TX failures                    */
    QWORD   RxInput;            /* 1E8 RX input packets                     */
    QWORD   XonTxCount;         /* 1F0 XON packets transmitted              */
    QWORD   XoffTxCount;        /* 1F8 XOFF packets transmitted             */
    QWORD   XonRxCount;         /* 200 XON packets received                 */
    QWORD   XoffRxCount;        /* 208 XOFF packets received                */
    QWORD   ResetTimeMs;        /* 210 Milliseconds since stats reset       */
    /* This is where the version 1 structure ended */
    QWORD   LroFlushed;         /* 218 LRO offloads performed               */
    QWORD   LroQueued;          /* 220 LRO segments coalesced               */
    QWORD   LroSkipped;         /* 228 LRO offloads skipped (no resources)  */
    QWORD   TxInline;           /* 230 Transmits inlined                    */
    QWORD   TxOversized;        /* 238 Oversized transmits                  */
    /* This is where the version 2 structure ended */
    QWORD   SbufsUsed;          /* 240 SBUFs used                           */
    QWORD   SbufsSkipped;       /* 248 SBUFs skipped (unavailable)          */
    QWORD   SbufsCopies;        /* 250 SBUF copy operations performed       */
    QWORD   SbufsFilled;        /* 258 SBUFs under capacity                 */
    QWORD   TxTimeAvg;          /* 260 Average time to process a transmit   */
    QWORD   TxTimeAvgX10;       /* 268 Transmits taking more than 10x avg   */
    QWORD   TxTimeAvgX100;      /* 270 Transmits taking more than 100x avg  */
    /* This is where the version 4 structure ended */
    QWORD   IngressGeneral;     /* 278 General discards                     */
    QWORD   IngressPolicyEngine;/* 280 Policy engine discards               */
    QWORD   IngressVlanMembership;
                                /* 288 VLAN membership discards             */
    QWORD   IngressTagFrame;    /* 290 VLAN tag allowance discards          */
    QWORD   EgressVlanMembership;
                                /* 298 VLAN membership discards             */
    QWORD   LoopbackFilter;     /* 2A0 Loop-back filter discards            */
    QWORD   EgressGeneral;      /* 2A8 General discards                     */
    QWORD   EgressHoq;          /* 2B0 Head of queue discards               */
    QWORD   PortIsolation;      /* 2B8 Port isolation discards              */
    QWORD   EgressPolicyEngine; /* 2C0 Policy engine discards               */
    QWORD   IngressTxLinkDown;  /* 2C8 TX Link down discards                */
    QWORD   EgressStp;          /* 2D0 Spanning tree filter discards        */
    QWORD   EgressHoqStall;     /* 2D8 Head of queue stall discards         */
    QWORD   AlignmentErrors;    /* 2E0 Ethernet alignment errors            */
    QWORD   FcsErrors;          /* 2E8 FCS errors                           */
    QWORD   SingleCollisionFrames;
                                /* 2F0 Single collision frames              */
    QWORD   MultipleCollisionFrames;
                                /* 2F8 Multiple collision frames            */
    QWORD   SqeTestErrors;      /* 300 SQE test error received              */
    QWORD   DeferredTransmissions;
                                /* 308 Deferred transmissions, medium busy  */
    QWORD   LateCollisions;     /* 310 Late collision detection             */
    QWORD   ExcessiveCollisions;/* 318 Excessive collision frames           */
    QWORD   InternalMacTransmitErrors;
                                /* 320 Failed transmit, no errors detected  */
    QWORD   CarrierSenseErrors; /* 328 Carrier sense condition lost         */
    QWORD   FramesTooLong;      /* 330 Exceed maximum frame size            */
    QWORD   InternalMacReceiveErrors;
                                /* 338 Failed receive, no errors detected   */
    QWORD   InUnknownOpcodes;   /* 340 Unsupported opcode                   */
    QWORD   RxFrames[HWI_NPS_NUM_PRIO];
                                /* 348 RX frame counter per priority        */
    QWORD   RxOctets[HWI_NPS_NUM_PRIO];
                                /* 388 RX octet counter per priority        */
    QWORD   TxFrames[HWI_NPS_NUM_PRIO];
                                /* 3C8 TX frame counter per priority        */
    QWORD   TxOctets[HWI_NPS_NUM_PRIO];
                                /* 408 TX octet counter per priority        */
    QWORD   RxPause[HWI_NPS_NUM_PRIO];
                                /* 448 Pause frames received                */
    QWORD   TxPause[HWI_NPS_NUM_PRIO];
                                /* 488 Pause frames transmitted             */
    QWORD   RxPauseDuration[HWI_NPS_NUM_PRIO];
                                /* 4C8 RX Pause duration (us)               */
    QWORD   TxPauseDuration[HWI_NPS_NUM_PRIO];
                                /* 508 TX Pause duration (us)               */
    QWORD   RxPauseTransition[HWI_NPS_NUM_PRIO];
                                /* 548 Received Xoff to Xon transitions     */
    QWORD   TxPauseTransition[HWI_NPS_NUM_PRIO];
                                /* 588 Transmitted Xoff to Xon transitions  */
    QWORD   RxDiscards[HWI_NPS_NUM_PRIO];
                                /* 5C8 RX discard count                     */
    QWORD   TxQueueDepth[HWI_NPS_NUM_TC];
                                /* 608 TX queue depth per tc                */
    QWORD   PortTransmitWait;   /* 648 Time port has data, not sending (ns) */
    QWORD   PhysicalSymbolErrors;
                                /* 650 Physical layer symbol errors         */
    QWORD   SyncHeaderErrors;   /* 658 Sync header errors                   */
    QWORD   BipErrors[HWI_NPS_NUM_LANES];
                                /* 660 PRBS errors                          */
    QWORD   FcCorrectedBlocks[HWI_NPS_NUM_LANES];
                                /* 680 FEC correctable block per lane       */
    QWORD   RsCorrectedBlocks;  /* 6A0 RS-FEC correctable block             */
    QWORD   RsUncorrectableBlocks;
                                /* 6A8 RS-FEC uncorrectable block           */
    QWORD   RsNoErrorsBlocks;   /* 6B0 RS-FEC blocks with no errors         */
    QWORD   RsSingleErrorBlocks;/* 6B8 RS-FEC single error blocks           */
    QWORD   RsCorrectedSymbols[HWI_NPS_NUM_LANES];
                                /* 6C0 RS-FEC corrected symbols per lane    */
    QWORD   PhyReceivedBits;    /* 6E0 Traffic bits received                */
    QWORD   PhyCorrectedBits;   /* 6E8 Traffic bits corrected               */
    /* This is where the version 6 structure ended */
} HWIHBA_NET_PHY_STATS,         /* 6F0 Size of structure                    */
*PHWIHBA_NET_PHY_STATS;


typedef struct _HwiHbaSasPhyStats {
    DWORD   InvDwords;          /*  00 Invalid DWORD count                  */
    DWORD   DisparityErrs;      /*  04 Disparity error count                */
    DWORD   LossOfDwSync;       /*  08 Loss of DWORD sychronization         */ 
    DWORD   PhyRstErrs;         /*  0C PHY reset error count                */
    DWORD   CrcErrs;            /*  10 CRC error count                      */
    DWORD   BcChanges;          /*  14 Broadcast change count               */
    DWORD   LinkChanges;        /*  18 Link change count (up/down)          */
} HWIHBA_SAS_PHY_STATS,         /*  1C Size of structure                    */
*PHWIHBA_SAS_PHY_STATS;


typedef struct _HwiHbaFcPhyStats {
    QWORD   LinkFailErrs;       /*  00 Link failure error count             */
    QWORD   LossOfSignal;       /*  08 Loss of signal error count           */
    QWORD   LossOfSync;         /*  10 Loss of sychronization error count   */ 
    QWORD   PrimSeqErrs;        /*  18 Primitive sequence error count       */
    QWORD   PrimSeqEvtTmo;      /*  20 Primitive sequence event timeouts    */
    QWORD   CrcErrs;            /*  28 CRC error count                      */
    QWORD   InvXmitWordErrs;    /*  30 Invalid transmission word error count*/
    QWORD   ElasticOverrunErrs; /*  38 Elastic buffer overrun error count   */
    QWORD   ArbTmo;             /*  40 Arbitration timeout count            */
    QWORD   AdvRcvBBCredit;     /*  48 Advertised receive buffer credit     */
    QWORD   CurrRcvBBCredit;    /*  50 Current receive buffer credit        */
    QWORD   AdvXmitBBCredit;    /*  58 Advertised transmit buffer credit    */
    QWORD   CurrXmitBBCredit;   /*  60 Current transmit buffer credit       */
    QWORD   NosEvents;          /*  68 NOS event count                      */
    QWORD   LipEvents;          /*  70 LIP event count                      */
    QWORD   XmitBytes;          /*  78 Transmitted bytes                    */
    QWORD   RcvBytes;           /*  80 Received bytes                       */
    QWORD   XmitFrames;         /*  78 Transmitted frames                   */
    QWORD   RcvFrames;          /*  80 Received frames                      */
    QWORD   SecsSinceLastReset; /*  88 Seconds since stats last reset       */
} HWIHBA_FC_PHY_STATS,          /*  90 Size of structure                    */
*PHWIHBA_FC_PHY_STATS;


typedef struct _HwiHbaPciPhyStats {
    QWORD   BwIntervalUs;       /*  00 Bandwidth counter interval (us)      */
    QWORD   BwTlpPostedIn;      /*  08 Posted TLP byte count (ingress)      */
    QWORD   BwTlpPostedOut;     /*  10 Posted TLP byte count (egress)       */
    QWORD   BwTlpNonPostedIn;   /*  18 Non-posted TLP byte count (ingress)  */
    QWORD   BwTlpNonPostedOut;  /*  20 Non-posted TLP byte count (egress)   */
    QWORD   BwTlpCompletionIn;  /*  28 Completion TLP byte count (ingress)  */
    QWORD   BwTlpCompletionOut; /*  30 Completion TLP byte count (egress)   */
    DWORD   LatencyTlpMaxNs;    /*  38 Max TLP latency between ports (ns)   */
    DWORD   LatencyTlpLastNs;   /*  3C Last TLP latency between ports (ns)  */
    QWORD   ErrUnsuppReq;       /*  40 Unsupported request error count      */
    QWORD   ErrEcrc;            /*  48 ECRC error count                     */
    QWORD   ErrMalformedTlp;    /*  50 Malformed TLP error count            */
    QWORD   ErrRcvOverflow;     /*  58 Receiver overflow error count        */
    QWORD   ErrCompleterAbort;  /*  60 Completer abort error count          */
    QWORD   ErrPoisonedTlp;     /*  68 Poisoned TLP error count             */
    QWORD   ErrSurpriseDown;    /*  70 Surprise down error count            */
    QWORD   ErrDataLink;        /*  78 Data link protocol error count       */
    QWORD   ErrReplayTimerTmo;  /*  80 Replay timer timeout count           */
    QWORD   ErrBadDllp;         /*  88 Bad DLLP error count                 */
    QWORD   ErrBadTlp;          /*  90 Bad TLP error count                  */
    QWORD   ErrNak;             /*  98 Received NAK error count             */
    QWORD   ErrReceiver;        /*  A0 Receiver error count                 */
    QWORD   ErrRcvFatalMsg;     /*  A8 Receiver fatal error messages        */
    QWORD   ErrRcvNonFatalMsg;  /*  B0 Receiver non-fatal error messages    */
    QWORD   ErrRcvCorrectMsg;   /*  B8 Receiver correctable error messages  */
} HWIHBA_PCI_PHY_STATS,         /*  C0 Size of structure                    */
*PHWIHBA_PCI_PHY_STATS;


/****************************************************************************
                                 PHY info
*****************************************************************************/

/* PHY states - 'NOT_RPTD' and the others are mutually exclusive */

#define HWI_GPHI_PHS_NOT_RPTD       0x00    /* Not reported/supported       */
#define HWI_GPHI_PHS_DIS_USER       0x01    /* Disabled by user             */
#define HWI_GPHI_PHS_DIS_CBL_MGT    0x02    /* No cable plugged in          */
#define HWI_GPHI_PHS_DIS_ERR        0x03    /* Disabled due to errors       */
#define HWI_GPHI_PHS_LINK_DOWN      0x04    /* PHY enabled, link is down    */
#define HWI_GPHI_PHS_LINK_UP        0x05    /* PHY is up                    */

/* Discovery states - 'NOT_RPTD' and the others are mutually exclusive */

#define HWI_GPHI_DS_NOT_RPTD        0x00    /* Not reported/supported       */
#define HWI_GPHI_DS_DIS_USER        0x01    /* Disabled by user             */
#define HWI_GPHI_DS_DIS_ERR         0x02    /* Disabled due to errors       */
#define HWI_GPHI_DS_IDLE            0x03    /* Discovery is done            */
#define HWI_GPHI_DS_IN_PROG         0x04    /* Discovery in progress        */
#define HWI_GPHI_DS_NOT_APPL        0x05    /* Not applicable (link is down)*/


typedef struct _HwiHbaSasPhyInfo {
#ifdef HWSAS_H
    SAS_IDENT_FRAME IdentFrame; /*   00 PHY identify frame                  */
    SAS_IDENT_FRAME AttachIdentFrame;
                                /*   1C Attached Identify frame             */
#else
    BYTE    IdentFrame[0x1C];
    BYTE    AttachIdentFrame[0x1C];
#endif
#ifdef HWSATA_H
    SATA_REGISTER_DTH SataRegDth;
#else
    BYTE    SataRegDth[0x14];   /*   38 SATA Register Device to Host        */
#endif
    BYTE    PhyState;           /*   4C PHY state (HWI_GPHI_PHS_XXX)        */
    BYTE    PhyFlags;           /*   4D PHY flags                           */
        #define HWI_SPHI_PF_SATA    0x01    /* SATA device attached         */
        #define HWI_SPHI_PF_EXT     0x02    /* External PHY                 */
        #define HWI_SPHI_READ_ONLY  0x04    /* Read-Only PHY                */

    BYTE    SasPort;            /*   4E SAS port or HWI_SAS_PORT_ID_INV     */
    BYTE    DiscState;          /*   4F Discovery state (HWI_GPHI_DS_XXX)   */
    BYTE    NegLinkRate;        /*   50 Negotiated link rate                */
    BYTE    MinLinkRateHw;      /*   51 Minimum HW link rate                */
    BYTE    MaxLinkRateHw;      /*   52 Maximum HW link rate                */
    BYTE    MinLinkRateProg;    /*   53 Minimum programmed link rate        */
    BYTE    MaxLinkRateProg;    /*   54 Maximum programmed link rate        */
                                /*      Rates above use SMP_RATE_XXX values */
    BYTE    Reserved[3];        /*   55 Reserved                            */
    DWORD   TargetId;           /*   58 Attached target ID                  */
} HWIHBA_SAS_PHY_INFO,          /*   5C Size of structure                   */
*PHWIHBA_SAS_PHY_INFO;


typedef struct _HwiHbaFcPhyInfo {
#ifdef HWFC_H
    FC_ELS_SVC_PARAMS SvcParams;/*   00 ELS service parameters              */
    FC_ELS_SVC_PARAMS AttachSvcParams;
                                /*   70 Attached ELS service parameters     */
#else
    BYTE    SvcParams[0x70];    
    BYTE    AttachSvcParams[0x70];
#endif
    BYTE    PhyState;           /*   E0 PHY state (HWI_GPHI_PHS_XXX)        */
    BYTE    DiscState;          /*   E1 Discovery state (HWI_GPHI_DS_XXX)   */
    BYTE    NegLinkRate;        /*   E2 Negotiated link rate                */
    BYTE    MinLinkRateHw;      /*   E3 Minimum HW link rate                */
    BYTE    MaxLinkRateHw;      /*   E4 Maximum HW link rate                */
    BYTE    MinLinkRateProg;    /*   E5 Minimum programmed link rate        */
    BYTE    MaxLinkRateProg;    /*   E6 Maximum programmed link rate        */
                                /*      Rates use HWI_LINK_RATE_XXX values  */
    BYTE    Topology;           /*   E7 Topology                            */
        #define HWI_FCPHI_TOP_NOT_RPTD  0x00    /* Topology not reported    */
        #define HWI_FCPHI_TOP_NOT_APPL  0x01    /* Not applicable-link down */
        #define HWI_FCPHI_TOP_AL        0x02    /* Arbitrated loop          */
        #define HWI_FCPHI_TOP_AL_PUBLIC 0x03    /* Public loop              */
        #define HWI_FCPHI_TOP_AL_PRIV   0x04    /* Private loop             */
        #define HWI_FCPHI_TOP_PTP       0x05    /* Point-to-point           */
        #define HWI_FCPHI_TOP_PTP_FAB   0x06    /* Point-to-point fabric    */
        #define HWI_FCPHI_TOP_PTP_DIR   0x07    /* Point-to-point direct    */

    DWORD   FcPortId;           /*   E8 FC port ID                          */
    DWORD   TargetId;           /*   EC Attached target ID                  */
    BYTE    FeatEnb;            /*   F0 Features enabled                    */
    BYTE    FeatSupp;           /*   F1 Features supported                  */
        #define HWI_FCPHI_FEAT_INI      0x01    /* Initiator mode           */
        #define HWI_FCPHI_FEAT_TGT      0x02    /* Target mode              */
        #define HWI_FCPHI_FEAT_NPIV     0x04    /* NPIV                     */

    /* Note that IniFlags and TgtFlags were added after the initial version */
    /* of this payload.  If the fields are zero, they are interpreted as    */
    /* 'not reported'.                                                      */

    BYTE    IniFlags;           /*   F2 Initiator mode flags                */
        #define HWI_FCPHI_IF_SCSI       FC_IT_MODE_SCSI
        #define HWI_FCPHI_IF_NVME       FC_IT_MODE_NVME

    BYTE    TgtFlags;           /*   F3 Target mode flags                   */
        #define HWI_FCPHI_TF_SCSI       FC_IT_MODE_SCSI
        #define HWI_FCPHI_TF_NVME       FC_IT_MODE_NVME

    /* The following field is only valid for HWI_FUNC_GET_PORT_INFO */

    BYTE    PhyId;              /*   F4 PHY ID associated with the port     */
    BYTE    Reserved[3];        /*   F5 Reserved                            */
} HWIHBA_FC_PHY_INFO,           /*   F8 Size of structure                   */
*PHWIHBA_FC_PHY_INFO;


typedef struct _HwiHbaPciPhyInfo {
    BYTE    PciPort;            /*   00 PCI port or HWI_PCI_PORT_ID_INV     */
    BYTE    PhyState;           /*   01 PHY state (HWI_GPHI_PHS_XXX)        */
    BYTE    DiscState;          /*   02 Discovery state (HWI_GPHI_DS_XXX)   */
    BYTE    NegLinkRate;        /*   03 Negotiated link rate                */
    BYTE    MinLinkRateHw;      /*   04 Minimum HW link rate                */
    BYTE    MaxLinkRateHw;      /*   05 Maximum HW link rate                */
    BYTE    MinLinkRateProg;    /*   06 Minimum programmed link rate        */
    BYTE    MaxLinkRateProg;    /*   07 Maximum programmed link rate        */
                                /*      Rates above use PCI_RATE_XXX values */
    BYTE    NegLinkWidth;       /*   08 Negotiated link width               */
    BYTE    MinLinkWidthHw;     /*   09 Minimum HW link width               */
    BYTE    MaxLinkWidthHw;     /*   0A Maximum HW link width               */
    BYTE    MinLinkWidthProg;   /*   0B Minimum programmed link width       */
    BYTE    MaxLinkWidthProg;   /*   0C Maximum programmed link width       */
    BYTE    PortType;           /*   0D Flags                               */
        #define HWI_PCIPHI_DF_UNKNOWN   0x00    /* Unknown port type        */
        #define HWI_PCIPHI_DF_USP       0x01    /* Upstream port            */
        #define HWI_PCIPHI_DF_DSP       0x02    /* Downstream port          */

    WORD    PciCfgLen;          /*   0E Valid PCI configuration space data  */
    DWORD   TargetId;           /*   10 Attached target ID                  */
} HWIHBA_PCI_PHY_INFO,          /*   14 Size of structure                   */
*PHWIHBA_PCI_PHY_INFO;


typedef struct _HwiHbaNetPhyInfo {
    BYTE    PhyState;           /*   00 PHY state (HWI_GPHI_PHS_XXX)        */
    BYTE    NegLinkRate;        /*   01 Negotiated HW link rate             */
    BYTE    MinLinkRate;        /*   02 Minimum HW link rate                */
    BYTE    MaxLinkRate;        /*   03 Maximum HW link rate                */
    CHAR    IfcName[20];        /*   04 Interface name (enX)                */
} HWIHBA_NET_PHY_INFO,          /*   18 Size of structure                   */
*PHWIHBA_NET_PHY_INFO;


/****************************************************************************
                                 Port info
*****************************************************************************/

/* Port states - 'NOT_RPTD' and the others are mutually exclusive */

#define HWI_GPI_PS_NOT_RPTD         HWI_GPHI_PHS_NOT_RPTD
#define HWI_GPI_PS_OFFLINE          HWI_GPHI_PHS_LINK_DOWN
#define HWI_GPI_PS_ONLINE           HWI_GPHI_PHS_LINK_UP

/* Discovery states - 'NOT_RPTD' and the others are mutually exclusive */

#define HWI_GPI_DS_NOT_RPTD         HWI_GPHI_DS_NOT_RPTD
#define HWI_GPI_DS_IDLE             HWI_GPHI_DS_IDLE
#define HWI_GPI_DS_IN_PROG          HWI_GPHI_DS_IN_PROG
#define HWI_GPI_DS_NOT_APPL         HWI_GPHI_DS_NOT_APPL


typedef struct _HwiHbaSasPortInfo {
    DWORD   PhyMask;            /*   00 Mask of aggregated PHYs             */
    BYTE    PhyBias;            /*   04 PHY ID for bit 0 of PhyMask         */
    BYTE    PortState;          /*   05 Port state (HWI_GPI_PS_XXX)         */
    BYTE    DiscState;          /*   06 Discovery state (HWI_GPI_DS_XXX)    */
    BYTE    NegLinkRate;        /*   07 Negotiated link rate (SMP_RATE_XXX) */
    WORD    ConnMask;           /*   08 Mask of aggregated connectors       */
    BYTE    Reserved[2];        /*   0A Reserved                            */
} HWIHBA_SAS_PORT_INFO,         /*   0C Size of structure                   */
*PHWIHBA_SAS_PORT_INFO;


typedef struct _HwiHbaFcPhyInfo HWIHBA_FC_PORT_INFO, *PHWIHBA_FC_PORT_INFO;
typedef struct _HwiHbaPciPhyInfo HWIHBA_PCI_PORT_INFO, *PHWIHBA_PCI_PORT_INFO;
typedef struct _HwiHbaNetPhyInfo HWIHBA_NET_PORT_INFO, *PHWIHBA_NET_PORT_INFO;


/****************************************************************************
                                    DPM
*****************************************************************************/

/* HWIHBA_DPM_VAL is the psuedo-data payload.  Due to variable length strings
and binary blobs, this structure intentionally does not define byte/character
arrays for the sake of using sizeof() which would impose an arbitrary size
limit.  Values can have 'infinite' length.  This structure is largely for
applications to reserve a place to hold values which may allocate additional
memory.                                             
*/

typedef union _HwiHbaDpmVal {
    BYTE    AnyType[1];
    PBYTE   AnyTypePtr;
    PSTR    Ascii;
    PBYTE   Binary;
    DWORD   Bool;
        #define HWI_DPMV_TRUE   0x00000001
        #define HWI_DPMV_FALSE  0x00000000

    DWORD   Dword;
    QWORD   Qword;
} HWIHBA_DPM_VAL,
*PHWIHBA_DPM_VAL;


typedef union _HwiHbaDpmValPtr {
    PBYTE   AnyType;
    PSTR    Ascii;
    PBYTE   Binary;
    PDWORD  Bool;
    PDWORD  Dword;
    PQWORD  Qword;
} HWIHBA_DPM_VAL_PTR;


/****************************************************************************
                            Misc. adapter info
*****************************************************************************/

typedef struct _HWIHBA_TEMP_INFO {
    BYTE    TempSensor;     /* + 04 Temperature sensor index                */
    BYTE    TempState;      /*  -05 Temperature sensor state                */
        #define HWI_AC_TS_UNSUPP        0x00    /* No temp sensors          */
        #define HWI_AC_TS_UNKNOWN       0x01    /* Unknown state            */
        #define HWI_AC_TS_INIT_FAILED   0x02    /* Temp sensor init failed  */
        #define HWI_AC_TS_NORMAL        0x03    /* Normal operating cond    */
        #define HWI_AC_TS_OUT_OF_RANGE  0x04    /* Temp out of range        */
        #define HWI_AC_TS_FAULT         0x05    /* Temp sensor is inoperable*/

    SWORD   TempValue;      /*  -06 Current temp in degrees Celsius         */
    SWORD   TempLowerLim;   /*  -08 Valid temp lower limit, inclusive       */
    SWORD   TempUpperLim;   /*  -0A Valid temp upper limit, inclusive       */
    CHAR    TempDesc[32];   /*  -0C Temperature sensor description          */
    BYTE    Reserved2[20];  /*   2C Reserved                                */
} HWIHBA_TEMP_INFO, *PHWIHBA_TEMP_INFO;


#pragma pack ()


#endif  /* HWIDATA_H */
