/****************************************************************************

DESCRIPTION: Public hardware layer interface definitions.

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2012-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef HWPUBLIC_H
#define HWPUBLIC_H


#include <hwtypes.h>


/* TM:!! HW request status */
#define HWR_STS_SUCCESS             0x00 /* Success                         */
#define HWR_STS_ERROR               0x01 /* General request error           */
#define HWR_STS_BUSY                0x02 /* Busy (out of resources, etc)    */
#define HWR_STS_TIMEOUT             0x03 /* Request timed out               */
#define HWR_STS_ABORTED             0x05 /* Aborted (incl. task management) */
#define HWR_STS_BUS_RESET           0x06 /* Aborted (adapter/port reset)    */
#define HWR_STS_TARG_NOT_PRESENT    0x07 /* Target no longer present        */
#define HWR_STS_TARG_INV            0x08 /* Target never present or invalid */
#define HWR_STS_SCSI_ERROR          0x09 /* SCSI error                      */
#define HWR_STS_DATA_OVERRUN        0x0A /* Data overrun                    */
#define HWR_STS_DATA_UNDERRUN       0x0B /* Data underrun                   */
#define HWR_STS_PROT_TIMEOUT        0x40 /* Protocol timeout                */
#define HWR_STS_PROT_RETRY          0x41 /* Protocol error (recoverable)    */
#define HWR_STS_PROT_ERROR          0x42 /* Protocol error                  */
#define HWR_STS_FC_SFS_REJECT       0x48 /* FC SFS reject (SM-HBA)          */
#define HWR_STS_FC_SFS_FAILED       0x49 /* FC SFS failed (SM-HBA)          */
#define HWR_STS_SAS_IT_NEXUS_LOSS   0x4C /* SAS I_T Nexus loss              */
#define HWR_STS_SAS_OPEN_BREAK      0x4D /* SAS Open Break                  */
#define HWR_STS_SAS_OPEN_RETRY_TMO  0x4E /* SAS Open Retry Timeout          */
#define HWR_STS_SAS_OPEN_WRONG_DEST 0x4F /* SAS Open Wrong Destination      */
#define HWR_STS_PENDING             0xFF /* Request is pending (internal)   */
/* Tm:!! HW request status */


/* Logging levels.  Selecting a logging level logs that level and lower */

enum {
    HWLOG_LVL_DISABLE,      /* All logging disabled                         */
    HWLOG_LVL_ALWAYS,       /* Informational message, always logged         */
    HWLOG_LVL_ALERT,        /* Critical issue requiring immediate attention */
    HWLOG_LVL_CRITICAL,     /* Critical info/error, may accompany alert     */
    HWLOG_LVL_ERROR,        /* Error that may require attention             */
    HWLOG_LVL_WARNING,      /* All other errors                             */
    HWLOG_LVL_NOTICE,       /* Informational message, may require attention */
    HWLOG_LVL_INFO,         /* All other informational messages             */
    HWLOG_LVL_VERBOSE1,     /* Verbose messages, may degrade performance    */
    HWLOG_LVL_VERBOSE2,     /* Verbose messages, may degrade performance    */
    HWLOG_LVL_VERBOSE3,     /* Verbose messages, may degrade performance    */
    HWLOG_LVL_DEBUG,        /* Debug messages, may degrade performance      */
    HWLOG_LVL_CNT
};


/* Target states

Used by OsTargStateChanged().  OS layers will typically
handle the NOT_PRESENT, PRESENT, and LUN_CHANGE states to
manage device arrivals and removals.

The INACTIVE and ACTIVE states serve as bookends to the link down
timeout period when waiting for an unavailable device to become
available again.  Multipathing solutions can use these states to do
immediate failover and failback to keep the device active with minimal
delay.

ACTIVE_LOCATION and ACTIVE_INFO are special cases of ACTIVE to
indicate device parameters have changed.  ACTIVE_LOCATION indicates
the device's physical location in the topology has changed.
ACTIVE_INFO indicates the location is the same, but either the PHYs
in the port or link rate have changed.  Note that if all PHYs in the
activated port have changed, it is considered a location change.

IN_TRANSITION is used by the OS layer for ALUA failover, but is defined
here to coordinate with other states.

Note that the states below are defined in order of decreasing severity.
The HW layer reports the most severe events first to assist the OS layer
in any device mapping algorithms and prevent temporary situations where
the OS has knowledge of an invalid topology (e.g. two direct attached
devices reported as present on the same port - the removal of one will
be reported first).
*/

/* TM:!! HW target state */
#define HWTS_NOT_PRESENT        0x00 /* Device removed from topology        */
#define HWTS_INACTIVE           0x01 /* Device in link down timeout period  */
#define HWTS_ACTIVE_LOCATION    0x02 /* Device active in a new location     */
#define HWTS_ACTIVE_INFO        0x03 /* Device active but parameters changed*/
#define HWTS_ACTIVE             0x04 /* Device became active while retrying */
#define HWTS_PRESENT            0x05 /* Device added to topology            */
#define HWTS_LUN_CHANGE         0x06 /* Device LUN inventory changed        */
#define HWTS_IN_TRANSITION      0x07 /* Device state in transition          */
                                     
#define HWTS_COUNT              0x08 /* Number of states                    */
                                     
#define HWTS_INVALID            0xFF /* Internal invalid state              */
/* Tm:!! HW target state */


/* Target types

Used by OsTargStateChanged().  Indicates the target type for which the state
change is being reported.  

The value passed is a bit mask; however, the EXPANDER and END_DEVICE bits
are mutually exclusive.  The HW layer should set multiple END_DEVICE bits (when
appropriate) in one call if such devices are supported.  This facilitates
the OS layer option of scheduling a single work item for a target ID to service
the change, or processing each type independently.
*/

/* TM:!! HW target type */
#define HWTARG_TYPE_SCSI        0x01    /* SCSI device                  */
#ifdef HW_NVME_INI
#define HWTARG_TYPE_NVME        0x02    /* NVMe device                  */
#else
#define HWTARG_TYPE_NVME        0x00
#endif
#define HWTARG_TYPE_EXPANDER    0x80    /* SAS expander                 */
/* Tm:!! HW target type */
#define HWTARG_TYPE_END_DEVICE  (HWTARG_TYPE_SCSI | HWTARG_TYPE_NVME)


/* HWINIT_MSI contains all parameters for HwInitAdapCfg().  This structure  */
/* should be cleared before initializing it to allow additional parameters  */
/* to be added without the need for additional option flags.                */

typedef struct _HWINIT_MSI {

    /* The Options field is set by the OS layer to indicate it's            */
    /* capabilities to the HW layer.  Upon return, the Results field is     */
    /* filled in with the options that were applicable to the current       */
    /* MSI(-X) configuration.                                               */

    /* HWINITMSI_MULTI_VECTOR - Indicates whether the ISR can determine     */
    /* which MSI interrupted and pass the vector to HwIntMsi().             */

    /* HWINITMSI_MSIX_ENB/HWINITMSI_MSI_ENB - Result flags that indicate    */
    /* whether MSI-X or MSI is enabled.  Legacy interrupts are enabled when */
    /* both are clear.                                                      */

    BYTE    Options;            /* + 00 00 OS capabilities allowed          */
    BYTE    Results;            /*  -01 01 HW capabilities enabled          */

        #define HWINITMSI_MULTI_VECTOR     0x01
        #define HWINITMSI_MSIX_ENB         0x04
        #define HWINITMSI_MSI_ENB          0x08

        #define HWINITMSI_MSI_MSIX_ENB (HWINITMSI_MSIX_ENB | HWINITMSI_MSI_ENB)

    /* The following are returned by the HW layer after successfully        */
    /* initializing MSI(-X).  They indicate the range of MSI(-X) vectors    */
    /* assigned to I/O queues.                                              */

    BYTE    FirstIoMsi;         /*  -02 02 First 0-based vector used for I/O*/
    BYTE    LastIoMsi;          /*  -03 03 Last 0-based vector used for I/O */

    /* MsiCnt - The MSI-X vector count that was enabled.   The HW layer     */
    /* will limit the number of vectors used to this count.  OS layers may  */
    /* need to use this if the desired count is lower than the CPU count.   */
    /* Zero indicates the value is unspecified.                             */

    BYTE    MsiCnt;             /* +-04 04 Total vector count               */

    /* MsiCntIo - The MSI(-X) vector count that may be used for I/O.  Zero  */
    /* indicates the CPU count is used.  OS limitations may dictate         */
    /* setting this to something less such as the NUMA node count or 1.     */
    /* This setting has no effect if HWINITMSI_MULTI_VECTOR is not set.     */

    BYTE    MsiCntIo;           /* +-05 05 Vector count for I/O             */

    BYTE    Pad[2];             /*   06 06 System processor count           */
} HWINIT_MSI, *PHWINIT_MSI;     /*   08 08 Size of structure                */


#endif /* HWPUBLIC_H */
