/****************************************************************************

DESCRIPTION: SAS definitions.

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2008-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef HWSAS_H
#define HWSAS_H


#include <hwscsi.h>


/****************************************************************************
                               Definitions
*****************************************************************************/

/* Initiator/target modes - located in various structures including SMP
Discover and the Identify frame.  Defined generically here to use in various
interfaces.
*/

/* TF:!! SAS I_T mode */
#define SAS_IT_MODE_SATA        0x01    /* SATA initiator/target             */
#define SAS_IT_MODE_SMP         0x02    /* SMP  initiator/target             */
#define SAS_IT_MODE_STP         0x04    /* STP  initiator/target             */
#define SAS_IT_MODE_SSP         0x08    /* SSP  initiator/target             */
/* Tf:!! SAS I_T mode */

/* Link rates */

#define SAS_LINK_RATE_1_5GB             0x08
#define SAS_LINK_RATE_3GB               0x09
#define SAS_LINK_RATE_6GB               0x0A
#define SAS_LINK_RATE_12GB              0x0B
#define SAS_LINK_RATE_24GB              0x0C

/* Address Frame Types */

#define SAS_ADDR_FRAME_TYPE_IDENTIFY    0x00
#define SAS_ADDR_FRAME_TYPE_OPEN        0x01

/* Frame types */

#define SAS_FRAME_TYPE_SSP_DATA         0x01
#define SAS_FRAME_TYPE_SSP_XFER_RDY     0x05
#define SAS_FRAME_TYPE_SSP_CMD          0x06
#define SAS_FRAME_TYPE_SSP_RSP          0x07
#define SAS_FRAME_TYPE_SSP_TASK         0x16
#define SAS_FRAME_TYPE_SMP_REQ          0x40
#define SAS_FRAME_TYPE_SMP_RSP          0x41

/* Transport Layer Retry Control */

#define SAS_TLR_CTRL_MODE_PAGE1         0x00    /* Use mode page (SAS-1,2)  */
#define SAS_TLR_CTRL_ENABLE             0x01    /* Always enable (SAS-2)    */
#define SAS_TLR_CTRL_DISABLE            0x02    /* Always disable (SAS-2)   */
#define SAS_TLR_CTRL_MODE_PAGE2         0x03    /* Use mode page (SAS-2)    */

/* SNW-3 PHY Capabilities */

#define SAS_SNW3_LINK_RATE_MASK         DW(0x0000000F)
#define SAS_SNW3_SSC_TYPE_CENTER        DW(0x00000040)
#define SAS_SNW3_SSC_TYPE_DOWN          DW(0x00000000)
#define SAS_SNW3_START                  DW(0x00000080)
#define SAS_SNW3_G4_SSC_ENB             DW(0x00000100)
#define SAS_SNW3_G4_SSC_DIS             DW(0x00000200)
#define SAS_SNW3_G3_SSC_ENB             DW(0x00000400)
#define SAS_SNW3_G3_SSC_DIS             DW(0x00000800)
#define SAS_SNW3_G2_SSC_ENB             DW(0x00001000)
#define SAS_SNW3_G2_SSC_DIS             DW(0x00002000)
#define SAS_SNW3_G1_SSC_ENB             DW(0x00004000)
#define SAS_SNW3_G1_SSC_DIS             DW(0x00008000)
#define SAS_SNW3_PARITY                 DW(0x01000000)


/****************************************************************************
                               Structures
*****************************************************************************/

/* IDENTIFY Address Frame */

typedef struct SasIdentifyAddressFrame {
    BYTE        Type;           /* 00 Device/frame type                     */
        #define SIAF_FRAME_TYPE_MASK    0x0F    /* Frame type mask          */
        #define SIAF_DEV_TYPE_MASK      0x70    /* Device type mask         */
        #define SIAF_DT_END_DEVICE      0x10    /* End device               */
        #define SIAF_DT_EXPANDER        0x20    /* Expander (Edge in SAS-1) */
        #define SIAF_DT_FANOUT_EXP      0x30    /* Fanout expander (OBS)    */

    BYTE        Reason;         /* 01 Reset reason - Added in SAS-2         */
        #define SIAF_RSN_MASK           0x0F    /* Reason mask              */
        #define SIAF_RSN_UNKNOWN        0x00    /* Unknown reason           */
        #define SIAF_RSN_POWER_ON       0x01    /* Power on                 */
        #define SIAF_RSN_HARD_RST       0x02    /* Hard reset               */
        #define SIAF_RSN_SMP_LINK_RST   0x03    /* SMP link reset           */
        #define SIAF_RSN_LOSS_OF_SYNC   0x04    /* Loss of DWORD sync       */
        #define SIAF_RSN_MUX            0x05    /* MUX received             */
        #define SIAF_RSN_STP_NEXUS_LOSS 0x06    /* STP I_T nexus loss       */
        #define SIAF_RSN_BREAK_TIMEOUT  0x07    /* BREAK timer expired      */
        #define SIAF_RSN_PHY_TEST_DONE  0x08    /* PHY test stopped         */
        #define SIAF_RSN_EXP_RED_FUNC   0x09    /* Expander reduced func.   */

    BYTE        IniFlags;       /* 02 Initiator flags                       */
        #define SIAF_IF_SMP             0x02    /* SMP initiator            */
        #define SIAF_IF_STP             0x04    /* STP initiator            */
        #define SIAF_IF_SSP             0x08    /* SSP initiator            */

    BYTE        TgtFlags;       /* 03 Target flags                          */
        #define SIAF_TF_SMP             0x02    /* SMP target               */
        #define SIAF_TF_STP             0x04    /* STP target               */
        #define SIAF_TF_SSP             0x08    /* SSP target               */

    BYTE        DeviceName[8];  /* 04 Device name - Added in SAS-2          */
    BYTE        SasAddr[8];     /* 0C SAS address                           */
    BYTE        PhyId;          /* 14 PHY identifier                        */
    BYTE        Flags;          /* 15 Flags - Added in SAS-2                */
        #define SIAFF_BREAK_REP_CAP     0x01    /* Capable of BREAK_REPLY   */
        #define SIAFF_REQ_IN_ZPSDS      0x02    /* Requested inside ZPSDS   */
        #define SIAFF_IN_ZPSDS_PERS     0x04    /* Inside ZPSDS Persistent  */
        #define SIAFF_PARTIAL           0x08    /* Partial capable          */
        #define SIAFF_SLUMBER           0x10    /* Slumber capable          */
        #define SIAFF_POWER_MASK        0x60    /* Power capabilities mask  */
        #define SIAFF_POWER_SOURCE      0x40    /* Grant additional power   */
        #define SIAFF_POWER_CONSUMER    0x20    /* Consume additional power */

    BYTE        Reserved[6];    /* 06 Reserved                              */
} SAS_IDENT_FRAME,              /* 1C Size of structure                     */
*PSAS_IDENT_FRAME;


/* OPEN Address Frame */

typedef struct SasOpenAddressFrame {
    BYTE        Type;           /* 00 Device/frame type                     */
        #define SOAF_FRAME_INIT_MASK    0x80    /* Initiator port mask      */
        #define SOAF_FRAME_INIT_TARG    0x00    /* Target port              */
        #define SOAF_FRAME_INIT_INIT    0x80    /* Initiator port           */
        #define SOAF_FRAME_PROTO_MASK   0x70    /* Frame protocol mask      */
        #define SOAF_FRAME_PROTO_SMP    0x00    /* Frame is SMP             */
        #define SOAF_FRAME_PROTO_SSP    0x10    /* Frame is SSP             */
        #define SOAF_FRAME_PROTO_STP    0x20    /* Frame is STP             */
        #define SOAF_FRAME_TYPE         0x01    /* Address frame type       */

    BYTE        FeatConn;       /* 01 Features and Connection rate          */
        #define SOAF_FEATURES_MASK      0xF0    /* Features Mask            */
        #define SOAF_CONN_RATE_MASK     0x0F    /* Connection Rate Mask     */

    WORD        InitTag;        /* 02 Initiator connection Tag              */
    BYTE        DestSasAddr[8]; /* 04 Destination SAS Address               */
    BYTE        SrcSasAddr[8];  /* 0C Source SAS Address                    */
    BYTE        SrcZoneGrp;     /* 14 Source Zone Group                     */
    BYTE        PathBlckCnt;    /* 15 Pathway blocked count                 */
    WORD        ArbWaitTime;    /* 16 Arbitration Wait time                 */
    DWORD       MoreCompFeat;   /* 18 More compatible features              */
} SAS_OPEN_FRAME,               /* 1C Size of structure                     */
*PSAS_OPEN_FRAME;


/* SSP Frame Header */

typedef struct SasSspFrameHeader {
    BYTE        FrameType;              /* 00 Frame type - defines the IU   */
    BYTE        HashedDestSasAddr[3];   /* 01 Destination SAS Address       */
    BYTE        Reserved1;              /* 04 Reserved                      */
    BYTE        HashedSrcSasAddr[3];    /* 05 Source SAS Address            */
    BYTE        Reserved2[2];           /* 08 Reserved                      */
    BYTE        SspFrmFlags;            /* 0A Flags                         */
        #define SSFH_CHG_DATA_PTR       0x01    /* Change Data Pointer      */
        #define SSFH_RETRANSMIT         0x02    /* Retransmit previous frame*/
        #define SSFH_RTRY_DATA_FRMS     0x04    /* Enable TLR for write DATA*/
        #define SSFH_TLR_CTRL_SH        3       /* SAS_TLR_CTRL_XXX         */

    BYTE        NumFillBytes;           /* 0B Number of fill bytes          */
        #define SSFH_NFB_MASK           0x03

    BYTE        Reserved3[4];           /* 0C Reserved                      */
    WORD        IniTag;                 /* 10 Initiator port transfer tag   */
    WORD        TgtTag;                 /* 12 Target port transfer tag      */
    DWORD       DataOffset;             /* 14 Data offset                   */
} SAS_SSP_FRM_HDR,                      /* 18 Size of structure             */
*PSAS_SSP_FRM_HDR;


/* SSP Command Information Unit */

typedef struct SasSspCmdIU {
    BYTE        ScsiLun[8];     /* 00 LUN                                   */
    BYTE        Reserved;       /* 08 Reserved                              */
    BYTE        Flags;          /* 09 Flags                                 */
        #define SSCIUF_ATTR_MASK    0x07    /* Task attribute mask          */
        #define SSCIUF_ATTR_SIMPLE  0x00    /* Simple queue                 */
        #define SSCIUF_ATTR_HEAD_Q  0x01    /* Head of queue                */
        #define SSCIUF_ATTR_ORDERED 0x02    /* Ordered queue                */
        #define SSCIUF_ATTR_ACA     0x04    /* ACA                          */
        #define SSCIUF_PRIORTY_MASK 0x78    /* Command priority mask        */
        #define SSCIUF_ENB_1ST_BRST 0x80    /* Enable first burst           */

    BYTE        Reserved2;      /* 0A Reserved                              */
    BYTE        AddCdbLen;      /* 0B Additional CDB length                 */
        #define SSCIU_ACL_MASK      0x1C    /* Additional CDB length mask   */
        #define SSCIU_ACL_SHIFT        2    /* Additional CDB length shift  */

    BYTE        Cdb[16];        /* 0C CDB                                   */
} SAS_SSP_CMD_IU,               /* 1C Size of structure                     */
*PSAS_SSP_CMD_IU;


/* SSP Extended Command Information Unit */

typedef struct SasExtSspCmdIU {
    BYTE        ScsiLun[8];     /* 00 LUN                                   */
    BYTE        Reserved;       /* 08 Reserved                              */
    BYTE        Flags;          /* 09 Flags                                 */
    BYTE        Reserved2;      /* 0A Reserved                              */
    BYTE        AddCdbLen;      /* 0B Additional CDB length                 */
    BYTE        Cdb[32];        /* 0C CDB                                   */
} SAS_EXT_SSP_CMD_IU,           /* 2C Size of structure                     */
*PSAS_EXT_SSP_CMD_IU;


/* SSP Task Management Information Unit */

typedef struct SasSspTaskIU {
    BYTE        ScsiLun[8];     /* 00 LUN                                   */
    BYTE        Reserved[2];    /* 08 Reserved                              */
    BYTE        TskMgtFunc;     /* 0A Task management function              */
        #define SSTIU_FUNC_ABORT    0x01    /* Abort task                   */
        #define SSTIU_FUNC_ATS      0x02    /* Abort task set               */
        #define SSTIU_FUNC_CTS      0x04    /* Clear task set               */
        #define SSTIU_FUNC_LU_RST   0x08    /* Logical unit reset           */
        #define SSTIU_FUNC_TGT_RST  0x10    /* Target reset                 */
                                            /* I T Nexus Reset              */
        #define SSTIU_FUNC_CLR_ACA  0x40    /* Clear ACA                    */
        #define SSTIU_FUNC_QUERY    0x80    /* Query task                   */
        #define SSTIU_FUNC_QTS      0x81    /* Query task set               */
        #define SSTIU_FUNC_QUA      0x82    /* Query unit attention         */

    BYTE        Reserved2;      /* 0B Reserved                              */
    WORD        IniTag;         /* 0C Tag of task to be managed             */
    BYTE        Reserved3[14];  /* 0E Reserved                              */
} SAS_SSP_TASK_IU,              /* 1C Size of structure                     */
*PSAS_SSP_TASK_IU;


/* SSP Response Information Unit */

typedef struct SasSspRspIU {
    BYTE        Reserved[8];    /* 00 Reserved                              */
    WORD        StsQual;        /* 08 Status qualifier                      */
    BYTE        Flags;          /* 0A Flags                                 */
        #define SSRIUF_DATA_MASK    0x03    /* Data present mask            */
        #define SSRIUF_DP_NO_DATA   0x00    /* No data present              */
        #define SSRIUF_DP_RSP_DATA  0x01    /* Response data present        */
        #define SSRIUF_DP_SNS_DATA  0x02    /* Sense data present           */

    BYTE        ScsiStatus;     /* 0B SCSI Status                           */
    BYTE        Reserved2[4];   /* 0C Reserved                              */
    DWORD       SnsDataLen;     /* 10 Sense data length                     */
    DWORD       RspDataLen;     /* 14 Response data length                  */
} SAS_SSP_RSP_IU,               /* 18 Size of structure                     */
*PSAS_SSP_RSP_IU;

typedef struct SasSspRspIURspData {
    BYTE        AddRspInfo[3];  /* 00 Additional response information       */
    BYTE        RspCode;        /* 03 Response code                         */
        /* TM:!! SAS response code */
        #define SSRD_RC_TSKMGT_COMP 0x00    /* Task management complete     */
        #define SSRD_RC_INV_FRAME   0x02    /* Invalid frame                */
        #define SSRD_RC_TSKMGT_NA   0x04    /* Task management unsupported  */
        #define SSRD_RC_TSKMGT_FAIL 0x05    /* Task management failed       */
        #define SSRD_RC_TSKMGT_OK   0x08    /* Task management succeeded    */
        #define SSRD_RC_INV_LUN     0x09    /* Invalid LUN                  */
        #define SSRD_RC_INV_TAG     0x0A    /* Overlapped tag               */
        /* Tm:!! SAS response code */

} SAS_SRIU_RSP_DATA,            /* 04 Size of structure                     */
*PSAS_SRIU_RSP_DATA;


/* SSP XFER_RDY Information Unit   */

typedef struct SasSspXfrRdyIU {
    DWORD       ReqOffset;      /* 00 Requested offset                      */
    DWORD       WrtDataLen;     /* 04 Write data length                     */
    DWORD       Reserved;       /* 08 Reserved                              */
} SAS_SSP_XFRRDY_IU,            /* 0C Size of structure                     */
*PSAS_SSP_XFRRDY_IU;

#endif /* HWSAS_H */

