/****************************************************************************

DESCRIPTION: SAS SMP definitions.

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2006-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef HWSASSMP_H
#define HWSASSMP_H


/****************************************************************************
                               Definitions
*****************************************************************************/

/* Function codes */

/* TM:!! SMP function */
#define SMP_FUNC_RPT_GENERAL        0x00    /* Report General               */
#define SMP_FUNC_RPT_MFR            0x01    /* Report Manufacturer Info     */
#define SMP_FUNC_READ_GPIO          0x02    /* Read GPIO Register           */
#define SMP_FUNC_RPT_SELF_CFG_STS   0x03    /* Report Self Config Status    */
#define SMP_FUNC_RPT_ZN_PERM_TBL    0x04    /* Report Zone Permission Table */
#define SMP_FUNC_RPT_ZN_MGR_PASSWD  0x05    /* Report Zone Manager Password */
#define SMP_FUNC_RPT_BROADCAST      0x06    /* Report Broadcast             */

#define SMP_FUNC_DISC               0x10    /* Report PHY info              */
#define SMP_FUNC_RPT_PHY_ERR_LOG    0x11    /* Report PHY Error Log         */
#define SMP_FUNC_RPT_PHY_SATA       0x12    /* Report PHY SATA Info         */
#define SMP_FUNC_RPT_ROUTE_INFO     0x13    /* Report Route Table Info      */
#define SMP_FUNC_RPT_PHY_EVENT      0x14    /* Report PHY Event             */

#define SMP_FUNC_DISC_LIST          0x20    /* Report PHY Info              */
#define SMP_FUNC_RPT_PHY_EVENT_LIST 0x21    /* Report PHY Event List        */
#define SMP_FUNC_RPT_EXP_ROUTE_TBL  0x22    /* Report Expander Route Table  */

#define SMP_FUNC_CFG_GENERAL        0x80    /* Config General               */
#define SMP_FUNC_ENB_DIS_ZONING     0x81    /* Enable/Disable Zoning        */
#define SMP_FUNC_WRITE_GPIO         0x82    /* Write GPIO Register          */

#define SMP_FUNC_ZN_BROADCAST       0x85    /* Broadcast on specific zones  */
#define SMP_FUNC_ZN_LOCK            0x86    /* Lock a zone                  */
#define SMP_FUNC_ZN_ACTIVATE        0x87    /* Activate a zone              */
#define SMP_FUNC_ZN_UNLOCK          0x88    /* Unlock a zone                */
#define SMP_FUNC_CFG_ZN_MGR_PASSWD  0x89    /* Config Zone Manager Password */
#define SMP_FUNC_CFG_ZN_PHY_INFO    0x8A    /* Config Zone PHY Info         */
#define SMP_FUNC_CFG_ZN_PERM_TBL    0x8B    /* Config Zone Permission table */

#define SMP_FUNC_CFG_ROUTE_INFO     0x90    /* Config Route Table Info      */
#define SMP_FUNC_PHY_CTRL           0x91    /* Request PHY actions          */
#define SMP_FUNC_PHY_TEST           0x92    /* Request PHY test functions   */
#define SMP_FUNC_CFG_PHY_EVENT      0x93    /* Config PHY Event             */
/* Tm:!! SMP function */

/* Result codes */

/* TM:!! SMP result */
#define SMP_RSLT_SUCCESS            0x00    /* Success                      */
#define SMP_RSLT_UNK_FUNC           0x01    /* Function not supported       */
#define SMP_RSLT_FAILED             0x02    /* Function failed              */
#define SMP_RSLT_INV_REQ_LEN        0x03    /* Invalid request frame length */
#define SMP_RSLT_INV_EXP_CHG_CNT    0x04    /* Invalid expander change count*/
#define SMP_RSLT_BUSY               0x05    /* Zones being updated          */
#define SMP_RSLT_INV_DESC_LIST      0x06    /* Invalid descriptor list      */

#define SMP_RSLT_INV_PHY            0x10    /* Specified PHY does not exist */
#define SMP_RSLT_INV_INDEX          0x11    /* Invalid route index          */
#define SMP_RSLT_NOT_SATA           0x12    /* PHY not part of SATA port    */
#define SMP_RSLT_UNK_PHY_OP         0x13    /* Unknown PHY operation        */
#define SMP_RSLT_UNK_PHY_TEST       0x14    /* Unknown PHY test function    */
#define SMP_RSLT_PHY_TESTING        0x15    /* PHY test function in prog.   */
#define SMP_RSLT_PHY_VACANT         0x16    /* Unable to access PHY         */
#define SMP_RSLT_UNK_PHY_EVT        0x17    /* Unknown PHY event info       */
#define SMP_RSLT_UNK_DESC_TYPE      0x18    /* Unknown descriptor type      */
#define SMP_RSLT_UNK_PHY_FILTER     0x19    /* Unknown PHY filter           */
#define SMP_RSLT_AFFIL_VIOLATION    0x1A    /* Affiliation violation        */

#define SMP_RSLT_ZN_VIOLATION       0x20    /* Zone violation               */
#define SMP_RSLT_ZN_NO_ACCESS       0x21    /* No management access rights  */
#define SMP_RSLT_ZN_UNK_ENB_DIS     0x22    /* Unknown enb/dis zoning value */
#define SMP_RSLT_ZN_LOCK_VIOL       0x23    /* Zone lock violation          */
#define SMP_RSLT_ZN_NOT_ACTIVATED   0x24    /* Zone not activated           */
#define SMP_RSLT_ZN_INV_GRP         0x25    /* Invalid zone group           */
#define SMP_RSLT_ZN_NO_PHYS_PRES    0x26    /* No physical presence         */
#define SMP_RSLT_ZN_NO_SAVE         0x27    /* Saving not supported         */
#define SMP_RSLT_ZN_INV_SRC_GRP     0x28    /* Invalid zone source group    */
#define SMP_RSLT_ZN_DIS_PW_NOT_SUPP 0x29    /* Disabled password not supp.  */
/* Tm:!! SMP result */

/* Device types */

/* TM:!! SMP device type */
#define SMP_DEVTYPE_NONE            0x00    /* No device (attached)         */
#define SMP_DEVTYPE_END_DEV         0x01    /* End device                   */
#define SMP_DEVTYPE_EDGE_EXP        0x02    /* Edge expander device         */
#define SMP_DEVTYPE_FANOUT_EXP      0x03    /* Fanout expander device       */
/* Tm:!! SMP device type */

/* Link Rates */

/* TM:!! SMP link rate */
#define SMP_RATE_UNKNOWN            0x00    /* Enabled, rate unknown        */
#define SMP_RATE_DISABLED           0x01    /* PHY is disabled              */
#define SMP_RATE_NEG_FAIL           0x02    /* Speed negotiation failed     */
#define SMP_RATE_SATA_HOLD          0x03    /* SATA Spin-up hold state      */
#define SMP_RATE_PORT_SEL           0x04    /* Detected SATA port selector  */
#define SMP_RATE_RESET              0x05    /* Reset in progress            */
#define SMP_RATE_UNSUPPORTED_PHY    0x06    /* Unsupported PHY attached     */
#define SMP_RATE_RESERVED           0x07    /* Reserved                     */
#define SMP_RATE_1_5GB              0x08    /* 1.5 Gbps                     */
#define SMP_RATE_3GB                0x09    /* 3.0 Gbps                     */
#define SMP_RATE_6GB                0x0A    /* 6.0 Gbps                     */
#define SMP_RATE_12GB               0x0B    /* 12  Gbps                     */
#define SMP_RATE_24GB               0x0C    /* 24  Gbps                     */
/* Tm:!! SMP link rate */

/* Routing */

/* TM:!! SMP route attribute */
#define SMP_ROUTE_DIRECT            0x00    /* Direct routing attribute     */
#define SMP_ROUTE_SUBTRACTIVE       0x01    /* Subtractive routing          */
#define SMP_ROUTE_TABLE             0x02    /* Table routing                */
/* Tm:!! SMP route attribute */

/* Broadcast types */

#define SMP_BC_TYPE_CHANGE          0x00    /* Change                       */
#define SMP_BC_TYPE_RES_CHG0        0x01    /* Reserved change 0            */
#define SMP_BC_TYPE_RES_CHG1        0x02    /* Reserved change 1            */
#define SMP_BC_TYPE_SES             0x03    /* SES                          */
#define SMP_BC_TYPE_EXP             0x04    /* Expander                     */
#define SMP_BC_TYPE_AE              0x05    /* Async event                  */
#define SMP_BC_TYPE_RES3            0x06    /* Reserved 3                   */
#define SMP_BC_TYPE_RES4            0x07    /* Reserved 4                   */
#define SMP_BC_TYPE_ZN_ACT          0x08    /* Zone activate                */

#define SMP_BC_TYPE_GET(b)          ((b) & 0x0F)

/* Zoning save types */

#define SMP_ZN_SAVE_CURR            0x00    /* Current (shadow)             */
#define SMP_ZN_SAVE_SAVED           0x01    /* Saved                        */
#define SMP_ZN_SAVE_SAVED_OPT_CURR  0x02    /* Saved if supp. and current   */
#define SMP_ZN_SAVE_SAVED_CURR      0x03    /* Saved and current            */

#define SMP_ZN_SAVE_GET(b)          ((b) & 0x03)

/* Zoning max group count */

#define SMP_ZN_GRP_MAX_128          0x00    /* 128 zone groups supported    */
#define SMP_ZN_GRP_MAX_256          0x01    /* 256 zone groups supported    */

#define SMP_ZN_GRP_MAX_GET6(b)      (((b) & 0xC0) / LOBIT(0xC0))
#define SMP_ZN_GRP_MAX_VAL6(v)      ((v) << 6)

/* Zoning PHY info flags */

#define SMP_ZN_PHYF_ZN_ENB          0x01    /* Zoning enabled               */
#define SMP_ZN_PHYF_IN_ZPSDS        0x02    /* Inside ZPSDS                 */
#define SMP_ZN_PHYF_ZN_GRP_PERS     0x04    /* Zone group persistent        */
#define SMP_ZN_PHYF_REQ_IN_ZPSDS    0x10    /* Requested inside ZPSDS       */
#define SMP_ZN_PHYF_IN_ZPSDS_PERS   0x20    /* Inside ZPSDS persistent      */
#define SMP_ZN_PHYF_ZPSDS_CHG       0x40    /* Expander changed Req. ZPSDS  */

/* Zoning report types */

#define SMP_ZM_RPT_TYPE_CURR        0x00    /* Report current info          */
#define SMP_ZM_RPT_TYPE_SHADOW      0x01    /* Report shadow info           */
#define SMP_ZM_RPT_TYPE_SAVED       0x02    /* Report saved info            */
#define SMP_ZM_RPT_TYPE_DEF         0x03    /* Report default info          */

#define SMP_ZN_RPT_TYPE_GET(b)      ((b) & 0x03)


/****************************************************************************
                               Structures
*****************************************************************************/

/* Pack the data structures since some of them have SAS addresses (QWORDs)  */
/* but are only a DWORD-multiple size.                                      */

#pragma pack (1)

/* SMP request frame header for all SMP requests. */

typedef struct SmpReqHdr {
    BYTE    FrameType;          /* 00 Frame type                            */
        #define SMPREQ_FRAME_TYPE   0x40

    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
                                /*    Added in SAS-2, may be zero           */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
                                /*    Added in SAS-2, may be zero           */
} SMPREQ_HDR,                   /* 04 Size of structure                     */
*PSMPREQ_HDR;


/* SMP response frame header for all SMP responses. */

typedef struct SmpRspHdr {
    BYTE    FrameType;          /* 00 Frame type                            */
        #define SMPRSP_FRAME_TYPE   0x41

    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
                                /*    Added in SAS-2, may be zero           */
} SMPRSP_HDR,                   /* 04 Size of structure                     */
*PSMPRSP_HDR;


/* Report General */

typedef SMPREQ_HDR  SMPREQ_RPT_GEN,          *PSMPREQ_RPT_GEN;

typedef struct SmpRspRptGen {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander Change Count                 */
    WORD    ExpRouteIndexes;    /* 06 Max route indexes per PHY             */
    BYTE    Flags2;             /* 08 Flags2                                */
        #define SMPRGF2_LONG_RSP    0x80    /* Non-zero req/rsp length supp */

    BYTE    NumPhys;            /* 09 Total number of PHYs in device        */
    BYTE    ConfigFlags;        /* 0A Configuration flags                   */
        #define SMPRG_CF_RTE_TBL      0x01  /* Configurable route table     */
        #define SMPRG_CF_CONFIGURING  0x02  /* Configuring                  */
        #define SMPRG_CF_CFG_OTHERS   0x04  /* Configures others            */
        #define SMPRG_CF_OPNRJT_RETRY 0x08  /* Supports Open Reject retry   */
        #define SMPRG_CF_STP_CONT_AWT 0x10  /* STP continue AWT             */
        #define SMPRG_CF_SELF_CFG     0x20  /* Self configuring             */
        #define SMPRG_CF_ZN_CFG       0x40  /* Zone configuring             */
        #define SMPRG_CF_TBL_TO_TBL   0x80  /* Table to table supported     */

    BYTE    ConnFlags;          /* 0B Connection flags                      */
        #define SMPRG_CF_INI_SSP_CLOSE 0x01 /* Initiates SSP CLOSE          */
        #define SMPRG_CF_EXT_FAIRNESS  0x02 /* Extended fairness            */

    BYTE    EnclosureId[8];     /* 0C Enclosure logical identifier          */
    BYTE    Reserved2[8];       /* 14 Reserved                              */
                                /*    End of SAS 1.1 fields                 */
    WORD    SspConnTo;          /* 1C SSP connection time limit             */
    WORD    StpBusInactTo;      /* 1E STP bus inactivity time limit         */
    WORD    StpMaxConnTo;       /* 20 STP maximum connection time limit     */
    WORD    StpItNexusLossTime; /* 22 STP I_T nexus loss time               */
    BYTE    ZnFlags;            /* 24 Zoning configuration flags            */
        #define SMPRG_ZF_ENB        0x01    /* Zoning enabled               */
        #define SMPRG_ZF_SUPP       0x02    /* Zoning supported             */
        #define SMPRG_ZF_PP_ASSERT  0x04    /* Physical presence asserted   */
        #define SMPRG_ZF_PP_SUPP    0x08    /* Physical presence supported  */
        #define SMPRG_ZF_LOCK       0x10    /* Zone locked                  */
        /* See  SMP_ZN_GRP_MAX_XXX                                          */

    BYTE    ZnSaveFlags;        /* 25 Zoning save flags                     */
        #define SMPRG_ZSF_ENB       0x01    /* Saving zoning enabled supp   */
        #define SMPRG_ZSF_PERM_TBL  0x02    /* Saving permission table supp */
        #define SMPRG_ZSF_PHY_INFO  0x04    /* Saving PHY info supported    */
        #define SMPRG_ZSF_PASSWORD  0x08    /* Saving password supported    */
        #define SMPRG_ZSF_SAVING    0x10    /* Saving                       */

    WORD    NumSasAddr;         /* 26 Number of routed SAS addresses        */
    SASADDR ZnMgrSasAddr;       /* 28 Active zone manager SAS address       */
    WORD    ZnLockInactTo;      /* 30 Zone lock inactivity timeout (100ms)  */
    BYTE    Reserved3[2];       /* 32 Reserved                              */
    BYTE    PowerDoneTo;        /* 34 Power done timeout                    */
    BYTE    FirstElementIndex;  /* 35 1st enclosure connector element index */
    BYTE    NumElementIndexes;  /* 36 Number of element indexes             */
    BYTE    Reserved4;          /* 37 Reserved                              */
    BYTE    Flags;              /* 38 Flags                                 */
        #define SMPRGF_REDUCED_FUNC 0x80    /* Reduced functionality        */

    BYTE    RedFuncTime;        /* 39 Time to reduced functionality         */
    BYTE    InitRedFuncTime;    /* 3A Initial time to reduced functionality */
    BYTE    MaxRedFuncTime;     /* 3B Maximum reduced functionality time    */
    WORD    LastScsDescIndex;   /* 3C Last self-cfg status descriptor index */
    WORD    MaxScsDesc;         /* 3E Max self-cfg status descriptors       */
    WORD    LastPhyEvtDescIndex;/* 40 Last PHY event descriptor index       */
    WORD    MaxPhyEvtDesc;      /* 42 Max PHY event descriptors             */
    WORD    StpRjtToOpenLimit;  /* 44 STP reject to open limit              */
    BYTE    Reserved5[2];       /* 46 Reserved                              */
} SMPRSP_RPT_GEN,               /* 48 Size of structure                     */
*PSMPRSP_RPT_GEN;


/* Report Manufacturer */

typedef SMPREQ_HDR  SMPREQ_RPT_MFR,          *PSMPREQ_RPT_MFR;

typedef struct SmpRspRptMfr {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved[2];        /* 06 Reserved                              */
    BYTE    Flags;              /* 08 Flags                                 */
        #define SMPRMF_SAS_1_1      0x01    /* Conforms to SAS 1.1          */

    BYTE    Reserved2[3];       /* 09 Reserved                              */
    CHAR    VendorId[8];        /* 0C Vendor ID                             */
    CHAR    ProductId[16];      /* 14 Product ID                            */
    CHAR    Revision[4];        /* 24 Revision level                        */
    CHAR    ComponentVendor[8]; /* 28 Component Vendor ID                   */
    WORD    ComponentId;        /* 30 Component ID                          */
    BYTE    ComponentRev;       /* 32 Component Revision level              */
    BYTE    Reserved3;          /* 33 Reserved                              */
    BYTE    VendorSpecific[8];  /* 34 Vendor-specific info                  */
} SMPRSP_RPT_MFR,               /* 3C Size of structure                     */
*PSMPRSP_RPT_MFR;


/* Report Self Configuration Status */

typedef struct SmpReqRptSelfCfgSts {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved[2];        /* 04 Reserved                              */
    WORD    StartDescIndex;     /* 06 Starting descriptor index             */
} SMPREQ_RPT_SELF_CFG_STS,      /* 08 Size of structure                     */
*PSMPREQ_RPT_SELF_CFG_STS;

typedef struct SmpRspRptSelfCfgStsDesc {
    BYTE    StatusType;         /* 00 Status type                           */
    BYTE    Flags;              /* 01 Flags                                 */
        #define SMPRSCSF_FINAL      0x01    /* Info is final                */

    BYTE    Reserved;           /* 02 Reserved                              */
    BYTE    PhyId;              /* 03 PHY identifier                        */
    BYTE    Reserved2[4];       /* 04 Reserved                              */
    SASADDR SasAddr;            /* 08 SAS address                           */
} SMPRSP_RSCS_DESC,             /* 10 Size of structure                     */
*PSMPRSP_RSCS_DESC;

typedef struct SmpRspRptSelfCfgSts {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    WORD    StartDescIndex;     /* 06 Starting descriptor index             */
    WORD    NumDescriptors;     /* 08 Total number of decriptors            */
    WORD    LastDescIndex;      /* 0A Last descriptor index                 */
    BYTE    DescLength;         /* 0C Descriptor length                     */
    BYTE    Reserved2[6];       /* 0D Reserved                              */
    BYTE    NumDesc;            /* 13 Number of descriptors that follow     */

    SMPRSP_RSCS_DESC Desc[1];   /* 14 Variable list of descriptors          */

} SMPRSP_RPT_SELF_CFG_STS,      /* XX Size of structure                     */
*PSMPRSP_RPT_SELF_CFG_STS;


/* Report Zone Permission Table */

typedef struct SmpReqRptZnPermTable {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Flags;              /* 04 Flags                                 */
        /* See  SMP_ZM_RPT_TYPE_XXX                                         */

    BYTE    Reserved;           /* 05 Reserved                              */
    BYTE    StartZnGrp;         /* 06 Starting zone group                   */
    BYTE    MaxZnPermDesc;      /* 07 Maximum number of zone perm. desc.    */
} SMPREQ_RPT_ZN_PERM_TBL,       /* 08 Size of structure                     */
*PSMPREQ_RPT_ZN_PERM_TBL;


typedef struct SmpZnPermTableDesc128 {
    BYTE    ZnPerms[16];        /* 00 Zone permission mask                  */
} SMP_ZN_PMT_DESC_128,          /* 10 Size of structure                     */
*PSMP_ZN_PMT_DESC_128;


typedef struct SmpZnPermTableDesc256 {
    BYTE    ZnPerms[32];        /* 00 Zone permission mask                  */
} SMP_ZN_PMT_DESC_256,          /* 20 Size of structure                     */
*PSMP_ZN_PMT_DESC_256;


typedef struct SmpRspRptZnPermTable {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Flags;              /* 06 Flags                                 */
        /* See  SMP_ZM_RPT_TYPE_XXX                                         */
        #define SMPRZPT_ZF_LOCKED   0x80    /* Zone locked                  */

    BYTE    ZnFlags;            /* 07 Zoning configuration flags            */
        /* See  SMP_ZN_GRP_MAX_XXX                                          */

    BYTE    Reserved[5];        /* 08 Reserved                              */
    BYTE    DescLength;         /* 0D Zone permission descriptor length     */
    BYTE    StartZnGrp;         /* 0E Starting zone group                   */
    BYTE    NumDesc;            /* 0F Number of descriptors that follow     */

    union {                     /* 10 Variable list of descriptors          */
        SMP_ZN_PMT_DESC_128 Desc128[1];
        SMP_ZN_PMT_DESC_256 Desc256[1];
    };

} SMPRSP_RPT_ZN_PERM_TBL,       /* XX Size of structure                     */
*PSMPRSP_RPT_ZN_PERM_TBL;

#define SMP_RPT_ZN_PMT_MAX_DESC_128     63
#define SMP_RPT_ZN_PMT_MAX_DESC_256     31


/* Report Zone Manager Password */

typedef struct SmpReqRptZnMgrPassword {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Flags;              /* 04 Flags                                 */
        /* See  SMP_ZM_RPT_TYPE_XXX                                         */

    BYTE    Reserved[3];        /* 05 Reserved                              */
} SMPREQ_RPT_ZN_MGR_PW,         /* 08 Size of structure                     */
*PSMPREQ_RPT_ZN_MGR_PW;


typedef struct SmpRspRptZnMgrPassword {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Flags;              /* 06 Flags                                 */
        /* See  SMP_ZM_RPT_TYPE_XXX                                         */

    BYTE    Reserved;           /* 07 Reserved                              */
    BYTE    ZnMgrPw[32];        /* 08 Zone manager password                 */
} SMPRSP_RPT_ZN_MGR_PW,         /* 28 Size of structure                     */
*PSMPRSP_RPT_ZN_MGR_PW;


/* Report Broadcast */

typedef struct SmpReqRptBroadcast {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    BroadcastType;      /* 04 Broadcast type                        */
        /* See  SMP_BC_TYPE_XXX                                             */

    BYTE    Reserved[3];        /* 05 Reserved                              */
} SMPREQ_RPT_BROADCAST,         /* 08 Size of structure                     */
*PSMPREQ_RPT_BROADCAST;


typedef struct SmpRptBroadcastDesc {
    BYTE    BroadcastType;      /* 00 Broadcast type                        */
        /* See  SMP_BC_TYPE_XXX                                             */

    BYTE    PhyId;              /* 01 PHY identifier                        */
    BYTE    BroadcastRsn;       /* 02 Broadcast reason                      */
        #define SMPRBC_BC_RSN_MASK  0x0F    /* Broadcast reason mask        */

    BYTE    Reserved;           /* 03 Reserved                              */
    WORD    Bcc;                /* 04 Broadcast count                       */
    BYTE    Reserved2[10];      /* 06 Reserved                              */
} SMP_RPT_BROADCAST_DESC,       /* 10 Size of structure                     */
*PSMP_RPT_BROADCAST_DESC;


typedef struct SmpRspRptBroadcast {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    BroadcastType;      /* 06 Broadcast type                        */
        /* See  SMP_BC_TYPE_XXX                                             */

    BYTE    Reserved[3];        /* 07 Reserved                              */
    BYTE    DescLength;         /* 0A Broadcast descriptor length           */
    BYTE    NumDesc;            /* 0B Number of descriptors that follow     */

    SMP_RPT_BROADCAST_DESC
            Desc[1];            /* 0C Variable broadcast descriptor list    */

} SMPRSP_RPT_BROADCAST,         /* XX Size of structure                     */
*PSMPRSP_RPT_BROADCAST;
    

/* Discover */

typedef struct SmpReqDiscover {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved[4];        /* 04 Reserved                              */
    BYTE    Flags;              /* 08 Flags                                 */
        #define SMPDF_IGNORE_ZN_GRP 0x01    /* Ignore zone group            */

    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved2[2];       /* 0A Reserved                              */
} SMPREQ_DISC,                  /* 0C Size of structure                     */
*PSMPREQ_DISC;


typedef struct SmpRspDiscover {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved[3];        /* 06 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved2[2];       /* 0A Reserved                              */
    BYTE    AttachTypeReason;   /* 0C Attached device type and reason       */
        #define SMPD_ATR_REASON_MASK 0x0F   /* Attached reason              */
        #define SMPD_ATR_DEVTYP_MASK 0x70   /* Attached device type mask    */
        #define SMPD_ATR_DEVTYP_SHIFT   4   /* Attached device type shift   */

    BYTE    LinkRate;           /* 0D Negotiated link rate                  */
        #define SMPD_LR_MASK        0x0F    /* Negotiated rate mask         */
    
    BYTE    AttachIni;          /* 0E Attached initiator flags              */
        #define SMPD_AIF_SATA       0x01    /* Attached SATA initiator      */
        #define SMPD_AIF_SMP        0x02    /* Attached SMP  initiator      */
        #define SMPD_AIF_STP        0x04    /* Attached STP  initiator      */
        #define SMPD_AIF_SSP        0x08    /* Attached SSP  initiator      */
        #define SMPD_AIF_TRANS      0x0E    /* Transport Protos             */
        #define SMPD_AIF_ALL        0x0F    /* Attached initiator mask      */
        #define SMPD_AIF_ALL_SAS    0x0A    /* Attached SAS initiator mask  */
        #define SMPD_AIF_ALL_SATA   0x05    /* Attached SATA initiator mask */

    BYTE    AttachTgt;          /* 0F Attached target flags                 */
        #define SMPD_ATF_SATA       0x01    /* Attached SATA target         */
        #define SMPD_ATF_SMP        0x02    /* Attached SMP  target         */
        #define SMPD_ATF_STP        0x04    /* Attached STP  target         */
        #define SMPD_ATF_SSP        0x08    /* Attached SSP  target         */
        #define SMPD_ATF_STP_BUF_LO 0x10    /* STP buffer too small         */
        #define SMPD_ATF_TRANS      0x0E    /* Transport Protos             */
        #define SMPD_ATF_SATA_SEL   0x80    /* Attached SATA selector       */
        #define SMPD_ATF_ALL        0x8F    /* Attached target mask         */
        #define SMPD_ATF_ALL_SAS    0x0A    /* Attached SAS target mask     */
        #define SMPD_ATF_ALL_SATA   0x85    /* Attached SATA target mask    */

    SASADDR SasAddr;            /* 10 SAS Address                           */
    SASADDR AttachSasAddr;      /* 18 Attached to SAS Address               */
    BYTE    AttachPhyId;        /* 20 Attached PHY ID                       */
    BYTE    AttachFlags;        /* 21 Attached flags                        */
        #define SMPD_AF_BRK_REP_CAP     0x01    /* Break reply capable      */
        #define SMPD_AF_REQ_IN_ZPSDS    0x02    /* Requested inside ZPSDS   */
        #define SMPD_AF_IN_ZPSDS_PERS   0x04    /* Inside ZPSDS persistent  */
        #define SMPD_AF_PARTIAL         0x08    /* Partial capable          */
        #define SMPD_AF_SLUMBER         0x10    /* Slumber capable          */
        #define SMPD_AF_POWER_MASK      0x60    /* Power capabilities mask  */
        #define SMPD_AF_POWER_NONE      0x00    /* No power capabilites     */
        #define SMPD_AF_POWER_SOURCE    0x40    /* Grant additional power   */
        #define SMPD_AF_POWER_CONSUMER  0x20    /* Consume additional power */

    BYTE    AttachFlags2;       /* 22 Attached flags                        */
        #define SMPD_AF2_PWR_DIS_CAP    0x01    /* power disable capable    */
        #define SMPD_AF2_SMP_PRI_CAP    0x02    /* SMP priority capable     */
        #define SMPD_AF2_APTA_CAP       0x04    /* APTA capable             */

    BYTE    Reserved3[5];       /* 23 Reserved                              */
    BYTE    MinRate;            /* 28 Programmed/hardware minimum rate      */
        #define SMPD_MINR_HW        0x0F    /* Hardware   rate mask         */
        #define SMPD_MINR_PROG      0xF0    /* Programmed rate mask         */
        #define SMPD_MINR_PROG_SH   4       /* Programmed rate shift        */

    BYTE    MaxRate;            /* 29 Programmed/hardware maximum rate      */
        #define SMPD_MAXR_HW        0x0F    /* Hardware   rate mask         */
        #define SMPD_MAXR_PROG      0xF0    /* Programmed rate mask         */
        #define SMPD_MAXR_PROG_SH   4       /* Programmed rate shift        */

    BYTE    PhyChangeCnt;       /* 2A Num Broadcast (change)s initiated     */
    BYTE    PartPathTov;        /* 2B Partial path timeout, virt. PHY flag  */
        #define SMPD_PPTOV_MASK     0x0F    /* Partial Pathway Timeout      */
        #define SMPD_PPTOV_VPHY     0x80    /* Virtual PHY flag             */

    BYTE    RouteAttr;          /* 2C Routing attribute                     */
        #define SMPD_RATTR_MASK     0x0F    /* Route attribute mask         */

    BYTE    ConnType;           /* 2D Connector Type                        */
        #define SMPD_CT_MASK        0x7F    /* Connector type mask          */

    BYTE    ConnElement;        /* 2E Connector Element Index               */
    BYTE    ConnPhysLink;       /* 2F Connector Physical Link               */
    BYTE    PowerCaps;          /* 30 Power capabilities                    */
        #define SMPD_PC_SATA_PART   0x01    /* SATA partial                 */
        #define SMPD_PC_SATA_SLUMBR 0x02    /* SATA slumber                 */
        #define SMPD_PC_SAS_PART    0x04    /* SAS partial                  */
        #define SMPD_PC_SAS_SLUMBR  0x08    /* SAS slumber                  */
        #define SMPD_PC_POWER_MASK  0x30    /* SAS power capabilities       */
        #define SMPD_PC_POWER_NONE  0x00    /* No power capabilites         */
        #define SMPD_PC_POWER_CONS  0x20    /* Consume additional power     */
        #define SMPD_PC_COND_MASK   0xC0    /* Power power condition        */
        #define SMPD_PC_COND_ACT    0x00    /* Active power condition       */
        #define SMPD_PC_COND_PART   0x40    /* Partial power condition      */
        #define SMPD_PC_COND_SLUMBR 0x80    /* Slumber power condition      */

    BYTE    PowerState;         /* 31 Power state                           */
        #define SMPD_PS_SATA_PART   0x01    /* SATA partial                 */
        #define SMPD_PS_SATA_SLUMBR 0x02    /* SATA slumber                 */
        #define SMPD_PS_SAS_PART    0x04    /* SAS partial                  */
        #define SMPD_PS_SAS_SLUMBR  0x08    /* SAS slumber                  */

    BYTE    VendorSpecific[2];  /* 32 Vendor specific                       */
                                /*    End of SAS 1.1 fields                 */
    BYTE    AttachDevName[8];   /* 34 Attached device name                  */
    BYTE    ZnPhyFlags;         /* 3C Zoning PHY flags (SMP_ZN_PHYF_XXX)    */
    BYTE    ZnReserved[2];      /* 3D Reserved                              */
    BYTE    ZnGrp;              /* 3F Zone group                            */
    BYTE    SelfCfgSts;         /* 40 Self configuration status             */
    BYTE    SelfCfgComp;        /* 41 Self configuration levels completed   */
    BYTE    Reserved6[2];       /* 42 Reserved                              */
    BYTE    SelfCfgSasAddr[8];  /* 44 Self configuration SAS address        */
    DWORD   ProgPhyCaps;        /* 4C Programmed PHY capabilities           */
    DWORD   CurrPhyCaps;        /* 50 Current PHY capabilities              */
    DWORD   AttachPhyCaps;      /* 54 Attached PHY capabilities             */
    BYTE    Reserved7[6];       /* 58 Reserved                              */
    BYTE    PhysLinkRate;       /* 5E Negotiated phys link rate and reason  */
        #define SMPD_PLR_MASK       0x0F    /* Negotiated rate mask         */
        #define SMPD_PLR_RSN_MASK   0xF0    /* Reason mask                  */

    BYTE    Flags;              /* 5F Flags                                 */
        #define SMPDF_HW_MUX_SUPP   0x01    /* Hardware muxing supported    */
        #define SMPDF_NEG_SSC       0x02    /* Negotiated SSC               */
        #define SMPDF_OPT_MODE      0x04    /* Optical mode enabled         */

    BYTE    DefZnPhyFlags;      /* 60 Default zoning PHY flags              */
                                /*    (SMP_ZN_PHYF_XXX)                     */
    BYTE    DefZnReserved[2];   /* 62 Reserved                              */
    BYTE    DefZnGrp;           /* 63 Default zone group                    */
    BYTE    SvZnPhyFlags;       /* 64 Saved zoning PHY flags                */
                                /*    (SMP_ZN_PHYF_XXX)                     */
    BYTE    SzZnReserved[2];    /* 65 Reserved                              */
    BYTE    SvZnGrp;            /* 67 Saved zone group                      */
    BYTE    ShdwZnPhyFlags;     /* 68 Shadow zoning PHY flags               */
                                /*    (SMP_ZN_PHYF_XXX)                     */
    BYTE    ShdwZnReserved[2];  /* 69 Reserved                              */
    BYTE    ShdwZnGrp;          /* 6B Shadow zone group                     */
    BYTE    DevSlotNum;         /* 6C Device slot number                    */
    BYTE    DevSlotGrpNum;      /* 6D Device slot group number              */
    CHAR    DevSlotGrpConn[6];  /* 6E Device slot group output connector    */
    WORD    StpBufSize;         /* 74 STP buffer size                       */
    BYTE    PhyBurstSize;       /* 76 Buffered PHY burst size (KB)          */
    BYTE    Reserved11;         /* 77 Reserved                              */
} SMPRSP_DISC,                  /* 78 Size of structure                     */
*PSMPRSP_DISC;


/* Report PHY Error Log */

typedef struct SmpReqRptPhyErrLog {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[5];       /* 04 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved3[2];       /* 0A Reserved                              */
} SMPREQ_RPT_PHY_ERR_LOG,       /* 0C Size of structure                     */
*PSMPREQ_RPT_PHY_ERR_LOG;


typedef struct SmpRspRptPhyErrLog {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved[3];        /* 06 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved2[2];       /* 0A Reserved                              */
    DWORD   InvalDwords;        /* 0C Invalid DWORD count                   */
    DWORD   DisparityErrCnt;    /* 10 Running Disparity error count         */
    DWORD   LossOfSyncCnt;      /* 14 Loss of DWORD sync count              */
    DWORD   ResetProblemCnt;    /* 18 PHY Reset Problem count               */
} SMPRSP_RPT_PHY_ERR_LOG,       /* 1C Size of structure                     */
*PSMPRSP_RPT_PHY_ERR_LOG;


/* Report PHY SATA */

typedef struct SmpReqRptPhySata {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[5];       /* 04 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    AffilContext;       /* 0A Affiliation context                   */
    BYTE    Reserved3;          /* 0B Reserved                              */
} SMPREQ_RPT_PHY_SATA,          /* 0C Size of structure                     */
*PSMPREQ_RPT_PHY_SATA;


typedef struct SmpRspRptPhySata {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved[3];        /* 06 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved2;          /* 0A Reserved                              */
    BYTE    Flags;              /* 0B Flags                                 */
        /* TF:!! SMP Report PHY SATA flags */
        #define SMPRPSF_AFF_VAL     0x01    /* Affiliation valid            */
        #define SMPRPSF_AFF_SUPP    0x02    /* Affiliations supported       */
        #define SMPRPSF_STP_NX_LOSS 0x04    /* STP I_T nexus loss occurred  */
        /* Tf:!! SMP Report PHY SATA flags */

    BYTE    Reserved3[4];       /* 0C Reserved                              */
    SASADDR StpSasAddr;         /* 10 SAS Address of STP target port        */
    BYTE    RegDevToHostFIS[20];/* 18 Register Device To Host FIS           */
    BYTE    Reserved4[4];       /* 2C Reserved                              */
    SASADDR AffilInitiator;     /* 30 Affiliated STP Initiator SAS Address  */
                                /*    End of SAS 1.1 fields                 */
    SASADDR StpNexLossSasAddr;  /* 38 STP I_T nexus loss SAS Address        */
    BYTE    Reserved5;          /* 40 Reserved                              */
    BYTE    AffilContext;       /* 41 Affiliation context                   */
    BYTE    CurrAffilContexts;  /* 42 Current affiliation contexts          */
    BYTE    MaxAffilContexts;   /* 43 Maximum affiliation contexts          */

} SMPRSP_RPT_PHY_SATA,          /* 44 Size of structure                     */
*PSMPRSP_RPT_PHY_SATA;


/* Report Route Information */

typedef struct SmpReqRptRouteInfo {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[2];       /* 04 Reserved                              */
    WORD    ExpRouteIndex;      /* 06 Expander Route Index                  */
    BYTE    Reserved3;          /* 08 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved4[2];       /* 0A Reserved                              */
} SMPREQ_RPT_ROUTE_INFO,        /* 0C Size of structure                     */
*PSMPREQ_RPT_ROUTE_INFO;


typedef struct SmpRspRptRouteInfo {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    WORD    ExpRouteIndex;      /* 06 Expander Route Index                  */
    BYTE    Reserved;           /* 08 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved2[2];       /* 0A Reserved                              */
    BYTE    Flags;              /* 0C Flags                                 */
        #define SMPRRIF_DISABLED    0x80    /* Route entry disabled         */

    BYTE    Reserved3[3];       /* 0D Reserved                              */
    SASADDR RoutedSasAddr;      /* 10 Routed SAS Address                    */
    BYTE    Reserved4[16];      /* 18 Reserved                              */
} SMPRSP_RPT_ROUTE_INFO,        /* 28 Size of structure                     */
*PSMPRSP_RPT_ROUTE_INFO;


/* Report PHY Event */

typedef struct SmpReqRptPhyEvent {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[5];       /* 04 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved3[2];       /* 0A Reserved                              */
} SMPREQ_RPT_PHY_EVENT,         /* 0C Size of structure                     */
*PSMPREQ_RPT_PHY_EVENT;


typedef struct SmpPhyEventDesc {
    BYTE    Reserved[3];        /* 00 Reserved                              */
    BYTE    EventSource;        /* 03 PHY event source                      */
    DWORD   EventInfo;          /* 04 Event info                            */
    DWORD   DetThreshold;       /* 08 Peak value detector threshold         */
} SMP_PHY_EVENT_DESC,           /* 0C Size of structure                     */
*PSMP_PHY_EVENT_DESC;


typedef struct SmpRspRptPhyEvent {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved[3];        /* 06 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved2[4];       /* 0A Reserved                              */
    BYTE    DescLength;         /* 0E Descriptor length                     */
    BYTE    NumDesc;            /* 0F Number of event descriptors           */

    SMP_PHY_EVENT_DESC Desc[1]; /* 10 Variable number of event descriptors  */

} SMPRSP_RPT_PHY_EVENT,         /* XX Size of structure                     */
*PSMPRSP_RPT_PHY_EVENT;


/* Discover List */

typedef struct SmpReqDiscList {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[4];       /* 04 Reserved                              */
    BYTE    StartPhyId;         /* 08 Starting PHY identifier               */
    BYTE    MaxDesc;            /* 09 Maximum number of descriptors         */
    BYTE    PhyFilter;          /* 0A PHY filter                            */
        #define SMPDL_PF_MASK       0x0F    /* PHY Filter mask              */
        #define SMPDL_PF_ALL        0x00    /* Return all PHYs              */
        #define SMPDL_PF_EXP        0x01    /* Return PHYs attached to exp. */
        #define SMPDL_PF_ATT        0x02    /* Return attached PHYs         */
        #define SMPDL_PF_IGN_ZN_GRP 0x80    /* Ignore zone group flag       */

    BYTE    DescType;           /* 0B Descriptor type                       */
        #define SMPDL_DT_MASK       0x0F    /* Descriptor type              */
        #define SMPDL_DT_DISC       0x00    /* Format of Discover command   */
        #define SMPDL_DT_SHORT      0x01    /* Short format                 */

    BYTE    Reserved3[4];       /* 0C Reserved                              */
    BYTE    VendorSpecific[12]; /* 10 Vendor specific                       */

} SMPREQ_DISC_LIST,             /* 1C Size of structure                     */
*PSMPREQ_DISC_LIST;


typedef struct SmpRspDiscListShortDesc {
    BYTE    PhyId;              /* 00 PHY identifier                        */
    BYTE    Result;             /* 01 Function result                       */
    BYTE    AttachTypeReason;   /* 02 Attached device type and reason       */
        #define SMPDLD_ATR_RSN_MASK 0x0F  /* Attached reason                */
        #define SMPDLD_ATR_DT_MASK  0x70  /* Attached device type           */

    BYTE    LinkRate;           /* 03 Negotiated link rate                  */
        #define SMPDLD_LR_MASK      0x0F    /* Negotiated rate mask         */

    BYTE    AttachIni;          /* 04 Attached initiator flags              */
        #define SMPDLD_AIF_SATA     0x01    /* Attached SATA initiator      */
        #define SMPDLD_AIF_SMP      0x02    /* Attached SMP  initiator      */
        #define SMPDLD_AIF_STP      0x04    /* Attached STP  initiator      */
        #define SMPDLD_AIF_SSP      0x08    /* Attached SSP  initiator      */

    BYTE    AttachTgt;          /* 05 Attached target flags                 */
        #define SMPDLD_ATF_SATA       0x01  /* Attached SATA target         */
        #define SMPDLD_ATF_SMP        0x02  /* Attached SMP  target         */
        #define SMPDLD_ATF_STP        0x04  /* Attached STP  target         */
        #define SMPDLD_ATF_SSP        0x08  /* Attached SSP  target         */
        #define SMPDLD_ATF_STP_BUF_LO 0x10  /* STP buffer too small         */
        #define SMPDLD_ATF_SATA_SEL   0x80  /* Attached SATA selector       */

    BYTE    RouteAttr;          /* 06 Routing attribute                     */
        #define SMPDLD_RATTR_MASK   0x0F    /* Route attribute mask         */
        #define SMPDLD_RATTR_VPHY   0x80    /* Virtual PHY flag             */

    BYTE    PhysLinkRate;       /* 5E Negotiated phys link rate and reason  */
        #define SMPD_PLR_MASK       0x0F    /* Negotiated rate mask         */
        #define SMPD_PLR_RSN_MASK   0xF0    /* Reason mask                  */

    BYTE    ZnGrp;              /* 08 Zone group                            */
    BYTE    ZnPhyFlags;         /* 09 Zoning PHY flags                      */
        /* See  SMP_ZN_PHYF_XXX                                             */

    BYTE    AttachPhyId;        /* 0A Attached PHY ID                       */
    BYTE    PhyChangeCnt;       /* 0B Num Broadcast (change)s initiated     */
    BYTE    AttachSasAddr[8];   /* 0C Attached to SAS Address               */
    BYTE    Reserved[4];        /* 14 Reserved                              */
} SMPRSP_DL_SHORT_DESC,         /* 18 Size of structure                     */ 
*PSMPRSP_DL_SHORT_DESC;


typedef struct SmpRspDiscList {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved[2];        /* 06 Reserved                              */
    BYTE    StartPhyId;         /* 08 Starting PHY identifier               */
    BYTE    NumDesc;            /* 09 Number of descriptors                 */
    BYTE    PhyFilter;          /* 0A PHY filter                            */
    BYTE    DescType;           /* 0B Descriptor type                       */
    BYTE    DescLength;         /* 0C Descriptor length                     */
    BYTE    Reserved2[3];       /* 0D Reserved                              */
    BYTE    Flags;              /* 10 Flags                                 */
        #define SMPDLF_CFG_RTE_TBL  0x01    /* Configurable route table     */
        #define SMPDLF_CONFIGURING  0x02    /* Configuring                  */
        #define SMPDLF_ZN_CFG       0x04    /* Zone configuring             */
        #define SMPDLF_SELF_CFG     0x08    /* Self configuring             */
        #define SMPDLF_ZN_ENB       0x40    /* Zoning enabled               */
        #define SMPDLF_ZN_SUPP      0x80    /* Zoning supported             */

    BYTE    Reserved3;          /* 11 Reserved                              */
    WORD    LastScsDescIndex;   /* 12 Last Self-Cfg Status descriptor index */
    WORD    LastPhyEvtDescIndex;/* 14 Last PHY event list descriptor index  */
    BYTE    Reserved4[10];      /* 11 Reserved                              */
    BYTE    VendorSpecific[16]; /* 20 Vendor specific                       */

    union {                     /* 30 Variable descriptor list              */
        SMPRSP_DISC          DiscDesc[1];
        SMPRSP_DL_SHORT_DESC ShortDesc[1];
    };

} SMPRSP_DISC_LIST,             /* XX Size of structure                     */
*PSMPRSP_DISC_LIST;


/* Report PHY Event List */

typedef struct SmpReqRptPhyEventList {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[2];       /* 04 Reserved                              */
    WORD    StartDescIndex;     /* 06 Starting descriptor index             */
} SMPREQ_RPT_PHY_EVENT_LIST,    /* 08 Size of structure                     */
*PSMPREQ_RPT_PHY_EVENT_LIST;


typedef struct SmpPhyEventListDesc {
    BYTE    Reserved[2];        /* 00 Reserved                              */
    BYTE    PhyId;              /* 02 PHY identifier                        */
    BYTE    EventSource;        /* 03 PHY event source                      */
    DWORD   EventInfo;          /* 04 Event info                            */
    DWORD   DetThreshold;       /* 08 Peak value detector threshold         */
} SMP_PHY_EVENT_LIST_DESC,      /* 0C Size of structure                     */
*PSMP_PHY_EVENT_LIST_DESC;


typedef struct SmpRspRptPhyEventList {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    WORD    StartDescIndex;     /* 06 Starting descriptor index             */
    WORD    LastDescIndex;      /* 08 Last descriptor index                 */
    BYTE    DescLength;         /* 0A Descriptor length                     */
    BYTE    Reserved[4];        /* 0B Reserved                              */
    BYTE    NumDesc;            /* 0F Number of event descriptors           */

    SMP_PHY_EVENT_LIST_DESC
            Desc[1];            /* 10 Variable number of event descriptors  */

} SMPRSP_RPT_PHY_EVENT_LIST,    /* XX Size of structure                     */
*PSMPRSP_RPT_PHY_EVENT_LIST;


/* Report Expander Route Table List */

typedef struct SmpReqRptExpRteTbl {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[4];       /* 04 Reserved                              */
    WORD    MaxDesc;            /* 08 Maximum number of descriptors         */
    WORD    StartSasAddrIx;     /* 0A Starting SAS address index            */
    BYTE    Reserved3[7];       /* 0C Reserved                              */
    BYTE    StartPhyId;         /* 13 Starting PHY identifier               */
    BYTE    Reserved4[8];       /* 14 Reserved                              */
} SMPREQ_RPT_EXP_RTE_TBL,       /* 1C Size of structure                     */
*PSMPREQ_RPT_EXP_RTE_TBL;


typedef struct SmpRspRptExpRteTblDesc {
    SASADDR SasAddr;            /* 00 Routed SAS Address                    */
    BYTE    PhyBitMap[6];       /* 08 PHY bit map                           */
    BYTE    Reserved;           /* 0E Reserved                              */
    BYTE    ZnGrp;              /* 0F Zone group                            */
} SMPRSP_RERT_DESC,             /* 10 Size of structure                     */
*PSMPRSP_RERT_DESC;


typedef struct SmpRspRptExpRteTbl {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    WORD    ExpRtChangeCnt;     /* 06 Expander route table change count     */
    BYTE    Flags;              /* 08 Flags                                 */
        #define SMPRERTF_ZN_ENB     0x01    /* Zoning enabled               */
        #define SMPRERTF_CONFIG     0x02    /* Configuring                  */
        #define SMPRERTF_ZN_CFG     0x04    /* Zone configuring             */
        #define SMPRERTF_SELF_CFG   0x08    /* Self configuring             */

    BYTE    Reserved;           /* 09 Reserved                              */
    BYTE    DescLength;         /* 0A Descriptor length                     */
    BYTE    NumDesc;            /* 0B Number of route table descriptors     */
    WORD    FirstSasAddrIx;     /* 0C First SAS address index               */
    WORD    LastSasAddrIx;      /* 0E Last SAS address index                */
    BYTE    Reserved3[3];       /* 10 Reserved                              */
    BYTE    StartPhyId;         /* 13 Starting PHY identifier               */
    BYTE    Reserved2[12];      /* 14 Reserved                              */

    SMPRSP_RERT_DESC Desc[1];   /* 20 Variable list of descriptors          */

} SMPRSP_RPT_EXP_RTE_TBL,       /* XX Size of structure                     */
*PSMPRSP_RPT_EXP_RTE_TBL;


/* Configure General */

typedef struct SmpReqCfgGeneral {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved2[2];       /* 06 Reserved                              */
    BYTE    Flags;              /* 08 Flags                                 */
        #define SMPCGF_ENB_SBIT     0x01    /* Set StpBusInactTo            */
        #define SMPCGF_ENB_SMCT     0x02    /* Set StpMaxConnTo             */
        #define SMPCGF_ENB_SINLT    0x04    /* Set StpItNexusLossTime       */
        #define SMPCGF_ENB_IRFT     0x08    /* Set InitRedFuncTime          */
        #define SMPCGF_ENB_SRTOL    0x10    /* Set StpRjtToOpenLimit        */
        #define SMPCGF_ENB_PDT      0x20    /* Set PowerDoneTo              */

    BYTE    Reserved3;          /* 09 Reserved                              */
    WORD    StpBusInactTo;      /* 0A STP bus inactivity time limit         */
    WORD    StpMaxConnTo;       /* 0C STP maximum connection time limit     */
    WORD    StpItNexusLossTime; /* 0E STP I_T nexus loss time               */
    BYTE    InitRedFuncTime;    /* 10 Initial time to reduced functionality */
    BYTE    PowerDoneTo;        /* 11 Power done timeout                    */
    WORD    StpRjtToOpenLimit;  /* 12 STP reject to open limit              */

} SMPREQ_CFG_GEN,               /* 14 Size of structure                     */
*PSMPREQ_CFG_GEN;


typedef SMPRSP_HDR  SMPRSP_CFG_GEN,          *PSMPRSP_CFG_GEN;


/* Enable/Disable Zoning */

typedef struct SmpReqEnbDisZoning {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    CfgFlags;           /* 06 Configuration flags                   */
        /* See  SMP_ZN_SAVE_XXX                                             */

    BYTE    Reserved2;          /* 07 Reserved                              */
    BYTE    Flags;              /* 08 Flags                                 */
        #define SMPEDZF_ENB_MASK    0x03    /* Enable/disable mask          */
        #define SMPEDZF_NO_CHG      0x00    /* No change                    */
        #define SMPEDZF_ENB         0x01    /* Enable zoning                */
        #define SMPEDZF_DIS         0x02    /* Disable zoning               */

    BYTE    Reserved3[3];       /* 09 Reserved                              */
} SMPREQ_ENB_DIS_ZN,            /* 0C Size of structure                     */
*PSMPREQ_ENB_DIS_ZN;


typedef SMPRSP_HDR  SMPRSP_ENB_DIS_ZN,          *PSMPRSP_ENB_DIS_ZN;


/* Zoned Broadcast */

typedef struct SmpReqZnBroadcast {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[2];       /* 04 Reserved                              */
    BYTE    BroadcastType;      /* 06 Broadcast type                        */
        /* See  SMP_BC_TYPE_XXX                                             */

    BYTE    NumGrps;            /* 07 Reserved                              */
    BYTE    ZnGrps[1];          /* 08 Variable list of source zone groups   */

} SMPREQ_ZN_BROADCAST,          /* XX Size of structure                     */
*PSMPREQ_ZN_BROADCAST;


typedef SMPRSP_HDR  SMPRSP_ZN_BROADCAST,          *PSMPRSP_ZN_BROADCAST;


/* Zone Lock */

typedef struct SmpReqZnLock {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    WORD    ZnLockInactTo;      /* 06 Zone lock inactivity timeout (100ms)  */
    BYTE    ZnMgrPw[32];        /* 08 Zone manager password                 */
} SMPREQ_ZN_LOCK,               /* 28 Size of structure                     */
*PSMPREQ_ZN_LOCK;


typedef struct SmpRspZnLock {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved[4];        /* 04 Reserved                              */
    SASADDR ZnMgrSasAddr;       /* 08 Zone manager SAS address              */
} SMPRSP_ZN_LOCK,               /* 10 Size of structure                     */
*PSMPRSP_ZN_LOCK;


/* Zone Activate */

typedef struct SmpReqZnActivate {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved2[2];       /* 06 Reserved                              */
} SMPREQ_ZN_ACTIVATE,           /* 08 Size of structure                     */
*PSMPREQ_ZN_ACTIVATE;


typedef SMPRSP_HDR  SMPRSP_ZN_ACTIVATE,          *PSMPRSP_ZN_ACTIVATE;


/* Zone Unlock */

typedef struct SmpReqZnUnlock {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    BYTE    Reserved2[2];       /* 04 Reserved                              */
    BYTE    Flags;              /* 06 Flags                                 */
        #define SMPZU_ACT_REQD      0x01    /* Activate required            */

    BYTE    Reserved3;          /* 07 Reserved                              */
} SMPREQ_ZN_UNLOCK,             /* 08 Size of structure                     */
*PSMPREQ_ZN_UNLOCK;


typedef SMPRSP_HDR  SMPRSP_ZN_UNLOCK,          *PSMPRSP_ZN_UNLOCK;


/* Configure Zone Manager Password */

typedef struct SmpReqCfgZnMgrPassword {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    CfgFlags;           /* 06 Configuration flags                   */
        /* See  SMP_ZN_SAVE_XXX                                             */

    BYTE    Reserved2;          /* 07 Reserved                              */
    BYTE    ZnMgrPw[32];        /* 08 Current zone manager password         */
    BYTE    NewZnMgrPw[32];     /* 28 New zone manager password             */
} SMPREQ_CFG_ZN_PASSWORD,       /* 48 Size of structure                     */
*PSMPREQ_CFG_ZN_PASSWORD;


typedef SMPRSP_HDR  SMPRSP_CFG_ZN_PASSWORD,
                    *PSMPRSP_CFG_ZN_PASSWORD;


/* Configure Zone PHY Information */

typedef struct SmpReqCfgZnPhyInfoDesc {
    BYTE    PhyId;              /* 00 PHY identifier                        */
    BYTE    ZnPhyFlags;         /* 01 Zoning PHY flags (SMP_ZN_PHYF_XXX)    */
    BYTE    Reserved;           /* 02 Reserved                              */
    BYTE    ZnGrp;              /* 03 Zone group                            */
} SMPREQ_CZPI_DESC,             /* 04 Size of structure                     */
*PSMPREQ_CZPI_DESC;


typedef struct SmpReqCfgZnPhyInfo {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    CfgFlags;           /* 06 Configuration flags                   */
        /* See  SMP_ZN_SAVE_XXX                                             */
        #define SMPCZPI_CF_DSCL_MASK 0xFC   /* Descriptor length mask       */
        #define SMPCZPI_CF_DSCL_SHIFT   2   /* Descriptor length shift      */

    BYTE    NumDesc;            /* 07 Number of following descriptors       */

    SMPREQ_CZPI_DESC Desc[1];   /* 08 Variable list of descriptors          */

} SMPREQ_CFG_ZN_PHY_INFO,       /* XX Size of structure                     */
*PSMPREQ_CFG_ZN_PHY_INFO;


typedef SMPRSP_HDR  SMPRSP_CFG_ZN_PHY_INFO,
                    *PSMPRSP_CFG_ZN_PHY_INFO;



/* Configure Zone Permission Table */

typedef struct SmpReqCfgZnPermTable {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    StartZnGrp;         /* 06 Starting zone group                   */
    BYTE    NumDesc;            /* 07 Number of following descriptors       */
    BYTE    CfgFlags;           /* 06 Configuration flags                   */
        /* See  SMP_ZN_SAVE_XXX                                             */
        /* See  SMP_ZN_GRP_MAX_XXX                                          */

    BYTE    DescLength;         /* 09 Descriptor length                     */
    BYTE    Reserved2[6];       /* 0A Reserved                              */

    union {                     /* 10 Variable list of descriptors          */
        SMP_ZN_PMT_DESC_128 Desc128[1];
        SMP_ZN_PMT_DESC_256 Desc256[1];
    };

} SMPREQ_CFG_ZN_PERM_TBL,       /* XX Size of structure                     */
*PSMPREQ_CFG_ZN_PERM_TBL;


typedef SMPRSP_HDR  SMPRSP_CFG_ZN_PERM_TBL,
                    *PSMPRSP_CFG_ZN_PERM_TBL;


/* Configure Route Information */

typedef struct SmpReqCfgRouteInfo {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    WORD    ExpRouteIndex;      /* 06 Expander Route Index                  */
    BYTE    Reserved2;          /* 08 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    Reserved3[2];       /* 0A Reserved                              */
    BYTE    Flags;              /* 0C Flags                                 */
        #define SMPCRIF_DISABLE     0x80    /* Disable route entry          */

    BYTE    Reserved4[3];       /* 0D Reserved                              */
    SASADDR RoutedSasAddr;      /* 10 Routed SAS Address                    */
    BYTE    Reserved5[16];      /* 18 Reserved                              */
} SMPREQ_CFG_ROUTE_INFO,        /* 28 Size of structure                     */
*PSMPREQ_CFG_ROUTE_INFO;


typedef SMPRSP_HDR  SMPRSP_CFG_ROUTE_INFO,
                    *PSMPRSP_CFG_ROUTE_INFO;


/* PHY Control */

typedef struct SmpReqPhyControl {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved2[3];       /* 06 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    PhyOp;              /* 0A PHY operation                         */
        #define SMPPC_PO_NOP        0x00    /* No operation                 */
        #define SMPPC_PO_LNKRST     0x01    /* Link Reset                   */
        #define SMPPC_PO_HARDRST    0x02    /* Hard Reset                   */
        #define SMPPC_PO_DISABLE    0x03    /* Disable PHY                  */
        #define SMPPC_PO_CLR_ERR    0x05    /* Clear error log              */
        #define SMPPC_PO_CLR_AFF    0x06    /* Clear affiliation            */
        #define SMPPC_PO_PORT_SEL   0x07    /* Xmit SATA port sel signal    */
        #define SMPPC_PO_CLR_SNL    0x08    /* Clear STP nexus loss         */
        #define SMPPC_PO_SET_ADN    0x09    /* Set attached device name     */

    BYTE    Flags;              /* 0B Operation flags                       */
        #define SMPPFF_UPPTOV       0x01    /* Update partial path timeout  */

    BYTE    Reserved3[12];      /* 0C Reserved                              */
    BYTE    AttachDevName[8];   /* 18 Attached device name                  */
    BYTE    MinProgRate;        /* 20 Programmed minimum physical rate      */
        #define SMPPC_MINPR_MASK    0xF0    /* Programmed minimum rate mask */
        #define SMPPC_MINPR_SHIFT   4       /* Programmed minimum rate shift*/

    BYTE    MaxProgRate;        /* 21 Programmed maximum physical rate      */
        #define SMPPC_MAXPR_MASK    0xF0    /* Programmed maximum rate mask */
        #define SMPPC_MAXPR_SHIFT   4       /* Programmed maximum rate shift*/

    BYTE    PowerCfg;           /* 22 Power configuration                   */
        #define SMPPC_PC_SATA_PART   0x03   /* SATA partial                 */
        #define SMPPC_PC_SATA_SLUMBR 0x0C   /* SATA slumber                 */
        #define SMPPC_PC_SAS_PART    0x30   /* SAS partial                  */
        #define SMPPC_PC_SAS_SLUMBR  0xC0   /* SAS slumber                  */

    BYTE    Reserved4;          /* 23 Reserved                              */
    BYTE    PartPathTov;        /* 24 Partial pathway timeout value         */
        #define SMPPC_PPTOV_MASK    0x0F    /* Partial Pathway Timeout      */

    BYTE    Reserved5[3];       /* 25 Reserved                              */
} SMPREQ_PHY_CTRL,              /* 28 Size of structure                     */
*PSMPREQ_PHY_CTRL;


typedef SMPRSP_HDR  SMPRSP_PHY_CTRL,          *PSMPRSP_PHY_CTRL;


/* PHY Test Function */

typedef struct SmpReqPhyTestFunction {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Reserved2[3];       /* 06 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    TestFunc;           /* 0A Test function                         */
        #define SMPPT_TF_STOP       0x00    /* Stop test function           */
        #define SMPPT_TF_XMIT_PAT   0x01    /* Transmit pattern             */

    BYTE    TestPattern;        /* 0B Test pattern                          */
    BYTE    Reserved3[3];       /* 0C Reserved                              */
    BYTE    TestRate;           /* 0F Test pattern link rate and flags      */
        #define SMPPT_TR_RATE_MASK  0x0F    /* Rate mask                    */
        #define SMPPT_TR_SSC_MASK   0x30    /* SSC mask                     */
        #define SMPPT_TR_SATA       0x40    /* PHY test pattern SATA        */

    BYTE    Reserved4[3];       /* 10 Reserved                              */
    BYTE    TestDwCtrl;         /* 13 Test pattern DWORDs control           */
    BYTE    TestDwPattern[8];   /* 14 Test pattern DWORDs                   */
    BYTE    Reserved5[12];      /* 1C Reserved                              */
} SMPREQ_PHY_TEST,              /* 28 Size of structure                     */
*PSMPREQ_PHY_TEST;


typedef SMPRSP_HDR  SMPRSP_PHY_TEST,          *PSMPRSP_PHY_TEST;


/* Configure PHY Event */

typedef struct SmpReqCfgPhyEvent {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RspLength;          /* 02 Number of response frame DWORDs w/o   */
                                /*    CRC or header.                        */
    BYTE    Length;             /* 03 Number of following DWORDs w/o CRC    */
    WORD    ExpChangeCnt;       /* 04 Expander change count                 */
    BYTE    Flags;              /* 05 Flags                                 */
        #define SMPCPEF_CLR_PEAKS   0x01    /* Clear peaks                  */

    BYTE    Reserved[2];        /* 06 Reserved                              */
    BYTE    PhyId;              /* 09 PHY identifier                        */
    BYTE    DescLength;         /* 0A Descriptor length                     */
    BYTE    NumDesc;            /* 0B Number of event descriptors           */

    SMP_PHY_EVENT_DESC Desc[1]; /* 0C Variable number of event descriptors  */

} SMPREQ_CFG_PHY_EVENT,         /* XX Size of structure                     */
*PSMPREQ_CFG_PHY_EVENT;


typedef SMPRSP_HDR  SMPRSP_CFG_PHY_EVENT,
                    *PSMPRSP_CFG_PHY_EVENT;


/****************************************************************************
                        Read/Write GPIO (from SFF-8485)
*****************************************************************************/

#define SMP_GPIO_TYPE_CFG           0x00    /* Configuration registers      */
#define SMP_GPIO_TYPE_RX            0x01    /* Receive registers            */
#define SMP_GPIO_TYPE_RX_GP         0x02    /* General purpose receive regs */
#define SMP_GPIO_TYPE_TX            0x03    /* Transmit registers           */
#define SMP_GPIO_TYPE_TX_GP         0x04    /* General purpose transmit regs*/

/* Configuration registers */

#define SMP_GPIO_REG_CFG0           0x00    /* Configuration register 0     */
#define SMP_GPIO_REG_CFG1           0x01    /* Configuration register 1     */

typedef struct SmpGpioRegCfg0 {
    BYTE    Reserved;           /* 00 Reserved                              */
    BYTE    Version;            /* 01 Version                               */
        #define SMPGRC0_VER_MASK        0x0F    /* Version mask             */

    BYTE    GpioCfg;            /* 02 Register counts and enable flag       */
        #define SMPGRC0_GC_GP_CNT_MASK  0x0F    /* Gen purpose reg count    */
        #define SMPGRC0_GC_CFG_CNT_MASK 0x70    /* Configuration reg count  */
        #define SMPGRC0_GC_CFG_CNT_SHIFT   4    /* Configuration reg shift  */
        #define SMPGRC0_GC_GPIO_ENB     0x80    /* GPIO enabled             */

    BYTE    DriveCnt;           /* 03 Supported drive count                 */
} SMPGPIOREG_CFG0,              /* 04 Size of structure                     */
*PSMPGPIOREG_CFG0;


typedef struct SmpGpioRegCfg1 {
    BYTE    Reserved;           /* 00 Reserved                              */
    BYTE    BlinkRate;          /* 01 Blink rate in (n+1)/8 of a second     */
        #define SMPGRC1_BR_A_MASK       0x0F    /* Blink rate A mask        */
        #define SMPGRC1_BR_B_MASK       0xF0    /* Blink rate B mask        */
        #define SMPGRC1_BR_B_SHIFT      4       /* Blink rate B shift       */

    BYTE    ActivityCtrl;       /* 02 Activity control in [] of a second    */
        #define SMPGRC1_AC_MAX_ON_MASK    0x0F  /* Max activity on [n/4]    */
        #define SMPGRC1_AC_FORCE_OFF_MASK 0xF0  /* Force activity off [n/8] */
        #define SMPGRC1_AC_FORCE_OFF_SHIFT   4  /* Force activity off shift */

    BYTE    StretchActCtrl;     /* 03 Stretch activity ctrl in [] of a sec  */
        #define SMPGRC1_SAC_ON_MASK     0x0F    /* Activity on [n+1/64]     */
        #define SMPGRC1_SAC_OFF_MASK    0xF0    /* Activity off [n/64]      */
        #define SMPGRC1_SAC_OFF_SHIFT   4       /* Activity off shift       */

} SMPGPIOREG_CFG1,              /* 04 Size of structure                     */
*PSMPGPIOREG_CFG1;


/* Receive registers */

#define SMPGRRX_DRIVE_MASK          0x07
#define SMPGRRX_DRIVE_INPUT0        0x01
#define SMPGRRX_DRIVE_INPUT1        0x02
#define SMPGRRX_DRIVE_INPUT2        0x04

#define SMPGRRX_DRIVE_TO_RX_IX(n)   \
        (3 - ((BYTE) (n) & 3) + ((BYTE) (n) & 0xFC))

/* Transmit registers */

#define SMPGRTX_ERROR_MASK          0x07
#define SMPGRTX_ERROR_SHIFT         0
  #define SMPGRTX_E_DISABLE         0x00
  #define SMPGRTX_E_ENABLE          0x01
  #define SMPGRTX_E_SELA_ON_FIRST   0x02
  #define SMPGRTX_E_SELA_ON_LAST    0x03
  #define SMPGRTX_E_ENABLE2         0x04
  #define SMPGRTX_E_ENABLE3         0x05
  #define SMPGRTX_E_SELB_ON_FIRST   0x06
  #define SMPGRTX_E_SELB_ON_LAST    0x07

#define SMPGRTX_LOCATE_MASK         0x18
#define SMPGRTX_LOCATE_SHIFT        3
  #define SMPGRTX_L_DISABLE         0x00
  #define SMPGRTX_L_ENABLE          0x08
  #define SMPGRTX_L_SELA_ON_FIRST   0x10
  #define SMPGRTX_L_SELA_ON_LAST    0x18

#define SMPGRTX_ACTIVITY_MASK       0xE0
#define SMPGRTX_ACTIVITY_SHIFT      5
  #define SMPGRTX_A_DISABLE         0x00
  #define SMPGRTX_A_ENABLE          0x20
  #define SMPGRTX_A_SELA_ON_FIRST   0x40
  #define SMPGRTX_A_SELA_ON_LAST    0x60
  #define SMPGRTX_A_BRIEF_ON_END    0x80
  #define SMPGRTX_A_BRIEF_ON_START  0xA0
  #define SMPGRTX_A_SELB_ON_FIRST   0xC0
  #define SMPGRTX_A_SELB_ON_LAST    0xF0

#define SMPGRTX_DRIVE_TO_TX_IX(n)   \
        (3 - ((BYTE) (n) & 3) + ((BYTE) (n) & 0xFC))

#define SMPGRTX_DRIVE_SHIFT(n)      ((3 - ((n) & 3)) * 8)

#define SMPGRTX_DRIVE_MASK          0xFF


/* Read GPIO command and response */

typedef struct SmpReqReadGpio {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RegType;            /* 02 Register type                         */
    BYTE    RegIndex;           /* 03 Register index                        */
    BYTE    RegCnt;             /* 04 Register count                        */
    BYTE    Reserved[3];        /* 05 Reserved                              */
} SMPREQ_READ_GPIO,             /* 08 Size of structure                     */
*PSMPREQ_READ_GPIO;


typedef struct SmpRspReadGpio {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Reserved;           /* 03 Reserved                              */
    DWORD   RegData[1];         /* 04 Variable list of register data        */
} SMPRSP_READ_GPIO,             /* XX Size of structure                     */
*PSMPRSP_READ_GPIO;


/* Write GPIO command and response */

typedef struct SmpReqWriteGpio {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    RegType;            /* 02 Register type                         */
    BYTE    RegIndex;           /* 03 Register index                        */
    BYTE    RegCnt;             /* 04 Register count                        */
    BYTE    Reserved[3];        /* 05 Reserved                              */
    DWORD   RegData[1];         /* 08 Variable list of register data        */
} SMPREQ_WRITE_GPIO,            /* XX Size of structure                     */
*PSMPREQ_WRITE_GPIO;


typedef struct SmpRspWriteGpio {
    BYTE    FrameType;          /* 00 Frame type                            */
    BYTE    Function;           /* 01 SMP function code                     */
    BYTE    Result;             /* 02 Function result                       */
    BYTE    Reserved;           /* 03 Reserved                              */
} SMPRSP_WRITE_GPIO,            /* 04 Size of structure                     */
*PSMPRSP_WRITE_GPIO;

/* Size macros */

#define SasSmpGetWrGpioCmdSz(RegCnt) (offsetof(SMPREQ_WRITE_GPIO, RegData)  \
                                    + ((RegCnt) * sizeof(DWORD)))

#define SasSmpGetRdGpioRspSz(RegCnt) (offsetof(SMPRSP_READ_GPIO, RegData)   \
                                    + ((RegCnt) * sizeof(DWORD)))

#pragma pack ()


/****************************************************************************
                    Unions of all requests and responses
*****************************************************************************/

typedef union {
    SMPREQ_HDR                  Hdr;
    SMPREQ_RPT_GEN              RptGen;
    SMPREQ_RPT_MFR              RptMfr;
    SMPREQ_RPT_SELF_CFG_STS     RptScs;
    SMPREQ_RPT_ZN_PERM_TBL      RptZpt;
    SMPREQ_RPT_ZN_MGR_PW        RptZmp;
    SMPREQ_RPT_BROADCAST        RptBc;
    SMPREQ_DISC                 Disc;
    SMPREQ_RPT_PHY_ERR_LOG      RptPeLog;
    SMPREQ_RPT_PHY_SATA         RptSata;
    SMPREQ_RPT_ROUTE_INFO       RptRi;
    SMPREQ_RPT_PHY_EVENT        RptPe;
    SMPREQ_DISC_LIST            DiscList;
    SMPREQ_RPT_PHY_EVENT_LIST   RptPeList;
    SMPREQ_RPT_EXP_RTE_TBL      RptErt;
    SMPREQ_CFG_GEN              CfgGen;
    SMPREQ_ENB_DIS_ZN           EnbDisZn;
    SMPREQ_ZN_BROADCAST         ZnBc;
    SMPREQ_ZN_LOCK              ZnLock;
    SMPREQ_ZN_ACTIVATE          ZnAct;
    SMPREQ_ZN_UNLOCK            ZnUnlock;
    SMPREQ_CFG_ZN_PASSWORD      CfgZnPw;
    SMPREQ_CFG_ZN_PHY_INFO      CfgZnPi;
    SMPREQ_CFG_ZN_PERM_TBL      CfgZnPt;
    SMPREQ_CFG_ROUTE_INFO       CfgRi;
    SMPREQ_PHY_CTRL             PhyCtrl;
    SMPREQ_PHY_TEST             PhyTest;
    SMPREQ_CFG_PHY_EVENT        CfgPe;
    SMPREQ_READ_GPIO            ReadGpio;
    SMPREQ_WRITE_GPIO           WriteGpio;
} SMPREQ, *PSMPREQ;


typedef union {
    SMPRSP_HDR                  Hdr;
    SMPRSP_RPT_GEN              RptGen;
    SMPRSP_RPT_MFR              RptMfr;
    SMPRSP_RPT_SELF_CFG_STS     RptScs;
    SMPRSP_RPT_ZN_PERM_TBL      RptZnPmt;
    SMPRSP_RPT_ZN_MGR_PW        RptZnPw;
    SMPRSP_RPT_BROADCAST        RptBc;
    SMPRSP_DISC                 Disc;
    SMPRSP_RPT_PHY_ERR_LOG      RptPeLog;
    SMPRSP_RPT_PHY_SATA         RptSata;
    SMPRSP_RPT_ROUTE_INFO       RptRi;
    SMPRSP_RPT_PHY_EVENT        RptPe;
    SMPRSP_DISC_LIST            DiscList;
    SMPRSP_RPT_PHY_EVENT_LIST   RptPeList;
    SMPRSP_RPT_EXP_RTE_TBL      RptErt;
    SMPRSP_CFG_GEN              CfgGen;
    SMPRSP_ENB_DIS_ZN           EnbDisZn;
    SMPRSP_ZN_BROADCAST         ZnBc;
    SMPRSP_ZN_LOCK              ZnLock;
    SMPRSP_ZN_ACTIVATE          ZnAct;
    SMPRSP_ZN_UNLOCK            ZnUnlock;
    SMPRSP_CFG_ZN_PASSWORD      CfgZnPw;
    SMPRSP_CFG_ZN_PHY_INFO      CfgZnPhi;
    SMPRSP_CFG_ZN_PERM_TBL      CfgZnPmt;
    SMPRSP_CFG_ROUTE_INFO       CfgRi;
    SMPRSP_PHY_CTRL             PhyCtrl;
    SMPRSP_PHY_TEST             PhyTest;
    SMPRSP_CFG_PHY_EVENT        CfgPe;
    SMPRSP_READ_GPIO            ReadGpio;
    SMPRSP_WRITE_GPIO           WriteGpio;
} SMPRSP, *PSMPRSP;

/* Request and response lengths are the number of DWORDs that follow the    */
/* header.                                                                  */

#define SasSmpGetDwordLen( ByteLen )  \
        ((BYTE)(((ByteLen) - 4) / sizeof(DWORD)))

#define SasSmpGetByteLen( DwordLen )    \
        (((DwordLen) + 1) * sizeof(DWORD))

#define SMP_MAX_PAYLOAD_LEN 1020
#define SMP_MAX_FRAME_LEN   1024


#endif /* HWSASSMP_H */
