/****************************************************************************

DESCRIPTION: SCSI general definitions.

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 1994-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef HWSCSI_H
#define HWSCSI_H


/****************************************************************************
                           SCSI STATUS VALUES
*****************************************************************************/

/* TM:!! SCSI status */
#define STATUS_GOOD     0x00    /* Status Good                              */
#define STATUS_CHKCOND  0x02    /* Check Condition                          */
#define STATUS_CONDMET  0x04    /* Condition Met                            */
#define STATUS_BUSY     0x08    /* Busy                                     */
#define STATUS_INTERM   0x10    /* Intermediate                             */
#define STATUS_INTCDMET 0x14    /* Intermediate-condition met               */
#define STATUS_RESCONF  0x18    /* Reservation conflict                     */
#define STATUS_COMTERM  0x22    /* Command Terminated                       */
#define STATUS_QFULL    0x28    /* Queue full                               */
#define STATUS_ACA_ACT  0x30    /* ACA active                               */
#define STATUS_ABORTED  0x40    /* Task aborted                             */
/* Tm:!! SCSI status */

/****************************************************************************
                             COMMAND OPCODES
*****************************************************************************/

#define SCSI_SERV_ACT_MASK  0x1F    /* Service action mask                  */

/* Command group codes */

#define SCSI_GROUP_CODE_MSK 0xE0    /* Group code mask                      */
#define SCSI_GROUP_6BYTE    0x00    /* 6  byte commands                     */
#define SCSI_GROUP_10BYTE_A 0x20    /* 10 byte commands                     */
#define SCSI_GROUP_10BYTE_B 0x40    /* 10 byte commands                     */
#define SCSI_GROUP_16BYTE   0x80    /* 16 byte commands                     */
#define SCSI_GROUP_12BYTE   0xA0    /* 12 byte commands                     */
#define SCSI_GROUP_VENDOR_A 0xC0    /* Vendor specific commands             */
#define SCSI_GROUP_VENDOR_B 0xE0    /* Vendor specific commands             */

/* Commands for all Device Types */

#define SCSI_ATA_PASSTHRU12 0xA1    /* ATA Pass-Through 12 (Optional)       */
#define SCSI_ATA_PASSTHRU16 0x85    /* ATA Pass-Through 16 (O)              */
#define SCSI_CHANGE_DEF     0x40    /* Change Definition (O)                */
#define SCSI_COMPARE        0x39    /* Compare (O)                          */
#define SCSI_COPY           0x18    /* Copy (O)                             */
#define SCSI_COP_VERIFY     0x3A    /* Copy and Verify (O)                  */
#define SCSI_EXTENDED_COPY  0x83    /* Extended Copy (O)                    */
#define SCSI_INQUIRY        0x12    /* Inquiry (MANDATORY)                  */
#define SCSI_LOG_SELECT     0x4C    /* Log Select (O)                       */
#define SCSI_LOG_SENSE      0x4D    /* Log Sense (O)                        */
#define SCSI_MAINT_IN       0xA3    /* Maintenance In                       */
#define SCSI_MAINT_OUT      0xA4    /* Maintenance Out                      */
#define SCSI_MODE_SEL6      0x15    /* Mode Select 6-byte (Device Specific) */
#define SCSI_MODE_SEL10     0x55    /* Mode Select 10-byte (Device Specific)*/
#define SCSI_MODE_SEN6      0x1A    /* Mode Sense 6-byte (Device Specific)  */
#define SCSI_MODE_SEN10     0x5A    /* Mode Sense 10-byte (Device Specific) */
#define SCSI_PERS_RSRV_IN   0x5E    /* Persistent Reserve In (O)            */
#define SCSI_PERS_RSRV_OUT  0x5F    /* Persistent Reserve Out (O)           */
#define SCSI_RD_MED_SER_NUM 0xAB    /* Read Media Serial Number (O)         */
#define SCSI_READ_ATTR      0x8C    /* Read Attribute (O)                   */
#define SCSI_READ_BUFF      0x3C    /* Read Buffer (O)                      */
#define SCSI_RELEASE        0x17    /* Release 6 (O)                        */
#define SCSI_RELEASE10      0x57    /* Release 10 (O)                       */
#define SCSI_REPORTLUNS     0xA0    /* Report Luns                          */
#define SCSI_REQ_SENSE      0x03    /* Request Sense (MANDATORY)            */
#define SCSI_RESERVE        0x16    /* Reserve 6 (O)                        */
#define SCSI_RESERVE10      0x56    /* Reserve 10 (O)                       */
#define SCSI_SEC_PROT_IN    0xA2    /* Security Protocol In (O)             */
#define SCSI_SEC_PROT_OUT   0xB5    /* Security Protocol Out (O)            */
#define SCSI_RCV_DIAG       0x1C    /* Recieve Diagnostic Result (0)        */
#define SCSI_SEND_DIAG      0x1D    /* Send Diagnostic (O)                  */
#define SCSI_TST_U_RDY      0x00    /* Test Unit Ready (MANDATORY)          */
#define SCSI_WRITE_BUFF     0x3B    /* Write Buffer (O)                     */
#define SCSI_SRVC_ACT_IN16  0x9E    /* Service Action In (16)               */
#define SCSI_SRVC_ACT_OUT16 0x9F    /* Service Action Out (16)              */

/* Commands Unique to Direct Access Devices */

#define SCSI_COMPARE        0x39    /* Compare (O)                          */
#define SCSI_COMP_WRITE     0x89    /* Compare and Write (O)                */
#define SCSI_FORMAT         0x04    /* Format Unit (MANDATORY)              */
#define SCSI_LCK_UN_CAC     0x36    /* Lock Unlock Cache (O)                */
#define SCSI_ORWRITE        0x8B    /* OR Write (O)                         */
#define SCSI_PREFETCH       0x34    /* Prefetch (O)                         */
#define SCSI_MED_REMOVL     0x1E    /* Prevent/Allow medium Removal (O)     */
#define SCSI_READ6          0x08    /* Read 6-byte (MANDATORY)              */
#define SCSI_READ10         0x28    /* Read 10-byte (MANDATORY)             */
#define SCSI_READ12         0xA8    /* Read 12-byte (O)                     */
#define SCSI_READ16         0x88    /* Read 16-byte (O)                     */
#define SCSI_RD_CAPAC       0x25    /* Read Capacity (MANDATORY)            */
#define SCSI_RD_CAPAC16     0x9E    /* Read Capacity 16 (O)                 */
#define SCSI_RD_DEFECT      0x37    /* Read Defect Data 10 (O)              */
#define SCSI_RD_DEFECT12    0xB7    /* Read Defect Data 12 (O)              */
#define SCSI_READ_LONG      0x3E    /* Read Long (O)                        */
#define SCSI_READ_LONG16    0x9E    /* Read Long 16 (O)                     */
#define SCSI_REASS_BLK      0x07    /* Reassign Blocks (O)                  */
#define SCSI_REZERO         0x01    /* Rezero Unit (O)                      */
#define SCSI_SANITIZE       0x48    /* Sanitize (O)                         */
#define SCSI_SRCH_DAT_E     0x1D    /* Search Data Equal (O)                */
#define SCSI_SRCH_DAT_H     0x1D    /* Search Data High (O)                 */
#define SCSI_SRCH_DAT_L     0x1D    /* Search Data Low (O)                  */
#define SCSI_SEEK6          0x0B    /* Seek 6-Byte (O)                      */
#define SCSI_SEEK10         0x2B    /* Seek 10-Byte (O)                     */
#define SCSI_SET_LIMIT      0x33    /* Set Limits (O)                       */
#define SCSI_START_STP      0x1B    /* Start/Stop Unit (O)                  */
#define SCSI_SYNC_CACHE     0x35    /* Synchronize Cache 10 (O)             */
#define SCSI_SYNC_CACHE16   0x91    /* Synchronize Cache 16 (O)             */
#define SCSI_UNMAP          0x42    /* Unmap (O)                            */
#define SCSI_VERIFY         0x2F    /* Verify (O)                           */
#define SCSI_VERIFY12       0xAF    /* Verify 12 (O)                        */
#define SCSI_VERIFY16       0x8F    /* Verify 16 (O)                        */
#define SCSI_WRITE6         0x0A    /* Write 6-Byte (MANDATORY)             */
#define SCSI_WRITE10        0x2A    /* Write 10-Byte (MANDATORY)            */
#define SCSI_WRITE12        0xAA    /* Write 12-Byte (O)                    */
#define SCSI_WRITE16        0x8A    /* Write 16-Byte (O)                    */
#define SCSI_WRT_VERIFY     0x2E    /* Write and Verify (O)                 */
#define SCSI_WRT_VERIFY12   0xAE    /* Write and Verify 12 (O)              */
#define SCSI_WRT_VERIFY16   0x8E    /* Write and Verify 16 (O)              */
#define SCSI_WRITE_LONG     0x3F    /* Write Long (O)                       */
#define SCSI_WRITE_LONG16   0x9F    /* Write Long 16 (O)                    */
#define SCSI_WRITE_SAME     0x41    /* Write Same (O)                       */
#define SCSI_WRITE_SAME16   0x93    /* Write Same 16 (O)                    */
#define SCSI_ZBC_OUT        0x94    /* ZBC Out                              */
#define SCSI_ZBC_IN         0x95    /* ZBC In                               */
#define SCSI_XDREAD         0x52    /* XD Read (O)                          */
#define SCSI_XDWRITE        0x50    /* XD Write (O)                         */
#define SCSI_XDWRT_READ     0x53    /* XD Write-Read (O)                    */
#define SCSI_XPWRITE        0x51    /* XP Write (O)                         */

/* Commands Unique to Sequential Access Devices */

#define SCSI_ERASE          0x19    /* Erase (MANDATORY)                    */
#define SCSI_ERASE16        0x93    /* Erase 16 (MANDATORY)                 */
#define SCSI_LOAD_UN        0x1B    /* Load/Unload (O)                      */
#define SCSI_LOCATE         0x2B    /* Locate (O)                           */
#define SCSI_PER_RSRV_IN    0x5E    /* Persistent Reservation In (MANDATORY)*/
#define SCSI_PER_RSRV_OUT   0x5F    /* Persistent Reservation Out(MANDATORY)*/
#define SCSI_RD_BLK_LIM     0x05    /* Read Block Limits (MANDATORY)        */
#define SCSI_READ_POS       0x34    /* Read Position (O)                    */
#define SCSI_READ_REV       0x0F    /* Read Reverse (O)                     */
#define SCSI_REC_BF_DAT     0x14    /* Recover Buffer Data (O)              */
#define SCSI_REWIND         0x01    /* Rewind (MANDATORY)                   */
#define SCSI_SPACE          0x11    /* Space (MANDATORY)                    */
#define SCSI_VERIFY_T       0x13    /* Verify (Tape) (O)                    */
#define SCSI_WRT_FILE       0x10    /* Write Filemarks (MANDATORY)          */

/* Commands Unique to Printer Devices */

#define SCSI_PRINT          0x0A    /* Print (MANDATORY)                    */
#define SCSI_SLEW_PNT       0x0B    /* Slew and Print (O)                   */
#define SCSI_STOP_PNT       0x1B    /* Stop Print (O)                       */
#define SCSI_SYNC_BUFF      0x10    /* Synchronize Buffer (O)               */

/* Commands Unique to Processor Devices */

#define SCSI_RECEIVE        0x08    /* Receive (O)                          */
#define SCSI_SEND           0x0A    /* Send (O)                             */

/* Commands Unique to Write-Once Devices */

#define SCSI_MEDIUM_SCN     0x38    /* Medium Scan (O)                      */
#define SCSI_SRCHDATE10     0x31    /* Search Data Equal 10-Byte (O)        */
#define SCSI_SRCHDATE12     0xB1    /* Search Data Equal 12-Byte (O)        */
#define SCSI_SRCHDATH10     0x30    /* Search Data High 10-Byte (O)         */
#define SCSI_SRCHDATH12     0xB0    /* Search Data High 12-Byte (O)         */
#define SCSI_SRCHDATL10     0x32    /* Search Data Low 10-Byte (O)          */
#define SCSI_SRCHDATL12     0xB2    /* Search Data Low 12-Byte (O)          */
#define SCSI_SET_LIM_10     0x33    /* Set Limits 10-Byte (O)               */
#define SCSI_SET_LIM_12     0xB3    /* Set Limits 10-Byte (O)               */
#define SCSI_VERIFY10       0x2F    /* Verify 10-Byte (O)                   */
#define SCSI_VERIFY12       0xAF    /* Verify 12-Byte (O)                   */
#define SCSI_WRITE12        0xAA    /* Write 12-Byte (O)                    */
#define SCSI_WRT_VER10      0x2E    /* Write and Verify 10-Byte (O)         */
#define SCSI_WRT_VER12      0xAE    /* Write and Verify 12-Byte (O)         */

/* Commands Unique to CD-ROM Devices */

#define SCSI_PLAYAUD_10     0x45    /* Play Audio 10-Byte (O)               */
#define SCSI_PLAYAUD_12     0xA5    /* Play Audio 12-Byte 12-Byte (O)       */
#define SCSI_PLAYAUDMSF     0x47    /* Play Audio MSF (O)                   */
#define SCSI_PLAYA_TKIN     0x48    /* Play Audio Track/Index (O)           */
#define SCSI_PLYTKREL10     0x49    /* Play Track Relative 10-Byte (O)      */
#define SCSI_PLYTKREL12     0xA9    /* Play Track Relative 12-Byte (O)      */
#define SCSI_READCDCAP      0x25    /* Read CD-ROM Capacity (MANDATORY)     */
#define SCSI_READHEADER     0x44    /* Read Header (O)                      */
#define SCSI_SUBCHANNEL     0x42    /* Read Subchannel (O)                  */
#define SCSI_READ_TOC       0x43    /* Read TOC (O)                         */

/* Commands Unique to Scanner Devices */

#define SCSI_GETDBSTAT      0x34    /* Get Data Buffer Status (O)           */
#define SCSI_GETWINDOW      0x25    /* Get Window (O)                       */
#define SCSI_OBJECTPOS      0x31    /* Object Postion (O)                   */
#define SCSI_SCAN           0x1B    /* Scan (O)                             */
#define SCSI_SETWINDOW      0x24    /* Set Window (MANDATORY)               */

/* Commands Unique to Optical Memory Devices */

#define SCSI_UpdateBlk      0x3D    /* Update Block (O)                     */

/* Commands Unique to Medium Changer Devices */

#define SCSI_EXCHMEDIUM     0xA6    /* Exchange Medium (O)                  */
#define SCSI_INITELSTAT     0x07    /* Initialize Element Status (O)        */
#define SCSI_POSTOELEM      0x2B    /* Position to Element (O)              */
#define SCSI_REQ_VE_ADD     0xB5    /* Request Volume Element Address (O)   */
#define SCSI_SENDVOLTAG     0xB6    /* Send Volume Tag (O)                  */

/* Commands Unique to Communication Devices */

#define SCSI_GET_MSG_6      0x08    /* Get Message 6-Byte (MANDATORY)       */
#define SCSI_GET_MSG_10     0x28    /* Get Message 10-Byte (O)              */
#define SCSI_GET_MSG_12     0xA8    /* Get Message 12-Byte (O)              */
#define SCSI_SND_MSG_6      0x0A    /* Send Message 6-Byte (MANDATORY)      */
#define SCSI_SND_MSG_10     0x2A    /* Send Message 10-Byte (O)             */
#define SCSI_SND_MSG_12     0xAA    /* Send Message 12-Byte (O)             */


/****************************************************************************
                           SCSI HIERARCHICAL LUN
*****************************************************************************/

typedef union _SCSI_LUN {
    BYTE  b[8];
    WORD  w[4];
    DWORD d[2];
} SCSI_LUN, *PSCSI_LUN;

/* Addressing method masks */

#define SCSI_LUN_ADDR_METH_B    0xC0
#define SCSI_LUN_ADDR_METH_W    0xC000

/* Maximum LUN counts */

#define SCSI_LUN_MAX_PERI_DEV   0x100   /* Single level, addressing method 0*/
#define SCSI_LUN_MAX_FLAT       0x4000  /* Single level, addressing method 1*/

/* Legacy */

#define MAX_SLVL_LUNS   256     /* Maximum Number of Single-level LUNs      */

#pragma pack (1)


/****************************************************************************
                     REQUEST SENSE COMMAND AND DATA
*****************************************************************************/

typedef struct CdbRequestSense {
    BYTE    Command;        /* 00 SCSI_REQ_SENSE (0x03)                     */
    BYTE    Flags;          /* 01 Flags/Reserved                            */
        #define REQSNSF_DESC    0x01    /* Descriptor data format           */

    BYTE    Reserved[2];    /* 02 Reserved                                  */
    BYTE    Length;         /* 04 Data length                               */
    BYTE    Control;        /* 05 Control                                   */
} CDB_REQ_SENSE,            /* 06 Size of structure                         */
*PCDB_REQ_SENSE;

/* Fixed format sense data */

typedef struct {
    BYTE    ErrorCode;          /* 00   Error Code (70H or 71H)             */
        #define SERROR_MASK     0x7F    /* Mask for error code              */
        #define SERROR_CURRENT  0x70    /* Current Errors                   */
        #define SERROR_DEFERRED 0x71    /* Deferred Errors                  */
        #define SENSE_VALID     0x80    /* Information field is valid       */

    BYTE    SegmentNum;         /* 01   # of current segment descriptor     */
    BYTE    SenseKey;           /* 02   Sense Key                           */
        #define SENSE_FILEMRK   0x80    /* Filemark                         */
        #define SENSE_EOM       0x40    /* End of medium                    */
        #define SENSE_ILI       0x20    /* Incorrect length indicator       */
        #define SENSE_OVERFLOW  0x10    /* Sense data overflow              */

        #define KEY_MASK        0x0F    /* Sense Key Mask                   */
        #define KEY_NOSENSE     0x00    /* No Sense                         */
        #define KEY_RECERROR    0x01    /* Recovered Error                  */
        #define KEY_NOTREADY    0x02    /* Not Ready                        */
        #define KEY_MEDIUMERR   0x03    /* Medium Error                     */
        #define KEY_HARDERROR   0x04    /* Hardware Error                   */
        #define KEY_ILLGLREQ    0x05    /* Illegal Request                  */
        #define KEY_UNITATT     0x06    /* Unit Attention                   */
        #define KEY_DATAPROT    0x07    /* Data Protect                     */
        #define KEY_BLANKCHK    0x08    /* Blank Check                      */
        #define KEY_VENDSPEC    0x09    /* Vendor Specific                  */
        #define KEY_COPYABORT   0x0A    /* Copy Abort                       */
        #define KEY_ABORTEDCMD  0x0B    /* Aborted Command                  */
        #define KEY_EQUAL       0x0C    /* Equal (Search)                   */
        #define KEY_VOLOVRFLW   0x0D    /* Volume Overflow                  */
        #define KEY_MISCOMP     0x0E    /* Miscompare (Search)              */
        #define KEY_COMPLETED   0x0F    /* Completed                        */

    BYTE    InfoByte0;          /* 03   Information MSB                     */
    BYTE    InfoByte1;          /* 04   Information MID                     */
    BYTE    InfoByte2;          /* 05   Information MID                     */
    BYTE    InfoByte3;          /* 06   Information LSB                     */
    BYTE    AddSenLen;          /* 07   Additional Sense Length             */
    BYTE    ComSpecInf0;        /* 08   Command Specific Info MSB           */
    BYTE    ComSpecInf1;        /* 09   Command Specific Info MID           */
    BYTE    ComSpecInf2;        /* 0a   Command Specific Info MID           */
    BYTE    ComSpecInf3;        /* 0b   Command Specific Info LSB           */
    BYTE    AddSenseCode;       /* 0c   Additional Sense Code               */
        #define SCODE_NOSENSE           0x00    /* No additional sense code */
        #define SCODE_WRITEFAULT        0x03    /* Write fault              */
        #define SCODE_NOTREADY          0x04    /* Unit not ready           */
        #define SCODE_NO_SELECTION      0x05    /* Does not respond to sel  */
        #define SCODE_LUNCOMMFAILURE    0x08    /* LUN Communication Fail   */
        #define SCODE_WRITEERROR        0x0C    /* Write error              */
        #define SCODE_LUNTIMEOUT        0x0E    /* LUN selection timeout    */
        #define SCODE_CRC_ECC           0x10    /* CRC or ECC error         */
        #define SCODE_READERROR         0x11    /* Unrecovered read error   */
        #define SCODE_DEFECT_LIST       0x19    /* Defect list error        */
        #define SCODE_PARAM_LIST_LEN    0x1A    /* Invalid param list length*/
        #define SCODE_INVALIDOPCODE     0x20    /* Invalid Command OpCode   */
        #define SCODE_LBA_RANGE         0x21    /* LBA out of range         */
        #define SCODE_INVALCDB          0x24    /* Invalid CDB Field        */
        #define SCODE_INV_LUN           0x25    /* Invalid LUN              */
        #define SCODE_PARAM_LIST_FIELD  0x26    /* Invalid param list field */
        #define SCODE_WRITE_PROTECT     0x27    /* Media write protected    */
        #define SCODE_MEDIA_CHANGE      0x28    /* Media changed            */
        #define SCODE_POWERRESET        0x29    /* Power On, Reset          */
        #define SCODE_PARAM_CHANGE      0x2A    /* Parameters changed       */
        #define SCODE_CMDSEQUENCE       0x2C    /* Command Sequence Error   */
        #define SCODE_CLEARED           0x2f    /* Commands cleared by      */
                                                /* another initiator        */
        #define SCODE_FORMAT            0x31    /* Format error             */
        #define SCODE_SAVING_NOT_SUPP   0x39    /* Save Params Not Suported */
        #define SCODE_NO_MEDIA          0x3A    /* Media not present        */
        #define SCODE_LUN_FAILURE       0x3E    /* Logical unit failures    */
        #define SCODE_TARG_CHANGE       0x3F    /* Operating conditions chg */
        #define SCODE_DIAG_FAILURE      0x40    /* Diagnostic failure       */
        #define SCODE_MSGERROR          0x43    /* Message error (reject?)  */
        #define SCODE_INT_TARG_FAIL     0x44    /* Internal target failure  */
        #define SCODE_SCSIPE            0x47    /* SCSI Bus Parity Error    */
        #define SCODE_IDE               0x48    /* Initiator Detected Error */
        #define SCODE_BADMSG            0x49    /* Invalid Message Received */
        #define SCODE_DATAPHASEERROR    0x4B    /* Data phase error.        */
        #define SCODE_OVERLAP           0x4E    /* Attempted overlapped cmds*/
        #define SCODE_SCSIIFERROR       0x54    /* SCSI Interface Error     */
        #define SCODE_MEDIA_RMV_REQ     0x5A    /* Media removal request    */
        #define SCODE_HW_IMPENDING_FAIL 0x5D    /* HW impending failure     */
        #define SCODE_POWER_CHANGE      0x5E    /* Power state changed      */
        #define SCODE_LUN_MGT_FAIL      0x67    /* LUN management failure   */
        #define SCODE_VENDOR_SPEC       0x80    /* Start vendor specific    */

    BYTE    AddSenQual;         /* 0d   Additional Sense Code Qualifier     */
        #define SQUAL_VENDOR_SPEC       0x80    /* Start vendor specific    */

    BYTE    FieldRepUCode;      /* 0e   Field Replaceable Unit Code         */
    BYTE    SenKeySpec15;       /* 0f   Sense Key Specific 15th byte        */
        #define SKS_VALID               0x80    /* Sense-key specific valid */

    BYTE    SenKeySpec16;       /* 10   Sense Key Specific 16th byte        */
    BYTE    SenKeySpec17;       /* 11   Sense Key Specific 17th byte        */
    BYTE    AddSenseBytes [1];  /* 12   Additional Sense Bytes              */
} SENSE_DATA_FMT,               /* 13   Size of sense data                  */
*PSENSE_DATA_FMT;


#define SENSE_DATA_LEN 0x12
#define SENSE_DATA_MAX 0xFC

/* Descriptor format sense data */

typedef struct {
    BYTE    ErrorCode;          /* 00   Error Code (72H or 73H)             */
        #define SDD_ERROR_CURRENT  0x72 /* Current Errors                   */
        #define SDD_ERROR_DEFERRED 0x73 /* Deferred Errors                  */

    BYTE    SenseKey;           /* 01   Sense Key                           */
        #define SDD_SK_MASK        0x0F /* Sense Key Mask                   */

    BYTE    AddSenseCode;       /* 02   Additional Sense Code               */
    BYTE    AddSenQual;         /* 03   Additional Sense Code Qualifier     */
    BYTE    Reserved[3];        /* 04   Reserved                            */
    BYTE    AddSenLen;          /* 07   Additional Sense Length             */
} SENSE_DATA_DESC,              /* 08   Size of sense data                  */
*PSENSE_DATA_DESC;


typedef struct SenseDescriptorHeader {
    BYTE    DescType;
    BYTE    AddLength;
} SNS_DESC_HDR, *PSNS_DESC_HDR;


typedef struct InfoDescriptor {
    BYTE    DescType;
        #define INFO_DESC_TYPE  0x00

    BYTE    AddLength;
    BYTE    Flags;
        #define INFO_VALID      0x80

    BYTE    Reserved;
    BYTE    Info[8];
} SNS_DESC_INFO, *PSNS_DESC_INFO;


typedef struct SenseKeySpecDescriptor {
    BYTE    DescType;
        #define SKS_DESC_TYPE   0x02

    BYTE    AddLength;
    BYTE    Reserved1;
    BYTE    Reserved2;
    BYTE    SenKeySpec1;
        #define SKSD_VALID      0x80

    BYTE    SenKeySpec2;
    BYTE    SenKeySpec3;

    BYTE    Reserved3;
} SNS_DESC_SKS, *PSNS_DESC_SKS;


typedef struct AtaReturnDescriptor {
    BYTE    DescType;
        #define ARD_DESC_TYPE   0x09

    BYTE    AddLength;
    BYTE    Flags;
        #define ARDF_EXTEND     0x01

    BYTE    Error;
    BYTE    SectorCntExp;
    BYTE    SectorCnt;
    BYTE    LbaLowExp;
    BYTE    LbaLow;
    BYTE    LbaMidExp;
    BYTE    LbaMid;
    BYTE    LbaHighExp;
    BYTE    LbaHigh;
    BYTE    Device;
    BYTE    Status;
} SNS_DESC_ATA_RETURN,
  *PSNS_DESC_ATA_RETURN,
  /* Legacy */
  ATA_RETURN_DESC,
  *PATA_RETURN_DESC;


/****************************************************************************
                        INQUIRY COMMAND AND DATA
*****************************************************************************/

typedef struct CdbInquiry {
    BYTE    Command;        /* 00 SCSI_INQUIRY (0x12)                       */
    BYTE    Flags;          /* 01 Flags/Reserved                            */
        #define INQF_ENB_VPD    0x01    /* Enable vital product data        */
        #define INQF_CMD_DATA   0x02    /* Return command support data      */

        #define INQ_Evpd        0x01    /* Old definition                   */
        #define INQ_CmdDt       0x02    /* Old definition                   */

    BYTE    PageCode;       /* 02 Page code when either flag is set         */
    BYTE    Reserved;       /* 03 Reserved                                  */
    BYTE    Length;         /* 04 Data length                               */
    BYTE    Control;        /* 05 Control                                   */
} CDB_INQUIRY,              /* 06 Size of structure                         */
*PCDB_INQUIRY;


typedef struct InqData {
    BYTE    PQType;             /* 00   Peripheral qualifier and type       */
        #define IPQ_TYPE        0x1f    /* Mask of device type              */

        #define IPQ_QUAL        0xe0    /* Mask of peripheral qualifier     */
        #define IPQ_QCONN       0x00    /* Device currently connected       */
        #define IPQ_NCONN       0x20    /* Device not connected             */
        #define IPQ_NOTCAP      0x60    /* Target not capable this LUN      */

        #define DTYPE_DASD      0x00    /* Disk Device                      */
        #define DTYPE_SEQD      0x01    /* Tape Device                      */
        #define DTYPE_PRNT      0x02    /* Printer                          */
        #define DTYPE_PROC      0x03    /* Processor                        */
        #define DTYPE_WORM      0x04    /* Write-once read-multiple         */
        #define DTYPE_CROM      0x05    /* CD-ROM device                    */
        #define DTYPE_SCAN      0x06    /* Scanner device                   */
        #define DTYPE_OPTI      0x07    /* Optical memory device            */
        #define DTYPE_JUKE      0x08    /* Medium Changer device            */
        #define DTYPE_COMM      0x09    /* Communications device            */
        #define DTYPE_SCC       0x0C    /* Storage array (RAID) controller  */
        #define DTYPE_SES       0x0D    /* Storage Enclosure Services (SES) */
        #define DTYPE_SDAD      0x0E    /* Simplified direct-access device  */
        #define DTYPE_OCRW      0x0F    /* Optical card reader/writer       */
        #define DTYPE_BCC       0x10    /* Bridge controller commands       */
        #define DTYPE_OSD       0x11    /* Object-based storage device      */
        #define DTYPE_ZBD       0x14    /* Zoned block device (host managed)*/
        #define DTYPE_RESL      0x0F    /* Reserved (low)                   */
        #define DTYPE_RESH      0x1E    /* Reserved (high)                  */
        #define DTYPE_UNKNOWN   0x1F    /* Unknown or no device type        */

    BYTE    RMB;                /* 01   Removable media in bit 7            */
        #define IRM_REM         0x80    /* The removable bit                */

    BYTE    Version;            /* 02   ANSI, ECMA, ISO version numbers     */
        #define IVER_ANSI       0x07    /* ANSI version mask                */
        #define IVER_ECMA       0x38    /* ECMA version mask                */
        #define IVER_ISO        0xc0    /* ISO  version mask                */

        #define ANSI_MAYBE      0x00    /* Device may or may not conform to */
                                        /* ANSI standard                    */
        #define ANSI_SCSI1      0x01    /* Device complies to SCSI-1        */
        #define ANSI_SCSI2      0x02    /* Device complies to SCSI-2        */
        #define ANSI_SPC1       0x03    /* Device complies to SPC-1         */
        #define ANSI_SPC2       0x04    /* Device complies to SPC-2         */
        #define ANSI_SPC3       0x05    /* Device complies to SPC-3         */
        #define ANSI_SPC4       0x06    /* Device complies to SPC-4         */
        #define ANSI_RESLO      0x07    /* Reserved (low)                   */
        #define ANSI_RESHI      0x07    /* Reserved (high)                  */

    BYTE    Format;             /* 03   Resp. format, TrmIOP and AENC       */
        #define IFMT_RESP       0x0f    /* Response data format             */
        #define IFMT_RESP_I     0x00    /* (OBS) SCSI I  format             */
        #define IFMT_RESP_CCS   0x01    /* (OBS) CCS     format             */
        #define IFMT_RESP_II    0x02    /* SCSI II format                   */

        #define IFMT_HISUPPORT  0x10    /* Hierarchical addressing          */
        #define IFMT_NormACA    0x20    /* Normal ACA Supported             */
        #define IFMT_TrmIOP     0x40    /* (OBS) Terminate I/O process      */
        #define IFMT_AENC       0x80    /* (OBS) Async event notifications  */

    BYTE    AddLength;          /*  04  Additional length of parameters     */
    BYTE    Flags4;             /*  05  Miscellaneous flags                 */

        #define IFLG4_PROTECT   0x01    /* Protection info supported        */
        #define IFLG4_3PC       0x08    /* Third party copy supported       */
        #define IFLG4_TPGS_MASK 0x30    /* Target Port Group Support        */
        #define IFLG4_TPGS_NONE 0x00    /* TPGS not supported               */
        #define IFLG4_TPGS_IMP  0x10    /* Implicit asymmetric TPGS         */
        #define IFLG4_TPGS_EXP  0x20    /* Explicit asymmetric TPGS         */
        #define IFLG4_TPGS_BOTH 0x30    /* Implicit and explicit TGPS       */
        #define IFLG4_ACC       0x40    /* Access Controls Coordinator      */
        #define IFLG4_SCCS      0x80    /* Storage Controller Component     */

    BYTE    Flags2;             /*  06  New miscellaneous flag bits         */
        #define IFLG2_BQue      0x80    /* Supports basic queueing          */
        #define IFLG2_EncServ   0x40    /* Contains encl svc component      */
        #define IFLG2_MultiP    0x10    /* Multi-port device                */
        #define IFLG2_MChngr    0x08    /* Attached to a medium changer     */
        #define IFLG2_ACKREQQ   0x04    /* (OBS)                            */
        #define IFLG2_Addr32    0x02    /* (OBS)                            */
        #define IFLG2_Addr16    0x01    /* SPI only                         */

    BYTE    Flags;              /* 07   Miscellaneous flag bits             */
        #define IFLG_RelAdr     0x80    /* (OBS) Supports rel. addressing   */
        #define IFLG_WBus32     0x40    /* (OBS) 32-bit transfers           */
        #define IFLG_WBus16     0x20    /* Supports 16-bit transfers        */
        #define IFLG_Sync       0x10    /* Supports synchronous transfers   */
        #define IFLG_Linked     0x08    /* Supports linked commands         */
        #define IFLG_TranDis    0x04    /* (OBS)                            */
        #define IFLG_CmdQue     0x02    /* Supports tagged command queuing  */
        #define IFLG_SftRe      0x01    /* Vendor specific                  */

    char    VendorID  [ 8];     /* 08   Vendor  identification              */
    char    ProductID [16];     /* 10   Product identification              */
    char    RevLevel  [ 4];     /* 20   Product revision level              */
    char    VendorStuff[20];    /* 24   Vendor specific info                */
    BYTE    Flags3;             /* 38   SPI-3 flags                         */
        #define IFLG3_ST_DT     0x0C    /* Mask for ST/DT capabiliities     */
        #define IFLG3_ST        0x00    /* ST only supported                */
        #define IFLG3_DT        0x04    /* DT only supported                */
        #define IFLG3_STDT      0x0C    /* ST & DT supported                */

        #define IFLG3_QAS       0x02    /* Quick Arbitration & Selection    */
        #define IFLG3_IUS       0x01    /* Information Units Supported      */
    
    BYTE    Reserved2;          /* 39   Don't expect anything here          */
    WORD    VersionDesc[8];     /* 3A   Version descriptors                 */  
    BYTE    Reserved3[22];      /* 4A   Don't expect anything here          */
    BYTE    VenSpecific2[4];    /* 60   Vendor specific info                */
} INQDATA, *PINQDATA;           /* 64   Total length                        */


#define INQLEN_MIN      0x24
#define INQLEN_STANDARD 0x60


typedef struct InquiryVPDHeader {
    BYTE    PeripheralDT;   /* 00 See INQUIRY.PQType                        */
    BYTE    PageCode;       /* 01 Page code                                 */
        #define INQ_VPD_PG00    0x00    /* Supported pages                  */
        #define INQ_VPD_PG80    0x80    /* Unit serial number               */
        #define INQ_VPD_PG83    0x83    /* Device identification            */
        #define INQ_VPD_PG86    0x86    /* Extended Inquiry Data            */
        #define INQ_VPD_PG87    0x87    /* Mode page policy                 */
        #define INQ_VPD_PG89    0x89    /* ATA information                  */
        #define INQ_VPD_PG8A    0x8A    /* Power Condition                  */
        #define INQ_VPD_PGB0    0xB0    /* Block limits                     */
        #define INQ_VPD_PGB1    0xB1    /* Block device characteristics     */
        #define INQ_VPD_PGB2    0xB2    /* Logical block provisioning       */
        #define INQ_VPD_PGB6    0xB6    /* Zoned block dev. characteristics */

    union {
        WORD    PageLen16;  /* 02 Length of data minus header               */

        struct {
            BYTE Reserved;  /* 02 Reserved                                  */
            BYTE PageLen8;  /* 03 Length of data minus header               */
        } s;
    } u;
} INQ_VPD_HDR,              /* 04 Size of structure                         */
*PINQ_VPD_HDR;

/* Device identification (SPC) */

typedef struct InquiryVPD83Desc {
    BYTE    CodeSet;        /* 00 Code set and protocol ID                  */

        #define PG83_CS_MASK    0x0F    /* Code set mask                    */
        #define PG83_CS_BINARY  0x01    /* Binary identifier                */
        #define PG83_CS_ASCII   0x02    /* ASCII identifier                 */
        #define PG83_CS_UTF8    0x03    /* ISO/IEC 10646-1 identifier       */

        #define PG83_PID_MASK   0xF0    /* Protocol ID mask                 */
        #define PG83_PID_FC     0x00
        #define PG83_PID_SPI    0x10
        #define PG83_PID_SSA    0x20
        #define PG83_PID_1394   0x30
        #define PG83_PID_SRP    0x40
        #define PG83_PID_ISCSI  0x50
        #define PG83_PID_SAS    0x60
        #define PG83_PID_ADT    0x70
        #define PG83_PID_ATA    0x80
        #define PG83_PID_USB    0x90
        #define PG83_PID_SOP    0xA0
        #define PG83_PID_PCIE   0xB0

    BYTE    AssocType;      /* 01 Association/type/flags                    */

        #define PG83_TYPE_MASK  0x0F    /* Identifier type mask             */
        #define PG83_TYPE_VEN   0x00    /* Vendor specific                  */
        #define PG83_TYPE_T10   0x01    /* T10 vendor                       */
        #define PG83_TYPE_EUI64 0x02    /* EUI-64                           */
        #define PG83_TYPE_NAA   0x03    /* NAA                              */
        #define PG83_TYPE_TREL  0x04    /* Relative target port             */
        #define PG83_TYPE_TGRP  0x05    /* Target port group                */
        #define PG83_TYPE_LGRP  0x06    /* LUN group                        */
        #define PG83_TYPE_LMD5  0x07    /* MD5 LUN                          */
        #define PG83_TYPE_NAME  0x08    /* SCSI name                        */

        #define PG83_ASSOC_MASK 0x30    /* Association mask                 */
        #define PG83_ASSOC_LUN  0x00    /* Addressed Logical unit           */
        #define PG83_ASSOC_PORT 0x10    /* SCSI target port                 */
        #define PG83_ASSOC_TARG 0x20    /* SCSI target device               */

        #define PG83F_PID_VALID 0x80    /* Protocol identifier valid        */

    BYTE    Reserved;       /* 02 Reserved                                  */
    BYTE    Length;         /* 03 Length of identifier                      */
    BYTE    Identifier[1];  /* 04 Identifier                                */
} INQ_VPD83_DESC,           /* xx Size of structure                         */
*PINQ_VPD83_DESC;

/* Extended Inquiry Data (SPC) */

typedef struct InquiryVPD86Data {
    BYTE    FeatSupp1;      /* 00 Feature support                           */
        #define PG86_FS1_REF_CHK    0x01    /* Device checks reference tag  */
        #define PG86_FS1_APP_CHK    0x02    /* Device checks application tag*/
        #define PG86_FS1_GRD_CHK    0x04    /* Device checks guard tag      */
        #define PG86_FS1_SPT_MASK   0x38    /* Protection level support mask*/
        #define PG86_FS1_SPT_1      0x00    /* Supports level  1            */
        #define PG86_FS1_SPT_12     0x08    /* Supports levels 1,2          */
        #define PG86_FS1_SPT_2      0x10    /* Supports level  2            */
        #define PG86_FS1_SPT_13     0x18    /* Supports levels 1,3          */
        #define PG86_FS1_SPT_3      0x20    /* Supports level  3            */
        #define PG86_FS1_SPT_23     0x28    /* Supports levels 2,3          */
        #define PG86_FS1_SPT_123    0x38    /* Supports levels 1,2,3        */
        #define PG86_FS1_AM_MASK    0xC0    /* Activate uCode mask          */
        #define PG86_FS1_BF_WR_BUFF 0x40    /* Act. uCode before Wr. Buffer */
        #define PG86_FS1_AF_RESET   0x80    /* Act. uCode after reset event */

    BYTE    FeatSupp2;      /* 01 Feature support                           */
        #define PG86_FS2_SIMPLE_Q   0x01    /* Supports simple queue attr   */
        #define PG86_FS2_ORDERED_Q  0x02    /* Supports ordered queue attr  */
        #define PG86_FS2_HEAD_OF_Q  0x04    /* Supports head of queue attr  */
        #define PG86_FS2_CMD_PRRTY  0x08    /* Supports command priority    */
        #define PG86_FS2_GROUP      0x10    /* Supports grouping function   */
        #define PG86_FS2_UA_SKS     0x20    /* Supports unit attn sense key */
                                            /* specific data                */
    BYTE    FeatSupp3;      /* 02 Feature support                           */
        #define PG86_FS3_VOL_CACHE  0x01    /* Supports volatile cache      */
        #define PG86_FS3_NV_CACHE   0x02    /* Supports non-volatile cache  */
        #define PG86_FS3_CORR_DIS   0x04    /* Supports correction-disable  */
        #define PG86_FS3_WR_UNC     0x08    /* Supports write uncorrectable */

    BYTE    FeatSupp4;      /* 03 Feature support                           */
        #define PG86_FS4_LUI_CLEAR  0x01    /* Supports LUN I_T nexus clear */
        #define PG86_FS4_PI_INTVL   0x10    /* Supports T10-DIF intervals   */
        #define PG86_FS4_NO_PI_CHK  0x20    /* No protection info checking  */

    BYTE    FeatSupp5;      /* 04 Feature support                           */
        #define PG86_FS5_CBCS       0x01    /* Supports capability-based    */
                                            /* command security             */
        #define PG86_FS5_HSSRELEF   0x02    /* History snapshots release eff*/
        #define PG86_FS5_REFERRAL   0x10    /* Supports referrals           */

    BYTE    FeatSupp6;      /* 05 Feature support                           */
        #define PG86_FS6_MMD_MASK   0x0F    /* Multi microcode download mask*/

    WORD    ExtSelfTestComp;/* 06 Number of minutes needed for a self-test  */
    BYTE    FeatSupp7;      /* 08 Feature support                           */
        #define PG86_FS7_VS_ACT     0x20    /* Supports vend spec activation*/
        #define PG86_FS7_HR_ACT     0x40    /* Supports hard reset act.     */
        #define PG86_FS7_PO_ACT     0x80    /* Supports power on activation */

    BYTE    MaxSenseLen;    /* 09 Maximum sense data length                 */
    BYTE    Reserved[50];   /* 0A Reserved                                  */
} INQ_VPD86_DATA,           /* 3C Size of structure                         */
*PINQ_VPD86_DATA;

/* Mode Policy (SPC) */

typedef struct InquiryVPD87Desc {
    BYTE    PageCode;       /* 00 Page code                                 */

        #define PG87_PC_MASK        0x3F    /* Page code mask               */
        #define PG87_PC_ALL         0x3F    /* All pages value              */

    BYTE    SubPageCode;    /* 01 Subpage code                              */
        #define PG87_SPC_ALL        0xFF    /* All pages value              */

    BYTE    PagePolicy;     /* 02 Page policy                               */
        #define PG87_PP_MASK        0x03    /* Page policy mask             */
        #define PG87_PP_SHARED      0x00    /* Shared                       */
        #define PG87_PP_TARG_PORT   0x01    /* Per target port              */
        #define PG87_PP_I_T_NEXUS   0x03    /* Per I_T nexus                */
        #define PG87_PP_MLUS        0x80    /* MLUS flag                    */

    BYTE    Reserved;       /* 03 Reserved                                  */
} INQ_VPD87_DESC,           /* 04 Size of structure                         */
*PINQ_VPD87_DESC;

/* ATA Information (SAT) */

typedef struct InquiryVPD89Data {
    BYTE    Reserved[4];    /* 00 Reserved                                  */
    BYTE    VendorId[8];    /* 04 SAT Vendor identification                 */
    BYTE    ProductId[16];  /* 0C SAT Product identification                */
    BYTE    RevLevel[4];    /* 1C SAT Product revision level                */
    BYTE    RegDth[20];     /* 20 Initial Register DTH (Signature)          */
    BYTE    CmdCode;        /* 34 Command code                              */
    BYTE    Reserved2[3];   /* 35 Reserved                                  */
    BYTE    IdentData[512]; /* 38 IDENTIFY DEVICE data                      */
} INQ_VPD89_DATA,           /*238 Size of structure                         */
*PINQ_VPD89_DATA;

/* Power Condition (SPC) */

typedef struct InquiryVPD8AData {
    BYTE    Flags;          /* 00 Flags                                     */
        #define PGBAF_STANDBY_Z     0x01
        #define PGBAF_STANDBY_Y     0x02

    BYTE    Flags2;         /* 01 Flags                                     */
        #define PGBAF2_IDLE_A       0x01
        #define PGBAF2_IDLE_B       0x02
        #define PGBAF2_IDLE_C       0x04

    WORD    StopRcvryTime;  /* 02 Stopped condition recovery time           */
    WORD    SbZRcvryTime;   /* 04 Standby Z condition recovery time         */
    WORD    SbYRcvryTime;   /* 06 Standby Y condition recovery time         */
    WORD    IdleARcvryTime; /* 08 Idle A condition recovery time            */
    WORD    IdleBRcvryTime; /* 0A Idle B condition recovery time            */
    WORD    IdleCRcvryTime; /* 0C Idle C condition recovery time            */
} INQ_VPD8A_DATA,           /* 0E Size of structure                         */
*PINQ_VPD8A_DATA;

/* Block Limits (SBC) */

typedef struct InquiryVPDB0Data {
    BYTE    Flags;          /* 00 Flags                                     */
        #define PGB0F_WSNZ          0x01    /* Write same no zero           */

    BYTE    MaxCmpWrLen;    /* 01 Max compare and write length              */
    WORD    OptXferLenGran; /* 02 Optimal transfer length granularity       */
    DWORD   MaxXferLen;     /* 04 Max transfer length                       */
    DWORD   OptXferLen;     /* 08 Optimal transfer length                   */
    DWORD   MaxPfXdXferLen; /* 0C Max prefetch XDREAD/WRITE transfer length */
    DWORD   MaxUnmapLbaCnt; /* 10 Max unmap LBA count                       */
    DWORD   MaxUnmapBDCnt;  /* 14 Max unmap block descriptor count          */
    DWORD   OptUnmapGran;   /* 18 Optimal unmap granularity                 */
    DWORD   UnmapGranAlign; /* 1C Unmap granularity alignment               */
        #define PGB0_UNMAP_ALIGN_VAL    0x80000000 
    DWORD   MaxWrSameLen[2];/* 20 Max WRITE SAME length                     */
    DWORD   MaxAtmcXferLen; /* 28 Max atomic transfer length                */
    DWORD   AtmcAlign;      /* 2C Atomic alignment                          */
    DWORD   AtmcXferLenGran;/* 30 Atomic transfer length granularity        */
    DWORD   MaxAtlWithBound;/* 34 Max atomic transfer length with boundary  */
    DWORD   MaxAtmcBoundSz; /* 38 Max atomic boundary size                  */
} INQ_VPDB0_DATA,           /* 3C Size of structure                         */
*PINQ_VPDB0_DATA;

/* Block Device Characteristics (SBC) */

typedef struct InquiryVPDB1Data {
    WORD    MediumRpm;      /* 00 Medium rotation rate                      */

        #define PGB1_MR_NOT_SUPP    0x0000  /* Not supported                */
        #define PGB1_MR_SOLID_STATE 0x0001  /* Solid state disk             */
        #define PGB1_MR_RPM_START   0x0401  /* Start of RPM values          */
        #define PGB1_MR_RPM_END     0xFFFE  /* End of RPM values            */

    BYTE    ProductType;    /* 02 Product type                              */
        #define PGB1_PT_NOT_INDICATED 0x00

    BYTE    FormFactor;     /* 03 Nominal form factor                       */
        #define PGB1_FF_MASK        0x0F    /* Form factor mask             */
        #define PGB1_FF_NOT_SUPP    0x00    /* Not supported                */
        #define PGB1_FF_5_25        0x01    /* 5.25 inch                    */
        #define PGB1_FF_3_5         0x02    /* 3.5 inch                     */
        #define PGB1_FF_2_5         0x03    /* 2.5 inch                     */
        #define PGB1_FF_1_8         0x04    /* 1.8 inch                     */
        #define PGB1_FF_LT_1_8      0x05    /* less than 1.8 inch           */
        #define PGB1_WACEREQ_MASK   0x30    /* Write after crypt erase reqd */
        #define PGB1_WACEREQ_UNK    0x00    /* Not supported                */
        #define PGB1_WACEREQ_GOOD   0x10    /* Return GOOD sts, data random */
        #define PGB1_WACEREQ_CRC    0x20    /* Return CRC/ECC error         */
        #define PGB1_WACEREQ_WASR   0x30    /* Return ASC-0x11 ASCQ-0x15    */
        #define PGB1_WABEREQ_MASK   0xC0    /* Write after block erase reqd */
        #define PGB1_WABEREQ_UNK    0x00    /* Not supported                */
        #define PGB1_WABEREQ_GOOD   0x40    /* Return GOOD sts, data random */
        #define PGB1_WABEREQ_CRC    0x80    /* Return CRC/ECC error         */
        #define PGB1_WABEREQ_WASR   0xC0    /* Return ASC-0x11 ASCQ-0x15    */

    BYTE    Flags;          /* 04 Flags                                     */
        #define PGB1F_VBULS         0x01    /* Verify byte check unmapped   */
                                            /* LBA supported                */
        #define PGB1F_FUAB          0x02    /* FUA behavior                 */
        #define PGB1F_BOCS          0x04    /* Background operation control */
        #define PGB1F_RBWZ          0x08    /* Reassign blocks write zero   */
        #define PGB1F_ZONED_MASK    0x30    /* Zoned block capabilities     */
        #define PGB1F_ZONED_UNK     0x00    /* Not reported                 */
        #define PGB1F_ZONED_HOST    0x10    /* Host aware zoned cap.        */
        #define PGB1F_ZONED_DEVICE  0x20    /* Device managed zoned cap.    */

    BYTE    Reserved2[55];  /* 05 Reserved                                  */
} INQ_VPDB1_DATA,           /* 3C Size of structure                         */
*PINQ_VPDB1_DATA,
  /* Legacy */
  INQ_VPDB1_DESC,           
*PINQ_VPDB1_DESC;

/* Logical Block Provisioning (SBC) */

typedef struct InquiryVPDB2Data {
    BYTE    ThresholdExp;   /* 00 Threshold exponent                        */
    BYTE    Flags;          /* 01 Flags                                     */
        #define PGB2F_DP        0x01  /* Descriptor present                 */
        #define PGB2F_ANC_SUP   0x02  /* Anchor supported                   */
        #define PGB2F_LBPRZ     0x04  /* Read zeros                         */
        #define PGB2F_LBPWS10   0x20  /* WRITE SAME 10 supported            */
        #define PGB2F_LBPWS16   0x40  /* WRITE SAME 16 supported            */
        #define PGB2F_LBPU      0x80  /* UNMAP supported                    */

    BYTE    ProvType;       /* 02 Provisioning type                         */
        #define PGB2_PT_MASK    0x07  /* Provisioning type mask             */
        #define PGB2_PT_NONE    0x00  /* No provisioning                    */
        #define PGB2_PT_RSRC    0x01  /* Resource provisioned               */
        #define PGB2_PT_THIN    0x02  /* Thin provisioned                   */

    BYTE    Reserved;       /* 03 Reserved                                  */
} INQ_VPDB2_DATA,           /* 04 Size of structure                         */
*PINQ_VPDB2_DATA;


/* Zoned Block Device Characteristics (ZBC) */

typedef struct InquiryVPDB6Data {
    BYTE    Flags;          /* 00 Flags                                     */
        #define PGB1F_URSWRZ        0x01    /* Not supported                */

    BYTE    Reserved[3];    /* 01 Reserved                                  */
    DWORD   OpenSwpZones;   /* 04 Optimal num. of open SWP zones            */
    DWORD   NswSwpZones;    /* 08 Optimal num. of non-seq. written SWP zones*/
    DWORD   OpenSwrZones;   /* 0C Max num. of open SWR zones                */
    BYTE    Reserved2[0x2C];/* 10 Reserved                                  */
} INQ_VPDB6_DATA,           /* 3C Size of structure                         */
*PINQ_VPDB6_DATA;


/****************************************************************************
                      READ CAPACITY COMMAND AND DATA
*****************************************************************************/

typedef struct ReadCapacity10 {
    BYTE    Command;                    /* SCSI_RD_CAPAC (0x25)             */
    BYTE    Reserved1;
    BYTE    Block[4];
    BYTE    Reserved2[2];
    BYTE    Flags;
        #define RCF_PMI     0x01        /* Partial medium indicator         */

    BYTE    Control;
} CDB_READ_CAP10, *PCDB_READ_CAP10;

typedef struct Capacity {
    DWORD   Block;                      /* Last logical block number        */
    DWORD   Bytes;                      /* Block length in bytes            */
} CAPACITY, CAPACITY10, *PCAPACITY, *PCAPACITY10;

typedef struct ReadCapacity16 {
    BYTE    Command;                    /* SCSI_RD_CAPAC16 (0x9E)           */
    BYTE    ServiceAction;
        #define RDCAP_SERV_ACTION   0x10
        #define RDCAP_SERV_MASK     0x1F

    BYTE    Block[8];
    BYTE    Length[4];
    BYTE    Flags;
        /* #define RCF_PMI     0x01        Partial medium indicator         */

    BYTE    Control;
} CDB_READ_CAP16, *PCDB_READ_CAP16;

typedef struct Capacity16 {
    QWORD   Block;                      /* Last logical block number        */
    DWORD   Bytes;                      /* Block length in bytes            */
    BYTE    Flags;                      /* Protection flags                 */
        #define RDCAPF_PROT_EN      0x01    /* DIF protection enabled       */
        #define RDCAPF_P_TYPE_MASK  0x0E    /* DIF protection type mask     */
        #define RDCAPF_P_TYPE_SHIFT 1       /* DIF protection type shift    */
        #define RDCAPF_P_TYPE_WGHT  0x01    /* DIF protection type weight   */
        #define RDCAPF_P_TYPE_1     0x00    /* DIF protection level 1       */
        #define RDCAPF_P_TYPE_2     0x02    /* DIF protection level 2       */
        #define RDCAPF_P_TYPE_3     0x04    /* DIF protection level 3       */
        #define RDCAPF_BASIS_MASK   0x30    /* Read capacity basis mask     */
        #define RDCAPF_BASIS_NO_SWR 0x00    /* Conventional and SWP only    */
        #define RDCAPF_BASIS_ALL    0x10    /* All LBAs including SWR       */

    BYTE    BlockExp;                   /* Logical to phys. exponent        */
        #define RDCAPBE_PI_EXP_MASK 0xF0    /* DIF PI exponent mask         */
        #define RDCAPBE_PI_EXP_SHFT 0x04    /* DIF PI exponent shift        */
        #define RDCAPBE_LBPPBE_MASK 0x0F    /* Log blocks/phys block exp    */

    WORD    LowestLba;                  /* Lowest aligned LBA               */
        #define RDCAPLL_LBPME       0x8000  /* Provisioning management      */
        #define RDCAPLL_LBPRZ       0x4000  /* Provisioning read zeroes     */

    BYTE    Rsrvd[16];
} CAPACITY16, *PCAPACITY16;


/****************************************************************************
                          START/STOP UNIT COMMAND
*****************************************************************************/

typedef struct StartUnit {
    BYTE    Command;                        /* SCSI_START_STP (0x1B)        */
    BYTE    Flags1;
        #define SUF_IMMEDIATE       0x01    /* Return before operation comp */

    BYTE    Reserved1;
    BYTE    PwrCondMod;
        #define SU_PCM_MASK         0x0F    /* Power condition modifier     */
        #define SU_PCM_IDLE_A       0x00
        #define SU_PCM_IDLE_B       0x01
        #define SU_PCM_IDLE_C       0x02
        #define SU_PCM_SB_Z         0x00
        #define SU_PCM_SB_Y         0x01

    BYTE    Flags;
        #define SUF_START               0x01    /* Start                    */
        #define SUF_LOAD_EJECT          0x02    /* Load/Eject               */
        #define SUF_NO_FLUSH            0x04    /* No Flush                 */
        #define SUF_PWR_COND_MASK       0xF0    /* Power condition mask     */
        #define SUF_PWR_COND_START_STOP 0x00    /* Start/stop power cond.   */
        #define SUF_PWR_COND_ACTIVE     0x10    /* Active power condition   */
        #define SUF_PWR_COND_IDLE       0x20    /* Idle power condition     */
        #define SUF_PWR_COND_STANDBY    0x30    /* Standby power condition  */
        #define SUF_PWR_COND_LU_CTRL    0x70    /* Start idle/standby timers*/
        #define SUF_PWR_COND_IDLE0      0xA0    /* Force idle timer to 0    */
        #define SUF_PWR_COND_STANDBY0   0xB0    /* Force standby timer to 0 */

        #define STRT_Start          0x01    /* Old definition               */
        #define STRT_LoEj           0x02    /* Old definition               */

    BYTE    Control;
} CDB_START_UNIT, *PCDB_START_UNIT;


/****************************************************************************
                            FORMAT UNIT COMMAND
*****************************************************************************/

typedef struct FormatUnit {
    BYTE    Command;                    /* SCSI_FORMAT (0x04)               */
    BYTE    Flags;
        #define FMTUF_DL_FMT_MASK   0x07    /* Defect list format           */
        #define FMTUF_DL_FMT_SHORT  0x00    /* Short block                  */
        #define FMTUF_DL_FMT_LONG   0x03    /* Long block                   */
        #define FMTUF_DL_FMT_INDEX  0x04    /* Bytes from index             */
        #define FMTUF_DL_FMT_PHYS   0x05    /* Physical sector              */
        #define FMTUF_DL_FMT_VENDOR 0x06    /* Vendor specific              */
        #define FMTUF_CMPLIST       0x08    /* Complete list                */
        #define FMTUF_FMTDATA       0x10    /* Format parameter list        */
        #define FMTUF_LONGLIST      0x20    /* Long parameter list header   */
        #define FMTUF_RTO_REQ       0x40    /* Reference tag own request    */
        #define FMTUF_FMTPINFO      0x80    /* Format protection info       */

    BYTE    VendorSpecific;
    BYTE    Obsolete[2];
    BYTE    Control;
} CDB_FORMAT_UNIT, *PCDB_FORMAT_UNIT;


typedef struct FormatUnitShortHdr {
    BYTE    Flags;
        #define FMTUPHF_PROT_USE_MASK  0x07 /* Protection field usage       */

    BYTE    Flags2;
        #define FMTUPHF2_IMMED         0x02 /* Immediate response           */
        #define FMTUPHF2_IP            0x08 /* Initialization pattern       */
        #define FMTUPHF2_STPF          0x10 /* Stop format                  */
        #define FMTUPHF2_DCRT          0x20 /* Disable certification        */
        #define FMTUPHF2_DPRY          0x40 /* Disable primary              */
        #define FMTUPHF2_FOV           0x80 /* Format options valid         */

    WORD    DefectListLen;
} FMTU_PARAM_HDR_SH, *PFMTU_PARAM_HDR_SH;


typedef struct FormatUnitLongHdr {
    BYTE    Flags;
    BYTE    Flags2;
    BYTE    Reserved;
    BYTE    ProtectInfo;
        #define FMTUPH_PI_INFO_MASK 0xF0    /* Information mask             */
        #define FMTUPH_PI_EXP_MASK  0x0F    /* Interval exponent mask       */

    DWORD   DefectListLen;
} FMTU_PARAM_HDR_LG, *PFMTU_PARAM_HDR_LG;


typedef struct FormatUnitInitPatternHdr {
    BYTE    Flags;
        #define FMTUIPF_IP_MOD_MASK 0xC0    /* IP modifier mask             */
        #define FMTUIPF_IP_MOD_NONE 0x00    /* No modification              */
        #define FMTUIPF_IP_MOD_LOG  0x40    /* Write LBA to logical block   */
        #define FMTUIPF_IP_MOD_PHYS 0x80    /* Write LBA to physical block  */
        #define FMTUIPF_IP_MOD_RSVD 0xC0    /* Reserved                     */
        #define FMTUIPF_SEC_INIT    0x20    /* Security initialize          */

    BYTE    PatternType;
        #define FMTUIP_PT_DEFAULT   0x00    /* Use default pattern          */
        #define FMTUIP_PT_REPEAT    0x01    /* Repeat pattern to fill block */

    WORD    PatternLen;
} FMTU_INIT_PATT_HDR, *PFMTU_INIT_PATT_HDR;


/****************************************************************************
                         READ/WRITE/VERIFY COMMANDS
*****************************************************************************/

typedef struct ReadWrite6 {
    BYTE    Command;                    /* SCSI_READ6   (0x08)              */
                                        /* SCSI_WRITE6  (0x0A)              */
    BYTE    Block[3];
        #define RW6_LBA_MASK    0x1F 

    BYTE    Length;
    BYTE    Control;
} CDB_READ6, CDB_WRITE6, *PCDB_READ6, *PCDB_WRITE6;


typedef struct ReadWriteVerify10 {
    BYTE    Command;                    /* SCSI_READ10      (0x28)          */
                                        /* SCSI_WRITE10     (0x2A)          */
                                        /* SCSI_VERIFY      (0x2F)          */
                                        /* SCSI_WRTVERIFY   (0x2E)          */
                                        /* SCSI_WRTVERIFY   (0x2E)          */
                                        /* SCSI_WRITE_SAME  (0x41)          */
                                        /* SCSI_READ_LONG   (0x3E)          */
                                        /* SCSI_WRITE_LONG  (0x3F)          */
    BYTE    Flags;
        #define RDWRF_NDOB              0x01
        #define RDWRF_FUA_NV            0x02
        #define RDWRF_BYTE_CHK          0x02
        #define RDWRF_LBDATA            0x02
        #define RDWRF_PBDATA            0x04
        #define RDWRF_FUA               0x08
        #define RDWRF_UNMAP             0x08
        #define RDWRF_ANCHOR            0x10
        #define RDWRF_WRT_PBLOCK        0x20
        #define RDWRF_WRT_UNC           0x40
        #define RDWRF_COR_DIS           0x80

        #define RDWRF_PROTECT           0xE0    /* T10-DIF */
        #define RDWRF_PROTECT_NONE      0x00
        #define RDWRF_PROTECT_ALL       0x20
        #define RDWRF_PROTECT_NO_GRD    0x40
        #define RDWRF_PROTECT_NO_CHK    0x60
        #define RDWRF_PROTECT_GRD_ONLY  0x80
        #define RDWRF_PROTECT_ALL_ATO   0xA0


    BYTE    Block[4];
    BYTE    GroupNumber;
    BYTE    Length[2];
    BYTE    Control;
} CDB_READ10,    *PCDB_READ10,
  CDB_WRITE10,   *PCDB_WRITE10,
  CDB_VERIFY10,  *PCDB_VERIFY10,
  CDB_WRTVER10,  *PCDB_WRTVER10,
  CDB_WRTSAME10, *PCDB_WRTSAME10,
  CDB_RDLONG10,  *PCDB_RDLONG10,
  CDB_WRTLONG10, *PCDB_WRTLONG10;


typedef struct ReadWriteVerify12 {
    BYTE    Command;                    /* SCSI_READ12      (0xA8)          */
                                        /* SCSI_WRITE12     (0xAA)          */
                                        /* SCSI_VERIFY12    (0xAF)          */
                                        /* SCSI_WRTVERIFY12 (0xAE)          */
    BYTE    Flags;
    BYTE    Block[4];
    BYTE    Length[4];
    BYTE    GroupNumber;
    BYTE    Control;
} CDB_READ12,   *PCDB_READ12,
  CDB_WRITE12,  *PCDB_WRITE12,
  CDB_VERIFY12, *PCDB_VERIFY12,
  CDB_WRTVER12, *PCDB_WRTVER12;


typedef struct ReadWriteVerify16 {
    BYTE    Command;                    /* SCSI_READ16       (0x88)         */
                                        /* SCSI_WRITE16      (0x8A)         */
                                        /* SCSI_VERIFY12     (0x8F)         */
                                        /* SCSI_WRTVERIFY12  (0x8E)         */
                                        /* SCSI_WRITE_SAME16 (0x93)         */
                                        /* SCSI_READ_LONG16  (0x9E)         */
                                        /* SCSI_WRITE_LONG16 (0x9F)         */
    BYTE    Flags;
        #define READLONG16_SERV_ACTION  0x11
        #define WRITELONG16_SERV_ACTION 0x11

    BYTE    Block[8];
    BYTE    Length[4];
    BYTE    GroupNumber;
    BYTE    Control;
} CDB_READ16,    *PCDB_READ16,
  CDB_WRITE16,   *PCDB_WRITE16,
  CDB_VERIFY16,  *PCDB_VERIFY16,
  CDB_WRTVER16,  *PCDB_WRTVER16,
  CDB_WRTSAME16, *PCDB_WRTSAME16,
  CDB_WRTLONG16, *PCDB_WRTLONG16;


typedef struct CompareAndWrite {
    BYTE    Command;
    BYTE    Flags;
    BYTE    Block[8];
    BYTE    Reserved[3];
    BYTE    Length;
    BYTE    GroupNumber;
    BYTE    Control;
} CDB_COMP_WRITE, *PCDB_COMP_WRITE;


/****************************************************************************
                          SYNCHRONIZE CACHE COMMANDS
*****************************************************************************/

typedef struct SyncCache10 {
    BYTE    Command;                    /* SCSI_SYNC_CACHE  (0x35)          */
    BYTE    Flags;
        #define SCF_SYNC_NV     0x04    /* Sync non-volatile                */
        #define SCF_IMMEDIATE   0x02    /* Return before operation complete */

    BYTE    Block[4];
    BYTE    GroupNumber;
    BYTE    NumBlocks[2];
    BYTE    Control;
} CDB_SYNC_CACHE10, *PCDB_SYNC_CACHE10;


typedef struct SyncCache16 {
    BYTE    Command;                    /* SCSI_SYNC_CACHE16 (0x91)         */
    BYTE    Flags;                      /* See Sync cache 10 command        */
    BYTE    Block[8];
    BYTE    NumBlocks[4];
    BYTE    GroupNumber;
    BYTE    Control;
} CDB_SYNC_CACHE16, *PCDB_SYNC_CACHE16;


/****************************************************************************
                     LOG SENSE/SELECT COMMAND AND DATA
*****************************************************************************/

#define LOG_PAGE_SUPPORTED_PAGES        0x00
#define LOG_PAGE_READ_ERR_COUNTERS      0x03
#define LOG_PAGE_TEMPERATURE            0x0D
#define LOG_PAGE_SELF_TEST_RESULTS      0x10
#define LOG_PAGE_SSD                    0x11
#define LOG_PAGE_PROTOCOL_SPECIFIC_PORT 0x18
#define LOG_PAGE_INFO_EXCEPTIONS        0x2F

typedef struct ScsiLogSense {
    BYTE    Command;                    /* SCSI_LOG_SENSE (0x4D)            */
    BYTE    Flags;
        #define LSNF_SP           0x01  /* Save parameters                  */
        #define LSNF_PPC          0x02  /* Parameter pointer control        */

    BYTE    PageCode;
        #define LSNPC_CODE_MASK   0x3F  /* Page code mask                   */
        #define LSNPC_CTRL_MASK   0xC0  /* Page code control mask           */
        #define LSNPCC_THRESHOLD  0x00  /* Threshold values                 */
        #define LSNPCC_CUMULATIVE 0x40  /* Cumulative values                */
        #define LSNPCC_DEFAULT_TH 0x80  /* Default threshold values         */
        #define LSNPCC_DEFAULT_CU 0xC0  /* Default cumulative values        */

    BYTE    SubPage;                    /* Added in SPC-3                   */
    BYTE    Reserved;
    BYTE    ParamPtr[2];
    BYTE    Length[2];
    BYTE    Control;
} CDB_LOG_SENSE, *PCDB_LOG_SENSE;

typedef struct ScsiLogSelect {
    BYTE    Command;                    /* SCSI_LOG_SENSE (0x4C)            */
    BYTE    Flags;
        #define LSLF_SP           0x01  /* Save parameters                  */
        #define LSLF_PCR          0x02  /* Parameter code reset             */

    BYTE    PageCode;
        #define LSLPC_CODE_MASK   0x3F  /* Page code mask                   */
        #define LSLPC_CTRL_MASK   0xC0  /* Page code control mask           */
        #define LSLPCC_CUMULATIVE 0x40  /* Cumulative values                */
        #define LSLPCC_DEFAULT_CU 0xC0  /* Default cumulative values        */

    BYTE    SubPage;                    /* Added in SPC-3                   */
    BYTE    Reserved[3];
    BYTE    Length[2];
    BYTE    Control;
} CDB_LOG_SELECT, *PCDB_LOG_SELECT;


/* Common Log Page Headers */

typedef struct LogPageHeader {
    BYTE    PageCode;
        #define LPGH_PC_SPF  0x40
        #define LPGH_PC_DS   0x80

    BYTE    SubPage;
    WORD    Length;
} LOG_PAGE_HDR, *PLOG_PAGE_HDR;


typedef struct LogParamHeader {
    WORD    ParamCode;
    BYTE    Flags;
        #define LPMHF_FMT_LINK_MASK 0x03
        #define LPMHF_FL_DATA_CNT0  0x00
        #define LPMHF_FL_ASCII      0x01
        #define LPMHF_FL_DATA_CNT2  0x02
        #define LPMHF_FL_BINARY     0x03
        #define LPMHF_TMC_MASK      0x0C
        #define LPMHF_ETC           0x10
        #define LPMHF_TSD           0x20
        #define LPMHF_DU            0x80

    BYTE    Length;
} LOG_PARAM_HDR, *PLOG_PARAM_HDR;


/* Solid State Media Log Data */

#define LOG_PARAM_CODE_SSD          0x0001

typedef struct SsdLogPage {
    BYTE    Rsvd[3];
    BYTE    PercentUsed;
} SSD_LOG_PAGE, *PSSD_LOG_PAGE;


/* Informational Exceptions Log Data */

typedef struct InfoExceptionsLogPage {
    BYTE    Asc;
    BYTE    Ascq;
    BYTE    Temperature;
        #define IELP_TEMP_UNAVAIL   0xFF

    BYTE    VendorSpecific[1];
} INFO_EXC_LOG_PAGE, *PINFO_EXC_LOG_PAGE;


/* Self Test Log Data */

typedef struct SeltTestLogPage {
    BYTE    Result;
        #define STLP_RSLT_SUCCESS   0x00
        #define STLP_RSLT_ABT_EXP   0x01
        #define STLP_RSLT_ABT_IMP   0x02
        #define STLP_RSLT_UNK_ERR   0x03
        #define STLP_RSLT_SEG_UNK   0x04
        #define STLP_RSLT_SEG_1     0x05
        #define STLP_RSLT_SEG_2     0x06
        #define STLP_RSLT_SEG_NUM   0x07
        #define STLP_RSLT_IN_PROG   0x0F

    BYTE    Number;
    WORD    PowerOnTime;
    BYTE    FailedLba[8];
    BYTE    SenseKey;
    BYTE    Asc;
    BYTE    Ascq;
    BYTE    VendorSpecific;
} SELF_TEST_LOG_PAGE, *PSELF_TEST_LOG_PAGE;

#define NUM_SELF_TEST_LOG_PAGES         0x14


/* Temperature Log Parameter Data */

#define LOG_PARAM_CODE_CURR_TEMP    0x0000  /* Current temperature          */
#define LOG_PARAM_CODE_REF_TEMP     0x0001  /* Reference temperature        */

typedef struct TempLogParam {
    BYTE    Reserved;
    BYTE    Temperature;
        #define TLP_TEMP_INVALID    0xFF    /* Invalid/unsupported temp     */

} TEMP_LOG_PARAM, *PTEMP_LOG_PARAM;


/* Protocol Specific Port Log Data */

typedef struct ProtocolSpecificPortLogParam {
    WORD    ParamCode;
    BYTE    Flags;
    BYTE    Length;
    BYTE    ProtocolId;
        #define PSPLP_PI_MASK   0x0F    /* Protocol Identifier mask         */
        #define PSPLP_PI_FC     0x00    /* Fibre Channel Protocol - FCP     */
        #define PSPLP_PI_SCSI   0x01    /* SCSI Parallel Interface - SPI    */
        #define PSPLP_PI_SSA    0x02    /* Serial Storage Arch. - SSA       */
        #define PSPLP_PI_SBP    0x03    /* Serial Bus Protocol 1394 - SBP   */
        #define PSPLP_PI_RDMA   0x04    /* SCSI RDMA Protocol - SRP         */
        #define PSPLP_PI_ISCSI  0x05    /* Internet SCSI - iSCSI            */
        #define PSPLP_PI_SAS    0x06    /* Serial Attached SCSI - SPL       */
        #define PSPLP_PI_ADT    0x07    /* Automation/Drive Interface - ADT */
        #define PSPLP_PI_ATA    0x08    /* AT Attachment Interface - ACS    */
        #define PSPLP_PI_USB    0x09    /* USB Attached SCSI - UAS          */
        #define PSPLP_PI_PCIE   0x0A    /* SCSI over PCIe - SOP             */
        #define PSPLP_PI_NONE   0x0F    /* No specific protocol             */

} PSP_LOG_PARAM, *PPSP_LOG_PARAM;


typedef struct SasPhyEventDescriptor {
    BYTE    Reserved[3];
    BYTE    EventSource;
    DWORD   EventInfo;
    DWORD   DetThreshold;
} SAS_PHY_EVT_DESC, *PSAS_PHY_EVT_DESC;


typedef struct SasPhyLogDescriptor {
    BYTE    Reserved1;
    BYTE    PhyId;
    BYTE    Reserved2;
    BYTE    Length;                     /* Added in SAS-2                   */

    BYTE    AttachTypeReason;           /* SMP_DEVTYPE_XXX                  */
    BYTE    LinkRate;                   /* SMP_RATE_XXX                     */
    BYTE    AttachIni;                  /* SMPD_AIF_XXX                     */
    BYTE    AttachTgt;                  /* SMPD_ATF_XXX                     */

    SASADDR SasAddr;
    SASADDR AttachSasAddr;

    BYTE    AttachPhyId;
    BYTE    Reserved3[7];

    DWORD   InvDwCnt;
    DWORD   DispErrCnt;
    DWORD   LossOfDwSyncCnt;
    DWORD   PhyResetErrCnt;

    BYTE    Reserved4[2];               /* Added in SAS-2                   */
    BYTE    PhyEventsLength;            /* Added in SAS-2                   */
    BYTE    PhyEventsCnt;               /* Added in SAS-2                   */

    SAS_PHY_EVT_DESC                    /* Added in SAS-2                   */
            PhyEvents[1];
} SAS_PHY_LOG_DESC, *PSAS_PHY_LOG_DESC;


typedef struct ProtocolSpecificPortLogParamSas {
    WORD    ParamCode;                  /* Relative target port identifier  */
    BYTE    Flags;
    BYTE    Length;
    BYTE    ProtocolId;                 /* PSPLP_PI_SAS                     */
    BYTE    Reserved;
    BYTE    Generation;
    BYTE    NumPhys;

    SAS_PHY_LOG_DESC
            PhyLogs[1];
} PSP_LOG_PORT_SAS, *PPSP_LOG_PORT_SAS;


/****************************************************************************
                     MODE SENSE/SELECT COMMAND AND DATA
*****************************************************************************/

#define MODE_PAGE_RD_WR_ERR_RECOVERY    0x01
#define MODE_PAGE_CACHING               0x08
#define MODE_PAGE_CONTROL               0x0A
#define MODE_PAGE_DATA_COMPRESSION      0x0F
#define MODE_PAGE_PROTO_LU              0x18
#define MODE_PAGE_PROTO_PORT            0x19
#define MODE_PAGE_PWR_COND              0x1A
#define MODE_PAGE_INFO_EXC_CTRL         0x1C
#define MODE_PAGE_ALL_PAGES             0x3F

/* Mode Sense CDBs */

typedef struct ModeSense6 {
    BYTE    Command;                    /* SCSI_MODE_SEN6 (0x1A)            */
    BYTE    Flags;
        #define MSNF_DBD        0x08    /* Disable block descriptors        */

    BYTE    PageCode;
        #define MSNPC_CODE_MASK 0x3F    /* Page code mask                   */
        #define MSNPC_CTRL_MASK 0xC0    /* Page code control mask           */
        #define MSNPCC_CURRENT  0x00    /* Current values                   */
        #define MSNPCC_CHANGE   0x40    /* Changeable values                */
        #define MSNPCC_DEFAULT  0x80    /* Default values                   */
        #define MSNPCC_SAVED    0xC0    /* Saved values                     */

    BYTE    SubPage;                    /* Added in SPC-3                   */
        #define MSNSP_ALL       0xFF    /* All subpages                     */

    BYTE    Length;
    BYTE    Control;
} CDB_MODE_SENSE6, *PCDB_MODE_SENSE6;


typedef struct ModeSense10 {
    BYTE    Command;                    /* SCSI_MODE_SEN10 (0x5A)           */
    BYTE    Flags;                      /* See Mode Sense 6                 */
        #define MSNF_LLBAA      0x10    /* Long LBA Accepted                */

    BYTE    PageCode;                   /* See Mode Sense 6                 */
    BYTE    SubPage;                    /* Added in SPC-3                   */
    BYTE    Reserved[3];
    BYTE    Length[2];
    BYTE    Control;
} CDB_MODE_SENSE10, *PCDB_MODE_SENSE10;


/* Mode Select CDBs */

typedef struct ModeSelect6
{
    BYTE    Command;                    /* SCSI_MODE_SEL6 (0x15)            */
    BYTE    Flags;
        #define MSLF_PF         0x10    /* Page Format                      */
        #define MSLF_SP         0x01    /* Save Pages                       */
    
    BYTE    Reserved[2];
    BYTE    Length;                     /* Parameter List Length            */
    BYTE    Control;
} CDB_MODE_SELECT6, *PCDB_MODE_SELECT6;


typedef struct ModeSelect10
{
    BYTE    Command;                    /* SCSI_MODE_SEL10 (0x55)           */
    BYTE    Flags;                      /* See Mode Select 6                */
    BYTE    Reserved[5];
    BYTE    Length[2];                  /* Parameter List Length            */
    BYTE    Control;
} CDB_MODE_SELECT10, *PCDB_MODE_SELECT10;


/* Mode page headers */

typedef struct MpHeader6 {
    BYTE    Length;                     /* Mode data length less this field */
    BYTE    MediumType;                 /* Medium type - device specific    */
        #define MPHMT_BLOCK     0x00    /* Block devices                    */

    BYTE    DeviceSpecific;             /* Device specific                  */
        #define MPHDS_BLOCK_WP  0x80    /* Write protected                  */
        #define MPHDS_BLOCK_DF  0x10    /* DFO and FUA supported            */

    BYTE    BlockDescLen;               /* Length of block descriptors      */
} MODE_PAGE_HDR6, *PMODE_PAGE_HDR6;

typedef struct MpHeader10 {
    WORD    Length;                     /* Mode data length less this field */
    BYTE    MediumType;                 /* Medium type - device specific    */
    BYTE    DeviceSpecific;             /* Device specific                  */
    BYTE    Flags;                      /* Flags                            */
        #define MPHF_LLBA       0x01    /* Long LBA                         */
    BYTE    Reserved;                   /* Reserved                         */
    WORD    BlockDescLen;               /* Length of block descriptors      */
} MODE_PAGE_HDR10, *PMODE_PAGE_HDR10;


/* Mode page block descriptors */

typedef struct MpBlockDescSlba {
    union {
        struct {
            BYTE DensityCode;           /* For non-block devices            */
            BYTE NumBlocks3[3];
        } s;
        DWORD NumBlocks;                /* For block devices                */
    }u;

    BYTE    Reserved;
    BYTE    BlockLen[3];
} MP_BLOCK_DESC_SLBA, *PMP_BLOCK_DESC_SLBA;

typedef struct MpBlockDescLlba {
    QWORD   NumBlocks;
    BYTE    DensityCode;                /* For non-block devices            */
    BYTE    Reserved[3];
    DWORD   BlockLen;
} MP_BLOCK_DESC_LLBA, *PMP_BLOCK_DESC_LLBA;


/* Mode pages */

#define PAGE_CODE_PS_FLAG   0x80
#define PAGE_CODE_SPF_FLAG  0x40
#define PAGE_CODE_MASK      0x3F

typedef struct MpHeaderPage0 {
    BYTE    PageCode;
    BYTE    PageLength;
} MODE_PAGE0_HDR, *PMODE_PAGE0_HDR;

typedef struct MpHeaderSubPage {
    BYTE    PageCode;
    BYTE    SubPage;
    WORD    PageLength;
} MODE_SUBPAGE_HDR, *PMODE_SUBPAGE_HDR;


/* Caching mode page */

typedef struct MpCaching{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    Flags;
        #define MPCACHEF_RCD    0x01    /* Read cache disable               */
        #define MPCACHEF_MF     0x02    /* Multiplication factor            */
        #define MPCACHEF_WCE    0x04    /* Write cache enabled              */
        #define MPCACHEF_SIZE   0x08    /* Size choice field                */
        #define MPCACHEF_DISC   0x10    /* Discontinuity allowed            */
        #define MPCACHEF_CAP    0x20    /* Caching analysis permitted       */
        #define MPCACHEF_ABPF   0x40    /* Abort pre-fetch enabled          */
        #define MPCACHEF_IC     0x80    /* Initiator control of caching     */

    BYTE    Retention;
        #define MPCACHERET_WR_MASK  0x0F    /* Write retention priority     */
        #define MPCACHERET_WR_NONE  0x00    /* No difference                */
        #define MPCACHERET_WR_LO    0x01    /* Write data has low priority  */
        #define MPCACHERET_WR_HI    0x0F    /* Write data has high priority */
        #define MPCACHERET_RD_MASK  0xF0    /* Read retention priority      */
        #define MPCACHERET_RD_NONE  0x00    /* No difference                */
        #define MPCACHERET_RD_LO    0x10    /* Read data has low priority   */
        #define MPCACHERET_RD_HI    0xF0    /* Read data has high priority  */

    WORD    DisPrefetchLen;
    WORD    MinPrefetch;
    WORD    MaxPrefetch;
    WORD    MaxPrefetchMax;
    BYTE    Flags1;
        #define MPCACHEF1_NVDIS     0x01    /* Disable non-volatile cache   */
        #define MPCACHEF1_SCP_MASK  0x06    /* Sync Cache progress mask     */
        #define MPCACHEF1_SCP_NONE  0x00    /* SCP not reported             */
        #define MPCACHEF1_SCP_FULL  0x02    /* SCP reported w/full operation*/
        #define MPCACHEF1_SCP_PART  0x04    /* SCP reported w/partial oper. */
        #define MPCACHEF1_DRA       0x20    /* Disable read ahead           */
        #define MPCACHEF1_LBCSS     0x40    /* Logical block cache seg size */
        #define MPCACHEF1_FSW       0x80    /* Force sequential write       */

    BYTE    NumCacheSegs;
    WORD    CacheSegLen;
    DWORD   Reserved;
} CACHING_MODE_PAGE, *PCACHING_MODE_PAGE;

/* Control mode page */

typedef struct MpControl{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    Flags;
        #define MPCTRLF_RLEC            0x01
        #define MPCTRLF_GLTSD           0x02
        #define MPCTRLF_D_SENSE         0x04
        #define MPCTRLF_DPICZ           0x08
        #define MPCTRLF_TMF_ONLY        0x10
        #define MPCTRLF_TST_MASK        0xE0
        #define MPCTRLF_TST_ONE         0x00
        #define MPCTRLF_TST_SEPARATE    0x20

    BYTE    Queuing;
        #define MPCTRL_QERR_MASK        0x06 
        #define MPCTRL_NUAR             0x08 
        #define MPCTRL_QAM_MASK         0xF0 
        #define MPCTRL_QAM_RESTRICTED   0x00 
        #define MPCTRL_QAM_UNRESTRICTED 0x10 

    BYTE    Flags1;
        #define MPCTRLF1_SWP            0x08
        #define MPCTRLF1_UA_CTRL_MASK   0x30
        #define MPCTRLF1_RAC            0x40
        #define MPCTRLF1_VS             0x80

    BYTE    Flags2;
        #define MPCTRLF2_AUTOLOAD_MASK  0x07
        #define MPCTRLF2_SBLP           0x80
        #define MPCTRLF2_RWWP           0x10
        #define MPCTRLF2_ATMPE          0x20
        #define MPCTRLF2_TAS            0x40
        #define MPCTRLF2_ATO            0x80

    WORD    Reserved;
    WORD    BusyTimeout;
        #define MPCTRL_BT_UNLIMITED     0xFFFF

    WORD    SelfTestTime;
} CONTROL_MODE_PAGE, *PCONTROL_MODE_PAGE;

/* Control Extension mode page */

typedef struct MpControlExt{
    BYTE    PageCode;
    BYTE    SubPage;
    WORD    PageLength;
    BYTE    Flags;
        #define MPCEF_IALUAE            0x01
        #define MPCEF_SCSIP             0x02
        #define MPCEF_TCMOS             0x04
        #define MPCEF_DLC               0x08

    BYTE    InitCmdPri;
        #define MPCE_ICP_MASK           0x0F 

    BYTE    MaxSenseLen;
    BYTE    Reserved[25];
} CONTROL_EXT_MODE_PAGE, *PCONTROL_EXT_MODE_PAGE;

/* Read Write Error Recovery mode page */

typedef struct MpReadWriteErrorRecovery{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    Flags;
        #define MPRWERF_DTE             0x02
        #define MPRWERF_PER             0x04
        #define MPRWERF_RC              0x10
        #define MPRWERF_TB              0x20
        #define MPRWERF_ARRE            0x40
        #define MPRWERF_AWRE            0x80

    BYTE    ReadRetryCnt;
    BYTE    Reserved[3];
    BYTE    Flags2;
        #define MPRWERF_LBPERE          0x80
        #define MPRWERF_MWR_MASK        0x60
        #define MPRWERF_MWR_DIS         0x00
        #define MPRWERF_MWR_ENB         0x20
        #define MPRWERF_MWR_TERM        0x40

    BYTE    WriteRetryCnt;
    BYTE    Reserved1;
    WORD    RcvrTimeLimit;
} RD_WR_ERR_RCVR_MODE_PAGE, *PRD_WR_ERR_RCVR_MODE_PAGE;

/* Informational Exceptions Control mode page */

typedef struct MpInfoExceptionsControl{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    Flags;
        #define MPIECF_LOGERR       0x01
        #define MPIECF_EBACKERR     0x02
        #define MPIECF_TEST         0x04
        #define MPIECF_DIS_EXCEPT   0x08
        #define MPIECF_EWASC        0x10
        #define MPIECF_EBF          0x20
        #define MPIECF_PERF         0x80

    BYTE    RptMethod;
        #define MPIEC_RM_MASK       0x0F
        #define MPIEC_RM_NONE       0x00
        #define MPIEC_RM_UA         0x02
        #define MPIEC_RM_C_RCVD_ERR 0x03
        #define MPIEC_RM_U_RCVD_ERR 0x04
        #define MPIEC_RM_NO_SENSE   0x05
        #define MPIEC_RM_ON_DEMAND  0x06

    DWORD   IntervalTimer;
    DWORD   ReportCnt;
} INFO_EXC_CTRL_MODE_PAGE, *PINFO_EXC_CTRL_MODE_PAGE;

/* Power Condition mode page */

typedef struct MpPowerCondition{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    Flags;
        #define MPPCF_STANDBY_Y     0x01
        #define MPPCF_PM_BG_MASK    0xC0
        #define MPPCF_BG_PREC       0x40
        #define MPPCF_PM_PREC       0x80

    BYTE    Flags2;
        #define MPPCF2_STANDBY_Z    0x01
        #define MPPCF2_IDLE_A       0x02
        #define MPPCF2_IDLE_B       0x04
        #define MPPCF2_IDLE_C       0x08

    DWORD   IdleATimer;
    DWORD   SbZTimer;
    DWORD   IdleBTimer;
    DWORD   IdleCTimer;
    DWORD   SbYTimer;
    BYTE    Reserved[15];
    BYTE    CcfFlags;
        #define MPPC_CCF_IDLE_MASK  0xC0
        #define MPPC_CCF_IDLE_DIS   0x40
        #define MPPC_CCF_IDLE_ENB   0x80
        #define MPPC_CCF_SB_MASK    0x30
        #define MPPC_CCF_SB_DIS     0x10
        #define MPPC_CCF_SB_ENB     0x20
        #define MPPC_CCF_STOP_MASK  0x0C
        #define MPPC_CCF_STOP_DIS   0x04
        #define MPPC_CCF_STOP_ENB   0x08

} PWR_COND_MODE_PAGE, *PPWR_COND_MODE_PAGE;

/* ATA Power Condition mode page */

typedef struct MpAtaPowerCondition{
    BYTE    PageCode;
    BYTE    SubPage;
    WORD    PageLength;
    BYTE    Reserved;
    BYTE    Flags;
        #define MPAPCF_APMP         0x01

    BYTE    ApmValue;
    BYTE    Reserved2[9];
} ATA_PWR_COND_MODE_PAGE, *PATA_PWR_COND_MODE_PAGE;

/* Data Compression mode page */

typedef struct DataCompression{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    Flags;
        #define MPDCF_DCE        0x80
        #define MPDCF_DCC        0x40
    
    BYTE    Flags2;
        #define MPDCF2_DDE       0x80
        #define MPDCF2_RED       0x00

    DWORD   CompAlgorithm;
    DWORD   DecompAlgorithm;
    DWORD   Reserved;
} DATA_COMP_MODE_PAGE, *PDATA_COMP_MODE_PAGE;

/* Protocol Specific Logical Unit mode pages */

typedef struct MpProtoLu{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    ProtoId;
        #define MPLU_PROTO_MASK     0x0F    /* Protocol ID mask             */
        #define MPLU_SAS_PROTO_ID   0x06    /* SAS protocol                 */

} PROTO_LU_MODE_PAGE, *PPROTO_LU_MODE_PAGE;

typedef struct MpSasLu{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    ProtoId;
        #define MPSLU_TLR   0x10    /* SAS Transport Layer Retries */

    BYTE    Reserved[5];
} SAS_LU_MODE_PAGE, *PSAS_LU_MODE_PAGE;

/* Protocol Specific Port mode pages */

typedef struct MpProtoPort{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    ProtoId;
        #define MPP_PROTO_MASK      0x0F    /* Protocol ID mask             */
        #define MPP_SAS_PROTO_ID    0x06    /* SAS protocol                 */

} PROTO_PORT_MODE_PAGE, *PPROTO_PORT_MODE_PAGE;

typedef struct MpSasPort{
    BYTE    PageCode;
    BYTE    PageLength;
    BYTE    ProtoId;
    BYTE    Reserved;
    WORD    ITNexusLossTime;
        #define MPSP_ITNL_VENDOR    0x0000  /* Vendor specific              */
        #define MPSP_ITNL_INFINITE  0xFFFF

    WORD    IniRspTimeout;
        #define MPSP_IRT_DISABLE    0x0000

    WORD    RejectToOpenLimit;
        #define MPSP_RTO_VENDOR     0x0000  /* Vendor specific              */

    BYTE    MaxXferRdy;
        #define MPSP_MXR_INFINITE   0x00

    BYTE    Reserved1[5];
} SAS_PORT_MODE_PAGE, *PSAS_PORT_MODE_PAGE;


typedef struct MpSasPhyModeDesc{
    /* See SMPRSP_DISC for definitions of these fields */

    BYTE    Reserved0;
    BYTE    PhyId;
    WORD    Reserved1;
    BYTE    DevType;
    BYTE    NegRate;
    BYTE    IniPort;
    BYTE    TgtPort;
    BYTE    SasAddr[8];
    BYTE    AttachSasAddr[8];
    BYTE    AttachPhyId;
    BYTE    AttachFlags;
    BYTE    AttachFlags2;
    BYTE    Reserved2[5];
    BYTE    MinRate;
    BYTE    MaxRate;
    BYTE    Reserved3[8];
    BYTE    VendSpec[2];
    BYTE    Reserved4[4];
} SAS_PHY_MODE_DESC, *PSAS_PHY_MODE_DESC;


typedef struct MpSasPhyCtrlDiscover{
    BYTE    PageCode;
    BYTE    SubPageCode;
    WORD    PageLength;
    BYTE    Reserved;
    BYTE    ProtoId;
    BYTE    GenCode;
    BYTE    PhyCount;
    SAS_PHY_MODE_DESC PhyModeDesc[1];
} SAS_PHY_CTRL_DISC_MODE_PAGE, *PSAS_PHY_CTRL_DISC_MODE_PAGE;


/****************************************************************************
                       REPORT LUNS COMMAND AND DATA
*****************************************************************************/

typedef struct ReportLuns {
    BYTE    Command;                    /* SCSI_REPORTLUNS (0x0A)           */
    BYTE    Reserved;
    BYTE    ReportType;                 /* Added in SPC-3                   */
        #define RLRT_NORMAL     0x00    /* Report "normal" LUNs             */
        #define RLRT_WELL_KNOWN 0x01    /* Report well known LUNs           */
        #define RLRT_ALL_LUNS   0x02    /* Report all LUNs                  */

    BYTE    Reserved2[3];
    BYTE    Length[4];
    BYTE    Reserved3;
    BYTE    Control;
} CDB_REPORT_LUNS, *PCDB_REPORT_LUNS;

typedef struct ReportLunsHdr {
    DWORD   Length;
    DWORD   Reserved;
} REPORT_LUNS_HDR, *PREPORT_LUNS_HDR;


/****************************************************************************
                 REPORT TARGET PORT GROUPS COMMAND AND DATA
*****************************************************************************/

typedef struct RptTargetPortGroups {
    BYTE    Command;                    /* SCSI_MAINT_IN (0xA3)             */
    BYTE    ServiceAction;              /* Service Action                   */
        #define RTPG_SRVC_ACT   0x0A

    BYTE    Reserved[4];
    BYTE    Length[4];
    BYTE    Reserved2;
    BYTE    Control;
} CDB_RPT_TARG_PORT_GRPS,
*PCDB_RPT_TARG_PORT_GRPS;

typedef struct RtpgDescHdr {
    DWORD Length;                       /* Length of data that follows      */
} RTPG_DESC_HDR, *PRTPG_DESC_HDR;

typedef struct RtpgGroupDesc {
    BYTE    AsymAccState;               /* Asymmetric Access state          */
        #define RTPGGD_AAS_PREFERRED    0x80    /* Preferred port flag      */
        #define RTPGGD_AAS_MASK         0x0F    /* Access state mask        */
        #define RTPGGD_AAS_ACT_OPT      0x00    /* Active/optimized         */
        #define RTPGGD_AAS_ACT_NONOPT   0x01    /* Active/non-optimized     */
        #define RTPGGD_AAS_STANDBY      0x02    /* Standby                  */
        #define RTPGGD_AAS_UNAVAILABLE  0x03    /* Unavailable              */
        #define RTPGGD_AAS_TRANSITION   0x0F    /* Between states           */

    BYTE    Flags;                      /* Flags                            */
        #define RTPGGDF_AO_SUP          0x01    /* Supports Active/opt      */
        #define RTPGGDF_AN_SUP          0x02    /* Supports Active/non-opt  */
        #define RTPGGDF_S_SUP           0x04    /* Supports standby         */
        #define RTPGGDF_U_SUP           0x08    /* Supports unavailable     */
        #define RTPGGDF_T_SUP           0x80    /* Supports transitioning   */

    WORD    GroupNum;                   /* Target port group                */
    BYTE    Reserved;                   /* Reserved                         */
    BYTE    StatusCode;                 /* Status Code                      */
        #define RTPGGD_STS_NONE         0x00    /* No status available      */
        #define RTPGGD_STS_STPG         0x01    /* Set TPG command sent     */
        #define RTPGGD_STS_IMPLICIT     0x02    /* Implicit access          */

    BYTE    VendorSpecific;             /* Vendor specific                  */
    BYTE    PortCount;                  /* Count of port descriptors        */
} RTPG_GROUP_DESC, *PRTPG_GROUP_DESC;

typedef struct RtpgPortDesc {
    WORD    Obsolete;                   /* Obsolete                         */
    WORD    PortNum;                    /* Relative target port identifier  */
} RTPG_PORT_DESC, *PRTPG_PORT_DESC;


/****************************************************************************
                  SET TARGET PORT GROUPS COMMAND AND DATA
*****************************************************************************/

typedef struct SetTargetPortGroups {
    BYTE    Command;                    /* SCSI_MAINT_OUT (0xA4)            */
    BYTE    ServiceAction;              /* Service Action                   */
        #define STPG_SRVC_ACT   0x0A

    BYTE    Reserved[4];
    BYTE    Length[4];
    BYTE    Reserved2;
    BYTE    Control;
} CDB_SET_TARG_PORT_GRPS,
*PCDB_SET_TARG_PORT_GRPS;

typedef struct StpgDescHdr {
    DWORD Reserved;
} STPG_DESC_HDR, *PSTPG_DESC_HDR;

typedef struct StpgGroupDesc {
    BYTE    AsymAccState;               /* Asymmetric Access state          */
        #define STPGGD_AAS_MASK         0x0F    /* Access state mask        */
        #define STPGGD_AAS_ACT_OPT      0x00    /* Active/optimized         */
        #define STPGGD_AAS_ACT_NONOPT   0x01    /* Active/non-optimized     */
        #define STPGGD_AAS_STANDBY      0x02    /* Standby                  */
        #define STPGGD_AAS_UNAVAILABLE  0x03    /* Unavailable              */

    BYTE    Reserved;                   /* Reserved                         */
    WORD    GroupNum;                   /* Target port group                */
} STPG_GROUP_DESC, *PSTPG_GROUP_DESC;


/****************************************************************************
                 SEND/RECEIVE DIAGNOSIC COMMANDS AND DATA
*****************************************************************************/

#define DIAG_PAGE_SUPPORTED             0x00
#define DIAG_PAGE_CONFIG                0x01
#define DIAG_PAGE_ENC_CTRL_STS          0x02
#define DIAG_PAGE_HELP_TEXT             0x03
#define DIAG_PAGE_STRING_IO             0x04
#define DIAG_PAGE_THRESHOLD_IO          0x05
#define DIAG_PAGE_ELEMENT_DESC          0x07
#define DIAG_PAGE_SHORT_ENC_STS         0x08
#define DIAG_PAGE_ENC_BUSY              0x09
#define DIAG_PAGE_ADD_ELEMENT_STS       0x0A
#define DIAG_PAGE_SUBENC_HELP_TEXT      0x0B
#define DIAG_PAGE_SUBENC_STRING_IO      0x0C
#define DIAG_PAGE_SES_SUPPORTED         0x0D
#define DIAG_PAGE_DOWNLOAD_CTRL_STS     0x0E
#define DIAG_PAGE_SUBENC_NN_CTRL_STS    0x0F

typedef struct SendDiagnostic {
    BYTE    Command;                    /* SCSI_SEND_DIAG (0x1D)            */
    BYTE    Flags;
        #define SNDDF_UNIT_OFFLINE      0x01
        #define SNDDF_DEV_OFFLINE       0x02
        #define SNDDF_SELFTEST          0x04
        #define SNDDF_PG_FORMAT         0x10
        #define SNDDF_ST_CODE_MASK      0xE0    
        #define SNDDF_ST_CODE_BG_SHORT  0x20
        #define SNDDF_ST_CODE_BG_EXT    0x40
        #define SNDDF_ST_CODE_ABORT_BG  0x80
        #define SNDDF_ST_CODE_FG_SHORT  0xA0
        #define SNDDF_ST_CODE_FG_EXT    0xC0

    BYTE    Reserved;
    BYTE    Length[2];
    BYTE    Control;
} CDB_SEND_DIAG, *PCDB_SEND_DIAG;

typedef struct ReceiveDiagnosticResults {
    BYTE    Command;                    /* SCSI_RCV_DIAG (0x1C)             */
    BYTE    Flags;
        #define RCVDRF_PCV          0x01

    BYTE    PageCode;
    BYTE    Length[2];
    BYTE    Control;
} CDB_RCV_DIAG_RSLTS, *PCDB_RCV_DIAG_RSLTS;

typedef struct DiagnosticHdr {
    BYTE    PageCode;
    BYTE    PageSpecific;
    WORD    Length;                  /* Length of data that follows      */
} DIAG_PARAM_HDR, *PDIAG_PARAM_HDR;


/****************************************************************************
                        ATA PASS-THROUGH COMMANDS
*****************************************************************************/

typedef struct AtaPassThrough12 {
    BYTE    OpCode;
    BYTE    ProtoCnt;
    BYTE    Flags;
    BYTE    Features;
    BYTE    SectorCnt;
    BYTE    LbaLow;
    BYTE    LbaMid;
    BYTE    LbaHigh;
    BYTE    Device;
    BYTE    Command;
    BYTE    Reserved;
    BYTE    Control;
} CDB_ATA_PASSTHRU12, *PCDB_ATA_PASSTHRU12;


typedef struct AtaPassThrough16 {
    BYTE    OpCode;
    BYTE    ProtoCnt;
        #define APT_PC_CNT_MASK         0xE0
        #define APT_PC_CNT_SHIFT        5
        #define APT_PC_PROTO_MASK       0x1E
        #define APT_PC_PROTO_HWRST      0x00
        #define APT_PC_PROTO_SRST       0x02
        #define APT_PC_PROTO_NONDATA    0x06
        #define APT_PC_PROTO_PIO_IN     0x08
        #define APT_PC_PROTO_PIO_OUT    0x0A
        #define APT_PC_PROTO_DMA        0x0C
        #define APT_PC_PROTO_DMA_Q      0x0E
        #define APT_PC_PROTO_DIAG       0x10
        #define APT_PC_PROTO_DEV_RST    0x12
        #define APT_PC_PROTO_UDMA_IN    0x14
        #define APT_PC_PROTO_UDMA_OUT   0x16
        #define APT_PC_PROTO_FPDMA      0x18
        #define APT_PC_PROTO_RSP        0x1E
        #define APT_PC_EXTEND           0x01

    BYTE    Flags;
        #define APTF_LEN_MASK           0x03
        #define APTF_LEN_NONE           0x00
        #define APTF_LEN_FEAT           0x01
        #define APTF_LEN_SECT           0x02
        #define APTF_LEN_TPSIU          0x03
        #define APTF_BLOCK              0x04
        #define APTF_READ_DATA          0x08
        #define APTF_XFER_TYPE          0x10
        #define APTF_CHK_COND           0x20
        #define APTF_OFFLINE_MASK       0xC0
        #define APTF_OFFLINE_SHIFT      6

    BYTE    FeaturesExp;
    BYTE    Features;
    BYTE    SectorCntExp;
    BYTE    SectorCnt;
    BYTE    LbaLowExp;
    BYTE    LbaLow;
    BYTE    LbaMidExp;
    BYTE    LbaMid;
    BYTE    LbaHighExp;
    BYTE    LbaHigh;
    BYTE    Device;
    BYTE    Command;
    BYTE    Control;
} CDB_ATA_PASSTHRU16, *PCDB_ATA_PASSTHRU16;


/****************************************************************************
                     READ/WRITE BUFFER COMMANDS AND DATA
*****************************************************************************/

typedef struct ReadWriteBuffer {
    BYTE    OpCode;
    BYTE    Mode;
        #define RWB_MODE_MASK           0x1F
        #define RWB_MODE_COMBINED       0x00
        #define RWB_MODE_VENDORSPECIFIC 0x01
        #define RWB_MODE_DATA           0x02
        #define RWB_MODE_DESCRIPTOR     0x03
        #define RWB_MODE_MCODE          0x04
        #define RWB_MODE_MCODE_SAVE     0x05
        #define RWB_MODE_MCODE_OFFS     0x06
        #define RWB_MODE_MCODE_OFFSAVE  0x07
        #define RWB_MODE_ECHOBUFFER     0x0A
        #define RWB_MODE_ECHODESCRIPTOR 0x0B
        #define RWB_MODE_MCODE_OFFS_SSD 0x0D
        #define RWB_MODE_MCODE_OFFS_SD  0x0E
        #define RWB_MODE_ACT_MCODE      0x0F
        #define RWB_MODE_SPEC_MASK      0xE0
        #define RWB_MODE_SPEC_VSE_ACT   0x20
        #define RWB_MODE_SPEC_HR_ACT    0x40
        #define RWB_MODE_SPEC_PO_ACT    0x80

    BYTE    BufId;
    BYTE    BufOffset[3];
    BYTE    Length[3];
    BYTE    Control;
} CDB_READ_BUFFER, *PCDB_READ_BUFFER, CDB_WRITE_BUFFER, *PCDB_WRITE_BUFFER;


typedef struct ReadWriteBufferDataDesc {
    BYTE    OffsetBoundary;
    BYTE    BufCapacity[3];
} RWBUFF_DATA_DESC, *PRWBUFF_DATA_DESC;


/****************************************************************************
                     RESERVE/RELEASE COMMANDS AND DATA
*****************************************************************************/

typedef struct CdbReserveRelease6 {
    BYTE    OpCode;
    BYTE    Reserved[4];
    BYTE    Control;
} CDB_RESERVE6, *PCDB_RESERVE6,
  CDB_RELEASE6, *PCDB_RELEASE6;


typedef struct CdbReserveRelease10 {
    BYTE    OpCode;
    BYTE    Flags;
        #define RR10F_LONG_ID       0x02
        #define RR10F_3RD_PARTY     0x10

    BYTE    Reserved;
    BYTE    ThirdPartyId;
    BYTE    Reserved2[3];
    BYTE    Length[2];
    BYTE    Control;
} CDB_RESERVE10, *PCDB_RESERVE10,
  CDB_RELEASE10, *PCDB_RELEASE10;


/****************************************************************************
                 PERSISTENT RESERVE OUT/IN COMMANDS AND DATA
*****************************************************************************/

#define PR_SCOPE_MASK       0xF0
#define PR_SCOPE_LU         0x00
#define PR_TYPE_MASK        0x0F
#define PR_TYPE_WR_EXCL     0x01
#define PR_TYPE_EXCL_ACC    0x03
#define PR_TYPE_WR_EXCL_RO  0x05
#define PR_TYPE_EXCL_ACC_RO 0x06
#define PR_TYPE_WR_EXCL_AR  0x07
#define PR_TYPE_EXCL_ACC_AR 0x08


typedef struct CdbPersistentReserveIn {
    BYTE    OpCode;
    BYTE    ServiceAction;
        #define PRI_SVC_ACT_MASK        0x1F
        #define PRI_SVC_ACT_READ_KEYS   0x00
        #define PRI_SVC_ACT_READ_RSRV   0x01
        #define PRI_SVC_ACT_RPT_CAPS    0x02
        #define PRI_SVC_ACT_READ_STS    0x03

    BYTE    Reserved[5];
    BYTE    Length[2];
    BYTE    Control;
} CDB_PERS_RSRV_IN, *PCDB_PERS_RSRV_IN;


typedef struct CdbPersistentReserveOut {
    BYTE    OpCode;
    BYTE    ServiceAction;
        #define PRO_SVC_ACT_MASK        0x1F
        #define PRO_SVC_ACT_REGISTER    0x00
        #define PRO_SVC_ACT_RESERVE     0x01
        #define PRO_SVC_ACT_RELEASE     0x02
        #define PRO_SVC_ACT_CLEAR       0x03
        #define PRO_SVC_ACT_PREEMPT     0x04
        #define PRO_SVC_ACT_PREEMPT_ABT 0x05
        #define PRO_SVC_ACT_REG_IGNORE  0x06
        #define PRO_SVC_ACT_REG_MOVE    0x07

    BYTE    ScopeType;
    BYTE    Reserved[4];
    BYTE    Length[2];
    BYTE    Control;
} CDB_PERS_RSRV_OUT, *PCDB_PERS_RSRV_OUT;


typedef BYTE PR_RSRV_KEY[8], *PPR_RSRV_KEY;


typedef struct PriReadHdr {
    DWORD   PrGeneration;
    DWORD   Length;
} PRI_READ_HDR, *PPRI_READ_HDR;


typedef struct PriReadRegisteredKeys {
    PRI_READ_HDR    Hdr;
    PR_RSRV_KEY     Keys[1];
} PRI_READ_REG_KEYS, *PPRI_READ_REG_KEYS;


typedef struct PriReadReservation {
    PRI_READ_HDR    Hdr;
    PR_RSRV_KEY     Key;
    BYTE            Reserved[5];
    BYTE            ScopeType;
    BYTE            Reserved2[2];
} PRI_READ_RSRV, *PPRI_READ_RSRV;


typedef struct PriReportCapabilities {
    WORD            Length;
    BYTE            Flags;
        #define PRIRCF_PTPL_C       0x01
        #define PRIRCF_ATP_C        0x04
        #define PRIRCF_SIP_C        0x08
        #define PRIRCF_CRH          0x10

    BYTE            Flags2;
        #define PRIRCF2_PTPL_A      0x01
        #define PRIRCF2_AC_MASK     0x70
        #define PRIRCF2_AC_MODE0    0x00
        #define PRIRCF2_AC_MODE1    0x10
        #define PRIRCF2_AC_MODE2    0x20
        #define PRIRCF2_AC_MODE3    0x30
        #define PRIRCF2_TMV         0x80

    WORD            TypeMask;
        #define PRIRC_TM_WR_EX      0x0002
        #define PRIRC_TM_EX_AC      0x0008
        #define PRIRC_TM_WR_EX_RO   0x0020
        #define PRIRC_TM_EX_AC_RO   0x0040
        #define PRIRC_TM_WR_EX_AR   0x0080
        #define PRIRC_TM_EX_AC_AR   0x0100
        
    BYTE            Reserved[2];
} PRI_RPT_CAPS, *PPRI_RPT_CAPS;


typedef struct PriStatusDescriptor {
    PR_RSRV_KEY     Key;
    BYTE            Reserved[4];
    BYTE            Flags;
        #define PRISDF_R_HOLDER     0x01
        #define PRISDF_ALL_TG_PT    0x02

    BYTE            ScopeType;
    BYTE            Reserved2[4];
    WORD            RelTargPortId;
    DWORD           AddLength;
    BYTE            TransportId[1];
} PRI_STS_DESC, *PPRI_STS_DESC;


typedef struct PriReadFullStatus {
    PRI_READ_HDR    Hdr;
    PRI_STS_DESC    StsDesc[1];
} PRI_READ_STS, *PPRI_READ_STS;


typedef struct ProParameterList {
    PR_RSRV_KEY     Key;
    PR_RSRV_KEY     SvcActKey;
    BYTE            Reserved[4];
    BYTE            Flags;
        #define PROPLF_APTPL        0x01
        #define PROPLF_ALL_TG_PT    0x04
        #define PROPLF_SPEC_I_PT    0x08

    BYTE            Reserved2[3];
} PRO_PARAM_LIST, *PPRO_PARAM_LIST;


typedef struct ProParameterListRegMove {
    PR_RSRV_KEY     Key;
    PR_RSRV_KEY     SvcActKey;
    BYTE            Reserved;
    BYTE            Flags;
        #define PROPLF_APTPL        0x01
        #define PROPLF_UNREG        0x02

    WORD            RelTargPortId;
    DWORD           AddLength;
    BYTE            TransportId[1];
} PRO_PARAM_LIST_RM, *PPRO_PARAM_LIST_RM;


/****************************************************************************
                         READ DEFECT DATA COMMANDS
*****************************************************************************/

typedef struct CdbRdDefectData10 {
    BYTE    Command;                        /* 00 RDD10 (0x37)              */
    BYTE    Reserved;                       /* 01 Reserved                  */
    BYTE    DefectListFmt;                  /* 02 List format, PLIST/GLIST  */
        #define RDD_DEFLSTFMT_MASK            0x1F
        #define RDD_DEFLSTFMT_SHORT_BLK       0x00
        #define RDD_DEFLSTFMT_XBYTES_FROM_IDX 0x01
        #define RDD_DEFLSTFMT_XSECTOR         0x02
        #define RDD_DEFLSTFMT_LONGBLK         0x03
        #define RDD_DEFLSTFMT_BYTES_FROM_IDX  0x04
        #define RDD_DEFLSTFMT_PHYS_SECTOR     0x05
        #define RDD_REQ_GLIST                 0x08
        #define RDD_REQ_PLIST                 0x10

    BYTE    Reserved2[4];                   /* 03 Reserved                  */
    BYTE    Length[2];                      /* 07 Allocation length         */
    BYTE    Control;                        /* 09 Control                   */
} CDB_RD_DEFECT10, *PCDB_RD_DEFECT10;       /* 0A Size of structure         */


typedef struct CdbRdDefectData12 {
    BYTE    Command;                        /* 00 RDD12 (0xB7)              */
    BYTE    DefectListFmt;                  /* 02 List format, PLIST/GLIST  */
    BYTE    AddrDescIx[4];                  /* 02 Addr. Descriptor Index    */
    BYTE    Length[4];                      /* 06 Allocation length         */
    BYTE    Reserved;                       /* 0A Reserved                  */
    BYTE    Control;                        /* 0B Control                   */    
} CDB_RD_DEFECT12, *PCDB_RD_DEFECT12;       /* 0C Size of structure         */


typedef struct Rdd10Hdr {
    BYTE    Reserved;                       /* 00 Reserved                  */
    BYTE    DefectListFmt;                  /* 01 List format, PLIST/GLIST  */
    WORD    Length;                         /* 02 Length                    */
} RDD10_HDR, *PRDD10_HDR;                   /* 04 Size of structure         */


typedef struct Rdd12Hdr {
    BYTE    Reserved;                       /* 00 Reserved                  */
    BYTE    DefectListFmt;                  /* 01 List format, PLIST/GLIST  */
    WORD    GenCode;                        /* 02 Generation code           */
    DWORD   Length;                         /* 04 Length                    */
} RDD12_HDR, *PRDD12_HDR;                   /* 08 Size of structure         */


/****************************************************************************
                          UNMAP COMMAND AND DATA
*****************************************************************************/

typedef struct CdbUnmap {
    BYTE    Command;                    /* SCSI_UNMAP   (0x42)              */
    BYTE    Flags;
        #define UNMAPF_ANCHOR   0x01 
    
    BYTE    Reserved[4];
    BYTE    GroupNum;
    BYTE    Length[2];
    BYTE    Control;
} CDB_UNMAP, *PCDB_UNMAP;


typedef struct UnmapDescHdr {
    WORD    Length;                     /* 00 Length following this field   */
    WORD    BlkDescLen;                 /* 02 Block descriptor length       */
    BYTE    Reserved[4];                /* 04 Reserved                      */
} UNMAP_DESC_HDR, *PUNMAP_DESC_HDR;


typedef struct UnmapDesc {
    QWORD   Lba;
    DWORD   NumBlocks;
    DWORD   Reserved;
} UNMAP_DESC, *PUNMAP_DESC;


/****************************************************************************
                          ZBC COMMANDS AND DATA
*****************************************************************************/

typedef struct CdbZbcOut {
    BYTE    Command;                    /* SCSI_ZBC_OUT (0x94)              */
    BYTE    ServiceAction;
        #define ZBC_OUT_SA_CLOSE_ZN     0x01
        #define ZBC_OUT_SA_FINISH_ZN    0x02
        #define ZBC_OUT_SA_OPEN_ZN      0x03
        #define ZBC_OUT_SA_RST_WR_PTR   0x04

    BYTE    ZnId[8];
    BYTE    Reserved[4];
    BYTE    Flags;
        #define ZBC_OUT_ALL             0x01 

    BYTE    Control;
} CDB_ZBC_OUT, *PCDB_ZBC_OUT;


typedef struct CdbZbcIn {
    BYTE    Command;                    /* SCSI_ZBC_IN (0x95)               */
    BYTE    ServiceAction;
        #define ZBC_IN_SA_RPT_ZNS       0x00

    BYTE    ZnStartLba[8];
    BYTE    Length[4];
    BYTE    Flags;
        #define ZBC_IN_RPT_OPT_MASK     0x3F
        #define ZBC_IN_RPT_OPT_ALL      0x00
        #define ZBC_IN_RPT_OPT_EMPTY    0x01
        #define ZBC_IN_RPT_OPT_IMP_OPEN 0x02
        #define ZBC_IN_RPT_OPT_EXP_OPEN 0x03
        #define ZBC_IN_RPT_OPT_CLOSED   0x04
        #define ZBC_IN_RPT_OPT_FULL     0x05
        #define ZBC_IN_RPT_OPT_RD_ONLY  0x06
        #define ZBC_IN_RPT_OPT_OFFLINE  0x07
        #define ZBC_IN_RPT_OPT_RWP_REC  0x10
        #define ZBC_IN_RPT_OPT_NSWRA    0x11
        #define ZBC_IN_RPT_OPT_NWP      0x3F
        #define ZBC_IN_PARTIAL          0x80 

    BYTE    Control;
} CDB_ZBC_IN, *PCDB_ZBC_IN;


typedef struct RptZonesDesc {
    BYTE    ZnType;                     /* 00 Zone type                     */
        #define RZD_ZT_TYPE_MASK        0x0F
        #define RZD_ZT_TYPE_CONV        0x01    /* Conventional             */
        #define RZD_ZT_TYPE_SWR         0x02    /* Sequential write required*/
        #define RZD_ZT_TYPE_SWP         0x03    /* Sequential write preferrd*/

    BYTE    ZnAttr;                     /* 01 Zone attributes               */
        #define RZD_ZA_RESET            0x01
        #define RZD_ZA_NON_SEQ          0x02
        #define RZD_ZA_COND_MASK        0xF0
        #define RZD_ZA_COND_NWP         0x00
        #define RZD_ZA_COND_EMPTY       0x10
        #define RZD_ZA_COND_IMP_OPEN    0x20
        #define RZD_ZA_COND_EXP_OPEN    0x30
        #define RZD_ZA_COND_CLOSED      0x40
        #define RZD_ZA_COND_RD_ONLY     0xD0
        #define RZD_ZA_COND_FULL        0xE0
        #define RZD_ZA_COND_OFFLINE     0xF0

    /* Note that drivers dereference the reserved fields! */

    BYTE    Reserved[6];                /* 02 Reserved                      */
    QWORD   ZnLen;                      /* 08 Zone length                   */
    QWORD   ZnStartLba;                 /* 10 Zone Start LBA                */
    QWORD   WrPtrLba;                   /* 18 Write pointer LBA             */
    BYTE    Reserved2[0x20];            /* 20 Reserved                      */
} RPT_ZONES_DESC, *PRPT_ZONES_DESC;


typedef struct RptZonesData {
    DWORD   ZnDescLen;                  /* 00 Length of zone descriptors    */
    BYTE    ZnDescAttr;                 /* 04 Zone descriptor attribute     */
        #define RZD_ZDA_SAME_MASK       0x0F
        #define RZD_ZDA_SAME_DT_DL      0x00    /* Different type and length*/
        #define RZD_ZDA_SAME_ST_SL      0x01    /* Same type and length     */
        #define RZD_ZDA_SAME_ST_DL      0x02    /* Same type, diff length   */
        #define RZD_ZDA_SAME_DT_SL      0x03    /* Diff type, same length   */

    /* Note that drivers dereference the reserved fields! */

    BYTE    Reserved[3];                /* 05 Reserved                      */
    QWORD   MaxLba;                     /* 08 Maximum LBA                   */
    BYTE    Reserved2[0x30];            /* 10 Reserved                      */

    RPT_ZONES_DESC ZnDesc[1];           /* 40 Zone descriptor list          */
} RPT_ZONES_DATA, *PRPT_ZONES_DATA;


/****************************************************************************
                                 SCSI CDB
*****************************************************************************/

#define SCSI_CDB_CTRL_LINK  0x01
#define SCSI_CDB_CTRL_NACA  0x04

typedef union scsicdb {
    BYTE    Byte [16];
    WORD    Word [ 8];

    BYTE    byte [16];              /* Old field */
    WORD    word [ 8];              /* Old field */

    struct  CMD6 {
        BYTE    Command;            /* 00   */
        BYTE    LUN_Block;          /* 01   */
        WORD    Block;              /* 02   */
        BYTE    Length;             /* 04   */
        BYTE    Control;            /* 05   */
    } cmd6;

    struct  CMD {
        BYTE    Command;            /* 00   */
        BYTE    LUN;                /* 01   */
        DWORD   Block;              /* 02   */
        BYTE    Reserved;           /* 06   */
        WORD    Length;             /* 07   */
        BYTE    Control;            /* 09   */
    } cmd;                          /* 10   */

    struct CMD16
    {
        BYTE    Command;
        BYTE    LUN;
        QWORD   Block;
        DWORD   Length;
        BYTE    GroupNumber;
        BYTE    Control;
    }cmd16;

    CDB_INQUIRY         inq;
    CDB_START_UNIT      strt;
    CDB_READ10          rd10;
    CDB_WRITE10         wr10;
    CDB_READ16          rd16;
    CDB_WRITE16         wr16;
    CDB_READ_CAP10      cap;
    CDB_READ_CAP10      cap10;
    CDB_READ_CAP16      cap16;
    CDB_MODE_SENSE6     msns6;
    CDB_MODE_SENSE10    msns10;
    CDB_MODE_SELECT6    msel6;
    CDB_MODE_SELECT10   msel10;
    CDB_SYNC_CACHE10    sc10;
    CDB_SYNC_CACHE16    sc16;
    CDB_REPORT_LUNS     rptluns;
    CDB_SEND_DIAG       sndDiag;
    CDB_RCV_DIAG_RSLTS  rcvDiag;
    CDB_PERS_RSRV_IN    perRsrvIn;
    CDB_PERS_RSRV_OUT   perRsrvOut;
    CDB_COMP_WRITE      cmpWr;

} SCSICDB, *PSCSICDB, FAR *LPSCSICDB;


#pragma pack ()


#endif
