/****************************************************************************

DESCRIPTION: Hardware-dependent data type definitions

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 1996-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef HWTYPES_H
#define HWTYPES_H

/* Make sure the OS types are defined first to override anything in here */

#include <ostypes.h>

/* Definitions that can be overridden by the OS layer   */

#ifndef FAR
#define FAR
#endif

#ifndef NEAR
#define NEAR
#endif

#ifndef HUGE_KW
#define HUGE_KW
#endif

#ifndef VOL
#define VOL     volatile
#endif

#ifndef FAST
#define FAST
#endif

#ifndef INLINE
#define INLINE  static inline
#endif

#ifndef UNALIGNED
#define UNALIGNED
#endif

#ifndef VOID
#define VOID        void
#endif

#ifndef OS_CHAR
#define OS_CHAR     char
#endif

#ifndef OS_WCHAR
#define OS_WCHAR    unsigned short
#endif

#ifndef OS_TCHAR
#define OS_TCHAR    char
#endif

#ifndef OS_STRCH
#define OS_STRCH    OS_CHAR
#endif

#ifndef OS_STRWCH
#define OS_STRWCH   OS_WCHAR
#endif

#ifndef OS_STRTCH
#define OS_STRTCH   OS_TCHAR
#endif

#ifndef OS_SINT8
#define OS_SINT8    signed char
#endif

#ifndef OS_SINT16
#define OS_SINT16   signed short
#endif

#ifndef OS_SINT32
#define OS_SINT32   signed int
#endif

#ifndef OS_SINT64
#define OS_SINT64   signed long long
#endif

#ifndef OS_UINT8
#define OS_UINT8    unsigned char
#endif

#ifndef OS_UINT16
#define OS_UINT16   unsigned short
#endif

#ifndef OS_UINT32
#define OS_UINT32   unsigned int
#endif

#ifndef OS_UINT64
#define OS_UINT64   unsigned long long
#endif

#ifndef OS_BOOL
#define OS_BOOL     int
#endif

#ifndef OS_SIZE_T
#define OS_SIZE_T   unsigned long
#endif

#ifndef OS_GUID
#define OS_GUID     struct { BYTE Byte[16]; }
#endif

#ifndef TRUE
#define TRUE        (1)
#endif

#ifndef FALSE
#define FALSE       (0)
#endif

#ifndef _T
#define _T(x) x
#endif

/* Define the HW types based on the OS types */

typedef OS_CHAR             CHAR;
typedef OS_WCHAR            WCHAR;
typedef OS_TCHAR            TCHAR;
typedef OS_UINT8            BYTE;
typedef OS_UINT16           WORD;
typedef OS_UINT32           DWORD;
typedef OS_UINT64           QWORD;
typedef OS_SIZE_T           SIZE_T;
typedef OS_GUID             GUID;

typedef OS_SINT8            SBYTE;
typedef OS_SINT16           SWORD;
typedef OS_SINT32           SDWORD;
typedef OS_SINT64           SQWORD;

typedef OS_BOOL             BOOL;

typedef const VOID          CNVOID;
#define CNCHAR              const CHAR
typedef const WCHAR         CNWCHAR;
#define CNTCHAR             const TCHAR
#define CNBYTE              const BYTE

/* Pointers */

typedef VOID                *PVOID;
typedef CNVOID              *PCNVOID;
typedef OS_CHAR             *PCHAR;
typedef OS_WCHAR            *PWCHAR;
typedef OS_TCHAR            *PTCHAR;
typedef OS_STRCH            *PSTR;
typedef OS_STRWCH           *PWSTR;
typedef OS_STRTCH           *PTSTR;
typedef CNCHAR              *PCNSTR;
typedef CNWCHAR             *PCNWSTR;
typedef CNTCHAR             *PCNTSTR;
typedef CNBYTE              *PCNBYTE;
typedef BYTE                *PBYTE;
typedef WORD                *PWORD;
typedef DWORD               *PDWORD;
typedef QWORD               *PQWORD;
typedef SBYTE               *PSBYTE;
typedef SWORD               *PSWORD;
typedef SDWORD              *PSDWORD;
typedef SQWORD              *PSQWORD;
typedef BOOL                *PBOOL;
typedef GUID                *PGUID;
typedef void                (*PFUNCTION)( void );

typedef VOID      FAR       *LPVOID;
typedef OS_CHAR   FAR       *LPCHAR;
typedef OS_STRCH  FAR       *LPSTR;
typedef OS_STRWCH FAR       *LPWSTR;
typedef CNCHAR    FAR       *LPCNSTR;
typedef CNWCHAR   FAR       *LPCNWSTR;
typedef BYTE      FAR       *LPBYTE;
typedef WORD      FAR       *LPWORD;
typedef DWORD     FAR       *LPDWORD;
typedef QWORD     FAR       *LPQWORD;
typedef BOOL      FAR       *LPBOOL;
typedef GUID      FAR       *LPGUID;
                          
typedef VOID      HUGE_KW   *HPVOID;
typedef CHAR      HUGE_KW   *HPCHAR;
typedef CHAR      HUGE_KW   *HPSTR;
typedef WCHAR     HUGE_KW   *HPWSTR;
typedef CNCHAR    HUGE_KW   *HPCNSTR;
typedef CNWCHAR   HUGE_KW   *HPCNWSTR;
typedef BYTE      HUGE_KW   *HPBYTE;
typedef WORD      HUGE_KW   *HPWORD;
typedef DWORD     HUGE_KW   *HPDWORD;
typedef QWORD     HUGE_KW   *HPQWORD;
typedef BOOL      HUGE_KW   *HPBOOL;
typedef GUID      HUGE_KW   *HPGUID;
                          
typedef VOID      VOL       *PVVOID;
typedef CHAR      VOL       *PVCHAR;
typedef CHAR      VOL       *PVSTR;
typedef WCHAR     VOL       *PVSTRW;
typedef CNCHAR    VOL       *PVCNSTR;
typedef CNWCHAR   VOL       *PVCNSTRW;
typedef BYTE      VOL       *PVBYTE;
typedef WORD      VOL       *PVWORD;
typedef DWORD     VOL       *PVDWORD;
typedef QWORD     VOL       *PVQWORD;
typedef BOOL      VOL       *PVBOOL;
                          
/* Pointers to Pointers */

typedef PVOID               *PPVOID;
typedef PCHAR               *PPCHAR;
typedef PSTR                *PPSTR;
typedef PWSTR               *PPWSTR;
typedef PBYTE               *PPBYTE;
typedef PWORD               *PPWORD;
typedef PDWORD              *PPDWORD;
typedef PQWORD              *PPQWORD;

typedef PVOID     VOL       *PVPVOID;
typedef PCHAR     VOL       *PVPCHAR;
typedef PSTR      VOL       *PVPSTR;
typedef PWSTR     VOL       *PVPWSTR;
typedef PBYTE     VOL       *PVPBYTE;
typedef PWORD     VOL       *PVPWORD;
typedef PDWORD    VOL       *PVPDWORD;
typedef PQWORD    VOL       *PVPQWORD;


/* Default some more types that may be derived from above */

#ifndef OS_LOCK_STATE
#define OS_LOCK_STATE   int
#endif

#ifndef OS_DPC_STATE
#define OS_DPC_STATE    int
#endif

#ifndef OS_SPIN_LOCK
#define OS_SPIN_LOCK    PVOID
#endif

#ifndef OS_DBG_LOCK
#define OS_DBG_LOCK     OS_SPIN_LOCK
#endif

#ifndef OS_PBASEADDR
#define OS_PBASEADDR    OS_UINT8 *
#endif

typedef OS_LOCK_STATE       LOCK_STATE, *PLOCK_STATE;
typedef OS_DPC_STATE        DPC_STATE,  *PDPC_STATE;
typedef OS_SPIN_LOCK        SPIN_LOCK;
typedef SPIN_LOCK *         PSPIN_LOCK;
typedef OS_DBG_LOCK         DBG_LOCK;
typedef DBG_LOCK *          PDBG_LOCK;

typedef OS_PBASEADDR        PBASEADDR;


/* The following are so we can use hwpnp.h and have the sizes and offsets   */
/* in structures correct.                                                   */

#ifdef OS_16BIT
typedef VOID (FAR  *FARPROC)  (VOID);
typedef VOID (NEAR *NEARPROC) (VOID);
typedef BYTE NEAR  *NPBYTE;
typedef NEARPROC    NEARPROC16;
typedef FARPROC     FARPROC16;
#else
#define FARPROC16   DWORD
#define NEARPROC16  WORD
#define NPBYTE      WORD
#endif

/* Fibre Channel Worldwide Name */

typedef union WorldWideName {
    BYTE    c[8];
    BYTE    b[8];
    WORD    w[4];
    DWORD   d[2];

    /* We would like to add the following field, but there are several      */
    /* structures that include this type that will be realigned.  These     */
    /* include target mode IOCTLs that must maintain their format for       */
    /* compatibility.  Instead of finding and packing all such structures,  */
    /* don't define the QWORD member.                                       */

    /* QWORD   q; */
} WWN, *PWWN;

/* SAS Address */

typedef union SasAddress {
    BYTE    c[8];
    BYTE    b[8];
    WORD    w[4];
    DWORD   d[2];
    QWORD   q;
} SASADDR, *PSASADDR;

/* MAC (Ethernet) Address */

typedef union MacAddress {
    BYTE    c[6];
    BYTE    b[6];
    WORD    w[3];
} MACADDR, *PMACADDR;

/* PCIe Address - Located in the serial number extended PCIe configuration  */
/* space header.                                                            */

typedef union PciAddress {
    BYTE    c[8];
    BYTE    b[8];
    WORD    w[4];
    DWORD   d[2];
    QWORD   q;
} PCIADDR, *PPCIADDR;

/* Define NULL for kicks */

#ifndef NULL
#define NULL ((VOID *) 0)
#endif

/* WORD component macros */

#ifndef LOBYTE
#define LOBYTE(w)   ((BYTE) (WORD) (w))
#endif

#ifndef HIBYTE
#define HIBYTE(w)   ((BYTE) (((WORD) (w)) >> 8))
#endif

#ifndef MAKEWORD
#define MAKEWORD(lo, hi) ((WORD) ((BYTE) (lo) | ((WORD) (BYTE) (hi) << 8)))
#endif

/* DWORD component macros */

#ifndef LOWORD
#define LOWORD(d)   ((WORD) (DWORD) (d))
#endif

#ifndef HIWORD
#define HIWORD(d)   ((WORD) (((DWORD) (d)) >> 16))
#endif

#ifndef MAKEDWORD
#define MAKEDWORD(lo, hi) ((DWORD) ((WORD) (lo) | ((DWORD) (WORD) (hi) << 16)))
#endif

/* QWORD component macros. Don't use right-shift due to the inefficient     */
/* code produced.                                                           */

#ifndef LODWORD
#ifdef OS_BIG_ENDIAN
#define LODWORD(pq)     (*((LPDWORD) (&(pq)) + 1))
#else
#define LODWORD(pq)     (*((LPDWORD) (&(pq)) + 0))
#endif
#endif

#ifndef HIDWORD
#ifdef OS_BIG_ENDIAN
#define HIDWORD(pq)     (*((LPDWORD) (&(pq)) + 0))
#else
#define HIDWORD(pq)     (*((LPDWORD) (&(pq)) + 1))
#endif
#endif

#ifndef LOCNDWORD
#define LOCNDWORD(q)    ((DWORD)(QWORD)(q))
#endif

#ifndef HICNDWORD
#define HICNDWORD(q)    ((DWORD)(((QWORD) (q)) >> 32))
#endif

#ifndef MAKEQWORD
#define MAKEQWORD(lo, hi) \
    ((QWORD) ((DWORD) (lo) | ((QWORD) (DWORD) (hi) << 32)))
#endif

#ifndef MAKEQUAD
#define MAKEQUAD(lo, hi) \
    ((QWORD) ((DWORD) (lo) | ((QWORD) (DWORD) (hi) << 32)))
#endif

/* Types and macros for physical addresses */

#ifndef _DOS_

  typedef QWORD     PHYSADDR;

  #ifndef MAKEPHYSADDR
    #define MAKEPHYSADDR(lo, hi)    ((PHYSADDR) MAKEQWORD (lo, hi))
  #endif
  #ifndef LOPHYSADDR
    #define LOPHYSADDR(pa)          LODWORD (pa)
  #endif
  #ifndef HIPHYSADDR
    #define HIPHYSADDR(pa)          HIDWORD (pa)
  #endif

#else   /* _DOS_ */

  typedef DWORD     PHYSADDR;

  #ifndef MAKEPHYSADDR
    #define MAKEPHYSADDR(lo, hi)    ((PHYSADDR) ((DWORD) (lo)))
  #endif
  #ifndef LOPHYSADDR
    #define LOPHYSADDR(pa)          ((DWORD) (pa))
  #endif
  #ifndef HIPHYSADDR
    #define HIPHYSADDR(pa)          (0)
  #endif

#endif  /* !_DOS_ */

typedef PHYSADDR                *PPHYSADDR;
typedef PHYSADDR    FAR         *LPPHYSADDR;
typedef PHYSADDR    HUGE_KW        *HPPHYSADDR;


/* Macro to get the lowest nonzero bit of a value */

#ifndef LOBIT
#define LOBIT(x)        ((x) & (0 - (x)))
#endif

/* Used on unreferenced function parameters */

#ifndef UNUSED_ARG
#define UNUSED_ARG(x)
#endif

/* Gets the number of elements in an array */

#ifndef numof
#define numof(x)        (sizeof (x) / sizeof (*x))
#endif

/* Signed/unsigned integers of pointer size */

#if defined(OS_64BIT)
typedef QWORD UINTPTR;
typedef QWORD UINTLPTR;
typedef SQWORD SINTPTR;
#elif defined(OS_32BIT)
typedef DWORD UINTPTR;
typedef DWORD UINTLPTR;
typedef SDWORD SINTPTR;
#elif defined(OS_16BIT)
typedef WORD  UINTPTR;
typedef DWORD UINTLPTR;
typedef SWORD SINTPTR;
#else
#error You must specify an OS_xxBIT compilation switch
#endif

/* Gets the offset of a structure field */

#ifndef _WATCOM_
#ifndef offsetof
#define offsetof(type, field) ((DWORD)(UINTPTR)&((type *)0)->field)
#endif
#endif

/* Gets the address of the structure containing the specified field */

#ifndef containerof
#define containerof(ptr, type, field) ((type *)((char *) (ptr) - \
					     (char *) &((type *)0)->field))
#endif

/* Gets the remaining size of a structure starting at the specified field */

#ifndef remainderof
#define remainderof(type, field) (sizeof(type) - offsetof(type, field))
#endif

/* For macros of DWORD size */

#define DW(dw)  ((DWORD)(dw))

/* Pointer conversion macros */

#define POINTER_TO_DWORD(x)    ((DWORD) (UINTPTR) (x))

/* Stringify and tokenize macros.  The layer of indirection allows macros   */
/* to be used as arguments.  Do not use the '2' versions in code.           */

#define STRINGIFY2(x)   #x
#define STRINGIFY(x)    STRINGIFY2(x)
#define TOKENIZE2(x,y)  x##y
#define TOKENIZE(x,y)   TOKENIZE2(x,y)


/****************************************************************************
                            Types for TSDK
*****************************************************************************/

typedef BYTE    bit8;
typedef WORD    bit16;
typedef DWORD   bit32;
typedef QWORD   bit64;

typedef SBYTE   sbit8;
typedef SWORD   sbit16;
typedef SDWORD  sbit32;
typedef SQWORD  sbit64;

typedef UINTPTR bitptr;

#ifndef GLOBAL
#define GLOBAL
#endif

#ifndef LOCAL
#define LOCAL
#endif

#endif
